/*
 * PAM ADLDAP module
 *
 * Description: PAM LDAP module for user authentication. This
 * module plugs into the existing PAM infrastructure.
 *
 */
#include <limits.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <netdb.h>
#include <security/pam_modules.h>

#include <sys/wait.h>
#include "pam_adldap_auth.h"
#include <sys/types.h>
#include <grp.h>

/* internal data */
static char conf_file[BUFFER_SIZE]; /* configuration file */
static adldap_server_t *live_server = NULL;
static int log_level = LOG_DEBUG;
int map_role(char* ldap_server_role, char* switch_role);

typedef unsigned long admap_t[ACCT_MAX_AD>>5];
int gid = -1;
#define CHASSIS_ID 255
/*
 * int _pam_ldap_log2()
 *
 * Description:
 * Logging function for this PAM module
 *
 * Input(s):
 * func	- Function name
 * line - Line number on which this is called
 * err  - Logging level
 * format - Content to be logged
 *
 * Output(s):
 *  NA
 *
 * Returns:
 * None
 */
static void 
_pam_ldap_log2(const char *func, int line, int err,
					  CONST char *format, ...)
{
	va_list args;
	char buffer[BUFFER_SIZE];

	va_start(args, format);
	vsprintf(buffer, format, args);
	if (err <= log_level) {
		fprintf(stderr, "%s(%d): %s\n", func, line, buffer);
	}
	va_end(args);
}

#define	PAM_LDAP_LOG(err, format...)				\
do {										\
	_pam_ldap_log2(__FUNCTION__, __LINE__, err, format);	\
} while (0)

/*
 * _pam_parse()
 *
 * Description:
 * This function parses the arguements passed on to
 * the pam module
 *
 * Input(s):
 * argc - Number of arguements
 * argv - Aruguements to PAM module
 * conf - configuration information
 *
 * Output(s):
 *  NA
 *
 * Returns:
 * 0 on success
 * Appropriate PAM error in case of failure
 */
static int 
_pam_parse(int argc, CONST char **argv, adldap_conf_t *conf)
{
	int ctrl=0;

	memset(conf, 0, sizeof(adldap_conf_t)); /* ensure it's initialized */

	strncpy(conf_file, CONF_FILE, sizeof(conf_file));

	/*
	 *  If either is not there, then we can't parse anything.
	 */
	if ((argc == 0) || (argv == NULL)) {
		return PAM_ABORT;
	}
  
	/* step through arguments */
	for (ctrl=0; argc-- > 0; ++argv) {

		/* generic options */
		if (!strncmp(*argv,"conf=",5)) {
			strncpy(conf_file, *argv+5, sizeof(conf_file));
		} else if (!strcmp(*argv, "use_first_pass")) {
			ctrl |= PAM_USE_FIRST_PASS;
		} else if (!strcmp(*argv, "try_first_pass")) {
			ctrl |= PAM_TRY_FIRST_PASS;
		} else if (!strcmp(*argv, "skip_passwd")) {
			ctrl |= PAM_SKIP_PASSWD;
		} else if (!strncmp(*argv,"retry=",6)) {
			int i = atoi(*argv+6);
			i &= 0x03;		/* keep the low 3 bits only */
			conf->retries = i;
			ctrl |= (i << 4);
		} else if (!strncmp(*argv, "client_id=", 10)) {
			if (conf->client_id) {
			PAM_LDAP_LOG(LOG_WARNING, "ignoring duplicate '%s'", *argv);
			} else {
				conf->client_id = (char *) *argv+10; /* point to the client-id */
			}
		} else if (!strcmp(*argv, "debug")) {
			ctrl |= PAM_DEBUG_ARG;
			conf->debug = 1;
		} else {
			PAM_LDAP_LOG(LOG_WARNING, "unrecognized option '%s'", *argv);
		}
	}

	return ctrl;
}

/*
 * _int_free()
 *
 * Description:
 * PAM callback function for freeing the memory
 *
 * Input(s):
 * pamh - Pointer to PAM handle
 * x	-  Memory to be freed
 * error_status - Error status
 *
 * Output(s):
 *  NA
 *
 * Returns:
 * None
 */
void
_int_free(pam_handle_t * pamh, void *x, int error_status)
{
    free(x);
}

/*
 * _char_free()
 *
 * Description:
 * PAM callback function for freeing the memory
 *
 * Input(s):
 * pamh - Pointer to PAM handle
 * x	-  Memory to be freed
 * error_status - Error status
 *
 * Output(s):
 *  NA
 *
 * Returns:
 * None
 */
void
_char_free(pam_handle_t *pamh, void *x, int error_status)
{
    free(x);
}

static int
setVfForRoleMap(char *rolemap,int *vfRoleMap)
{
        char    *p = NULL, *temp = NULL;
        char    *tempmap = NULL;
        int     ret = 0,i = 0,gid = -1;
        char    buffer[SMALL_BUFFER_SIZE];
        char    roleVfbuffer[SMALL_BUFFER_SIZE];
        char    *roleVfmap = NULL;
        char    *vf_list = NULL;
        static int len = 0;
        int     range_val1, range_val2, vfIndex;
	char *tmprole = NULL;

	if (!rolemap ||
		!vfRoleMap) {
		goto error;
	}

        PAM_LDAP_LOG(LOG_DEBUG,"rolemap %s\n",rolemap);
	memset(buffer,0,SMALL_BUFFER_SIZE);
	strcpy(buffer,rolemap);
        tmprole = strtok(buffer,"=");
        PAM_LDAP_LOG(LOG_DEBUG,"tmprole %s\n",tmprole);

        while (tmprole[i] != '\0') {
	        tmprole[i] = tolower(tmprole[i]);
                i++;
        }

        gid = getgroupid(tmprole);
        PAM_LDAP_LOG(LOG_DEBUG,"role %s gid %d\n",tmprole,gid);
        if (gid == INVALID_ROLE) {
	       goto error; 
        }

	tempmap = strtok(NULL,",");
        while(tempmap != NULL) {
		PAM_LDAP_LOG(LOG_DEBUG,"tmpmap %s\n",tempmap);
                if ((vf_list =  strchr(tempmap, '-')) != NULL) {
			PAM_LDAP_LOG(LOG_DEBUG,"tmpmap range\n");
                        *vf_list = '\0';
                        vf_list++;
                        if ((strchr(vf_list, '-')) != NULL) {
                                goto error;
                        }
                        if ((*vf_list == '\0') || (*tempmap == '\0')) {
                                goto error;
                        }

                        range_val1 = atoi(tempmap);
                        if ((range_val1 > PAM_ADLDAP_REAL_MAXVF) || (range_val1 < 1)) {
                                goto error;
                        }

                        range_val2 = atoi(vf_list);
                        if ((range_val2 > PAM_ADLDAP_REAL_MAXVF) || (range_val2 < 1)) {
                                goto error;
                        }
                        for (i = range_val1; i <= range_val2; i++) {
        	        	vfRoleMap[i] = gid;
                        }

                } else {
			PAM_LDAP_LOG(LOG_DEBUG,"tmpmap no\n");
                        vfIndex = atoi(tempmap);
                        if ((vfIndex > PAM_ADLDAP_REAL_MAXVF) || (vfIndex < 1)) {
                                goto error;
                        }
			vfRoleMap[vfIndex] = gid;
                }
                tempmap = strtok(NULL,",");
        }
        return (ret);
error:
        ret = -1;
        return (ret);
}


/*
 * get_host_port()
 *
 * Description:
 * Parses the hostname:port to provide separate server IP and port
 * as seperate entities
 *
 * Input(s):
 * str		- String containing hostname and port information
 *
 * Output(s):
 * hostname	- Hostname of the server
 * port		- Port of the server
 *
 * Returns:
 * None
 */
static void 
get_host_port(char* str, char* hostname, char* port)
{
#if 0
	char delims[] = ":";
	char *result = NULL, *strptr = NULL;
	int first = 1;

	result = strtok_r(str, delims, &strptr);

	/*
	 * Loop through all the server entries
	 */
	while(result != NULL) {
		if (first) {
			first = 0;
			strncpy(hostname, result, SMALL_BUFFER_SIZE);
		} else {
			strncpy(port, result, SMALL_BUFFER_SIZE);
		}
		result = strtok_r(NULL, delims, &strptr);
	} /*while ends*/
#endif
	char *pname = NULL;
	/* Search for the port number at the end of the hostname
	 * In case of IPv4, hostname will be of type a.b.c.d:port
	 * In case of IPv6, hostname will be a:b:c:d:e:f:g:h:port
	 *                  or with some combination of : and ::
	 *                  followed by port. 
	 * So, we always search for last colon
	 */
	
	pname = strrchr(str, ':');
	if (pname != NULL && isdigit(*(pname+1))) {
		strncpy(port, pname+1, SMALL_BUFFER_SIZE);
	}
	*pname = '\0';  /* end the host name without port */
	strncpy(hostname, str, SMALL_BUFFER_SIZE);
	*pname = ':';  /* revert original character back */
}

/*
 * get_basedn()
 *
 * Description:
 * This function parses the domain name of the server and 
 * creates the basedn as required by the ldapsearch APIs. 
 *
 * Input(s):
 * domain	- PAM handle
 *
 * Output(s):
 * basedn	- basedn of Active Directory server 
 *
 * Returns:
 * 0 on success
 * PAM error on failure
 */
int
get_basedn(char* domain, char* basedn)
{
	char *result = NULL;
	char delim[] = ".";
	char *strptr = NULL;
	char *localDomainVar = NULL; /*Local variable to be used in strtok_r() function */

	if (!domain) {
		return (-1);
	}
	/* 
	 * If it is already in dn=org,dc=com format, no conversion is neededm
	 * otherwise, add prepend all parts of abc.org.com with prefix dc e.g.
	 * abc.org.com becomes dc=abc,dc=org,dc=com
 	 */
	localDomainVar = strdup(domain);
	if(strchr(localDomainVar, '=') == NULL) {
		result = strtok_r(localDomainVar, delim, &strptr);
		strcpy(basedn, "dc=");
		strcat(basedn, result);

		while ( (result = (char *)strtok_r(NULL, delim, &strptr)) != NULL) {
			strcat(basedn, ",dc=");
			strcat(basedn, result);
		}
	} else {
		strcpy(basedn, domain);
	}
	if (localDomainVar != NULL) {
		free(localDomainVar);
		localDomainVar = NULL;
	}
	return 0;
}

/*
 * role_to_user()
 *
 * Description:
 * This function parses the output of the LDAP search and
 * extracts the user role
 *
 * Input(s):
 * pamh		- PAM handle
 * user		- User name
 *
 * Output(s):
 * role		- Role of the user
 *
 * Returns:
 * 0 on success
 * PAM error on failure
 */
static int
role_to_user(pam_handle_t *pamh, const char *user, char *role,
			char *ad_list, int *vfRoleMap, int homeVF, char* lf_id_list,
			char *lf_id, char* chassis_role)
{
	int	retval;
	char		cmd_buf[SMALL_BUFFER_SIZE + MAXPASS];
	const char	*pinst;
	char	adbuf[BUFFER_SIZE*2] = {0};
	char	homebuf[256] = {0};
	char	cmdbuf[BUFFER_SIZE*3];
	char	buffer[SMALL_BUFFER_SIZE];
	admap_t	admap = {0};
	unsigned long *mapdata;
	int		*pHome;
	int		homeAd = 0, roleValid;
	FILE	*pfile;
	char *strptr = NULL;
	char *rolemap = NULL;
	char *tmprole = NULL;
	char *tmpvfindex = NULL;
	char 	ldap_server_role[SMALL_BUFFER_SIZE], switch_role[SMALL_BUFFER_SIZE];
	char * vf_enabled_str = NULL, *end_ptr;
	int vf_enabled = 0;
	int *pHomeVF, novf_flag = 0, default_vf = -1;
	int noVfFlag = 0;
	char *result = NULL;
        int lf_id_min = 0;
        int lf_id_max = 0;
	int roleGid = -1;
	int i = 0;
	char *vfrange = NULL;

	/* Find whether VF is enabled or not from PAM envlist */
	vf_enabled_str = pam_getenv(pamh, "VF_ENABLED");
	if (vf_enabled_str) {
		/* In case vf environmental variable is not available and by
		 * default, it is assumed as non VF environment
		 */
		vf_enabled = strtol(vf_enabled_str , &end_ptr, 10);
	}
	PAM_LDAP_LOG(LOG_DEBUG, "VF_Enabled:%d", vf_enabled);
	PAM_LDAP_LOG(LOG_DEBUG, "User: %s Role: %s", user, role);
	PAM_LDAP_LOG(LOG_DEBUG, " lf_id_list: %s lf_id: %s chassis role:%s ",lf_id_list, lf_id, chassis_role);

	if (!pamh || !user || !role) {
		PAM_LDAP_LOG(LOG_DEBUG, "Returning error from inside role_to_user()..");
		return (PAM_AUTH_ERR);
	}

	memset(ldap_server_role, 0, SMALL_BUFFER_SIZE);
	memset(switch_role, 0, SMALL_BUFFER_SIZE);
	strncpy(ldap_server_role, role, strlen(role)+1);

	/*
	 * Get the switch instance from PAM envlist
	 */
	pinst = pam_getenv(pamh, "FABOS_SWITCHNO");
	if (pinst == NULL) {
		pinst = "0";
	}
 /*
 *  *      * Make sure the selected role is allowed by switch security policy.
 *   *           */
    snprintf(cmd_buf, sizeof (cmd_buf),
             "/fabos/libexec/secRoleCheck %s > /dev/null 2>&1",
             role);
    retval = system(cmd_buf);
    if ((retval != 0) && (errno != ECHILD)) {
        return (PAM_AUTH_ERR);
    }
    roleValid = WEXITSTATUS(retval);
    PAM_LDAP_LOG(LOG_DEBUG, "check role %s returns %d",
                role, roleValid);
    if ((strcmp(role, "root") == 0) || roleValid) {
        PAM_LDAP_LOG(LOG_DEBUG,"Invalid Role");
        return (PAM_AUTH_ERR);
    }


	retval = 0;

	if (!vf_enabled) {
		 PAM_LDAP_LOG(LOG_DEBUG,"NO VF ENABLE ");
		
		/* Fix for defect 640562 - Remote user with root role is not allowed */
		if (strcasecmp(role, "root") == 0) {
			return (INVALID_ROLE);
		}

		if (strlen(ad_list) > 0) {
			strncpy(adbuf, ad_list, strlen(ad_list));
    		strcpy(homebuf, strtok_r(ad_list, ",", &strptr));
			PAM_LDAP_LOG(LOG_DEBUG, "adbuf = %s homebuf = %s\n",
						adbuf, homebuf);
		}

		/*
	 	 * use popen to parse, convert & validate AD input buffer
	 	 */
		if (adbuf[0]) {
			if (homebuf[0]) {
				snprintf(cmdbuf, sizeof (cmdbuf),
					 "/fabos/libexec/secADParse -h \"%s\" -a \"%s\" "
					 "2> /dev/null", homebuf, adbuf);
			} else {
				snprintf(cmdbuf, sizeof (cmdbuf),
					 "/fabos/libexec/secADParse -a \"%s\" 2> /dev/null",
					 adbuf);
			}

			PAM_LDAP_LOG(LOG_DEBUG, "cmd: %s", cmdbuf);
			pfile = popen(cmdbuf, "r");
			if (!pfile) {
				PAM_LDAP_LOG(LOG_WARNING, "popen failed %d (%s)", errno,
							strerror(errno));
				retval = -1;
			} else {
				fscanf(pfile, "%lx:%lx:%lx:%lx:%lx:%lx:%lx:%lx",
						&admap[0], &admap[1], &admap[2], &admap[3],
						&admap[4], &admap[5], &admap[6], &admap[7]);
				fscanf(pfile, "%d", &homeAd);
				retval = pclose(pfile);
				/*
				 * If SIGCHLD is set to SIG_IGN, pclose will
				 * always return -1 and errno will be set to	
				 * ECHILD. So to ignore this case, if errno is
				 * ECHILD always set retval to 0
				 */
				if ((retval == -1) && (errno == ECHILD)) {
					retval = 0;
				}

				PAM_LDAP_LOG(LOG_DEBUG, 
							"admap: %lx:%lx:%lx:%lx:%lx:%lx:%lx:%lx, "
							"homeAD: %d", admap[0], admap[1], admap[2],
							admap[3], admap[4], admap[5], admap[6],
							admap[7], homeAd);
			}
		} else {
			/* empty AD member list, default to AD 0 */
			BM_SetBitN(0, admap);
			homeAd = 0;
			retval = PAM_SUCCESS;
		}

		/*
	 	 *If there is an error, just set AD bitmap to all zero
	 	 */
		if (retval) {
			memset(admap, 0, sizeof (admap));
			retval = PAM_SUCCESS;
		}

		/*
	 	 * set AD information to PAM data for other modules to use
	 	 */
		mapdata = calloc(sizeof (admap), sizeof(char));
		if (!mapdata) {
			retval = PAM_AUTH_ERR;
			return (retval);
		} else {
			memcpy(mapdata, admap, sizeof (admap));
			retval = pam_set_data(pamh, PAM_ADLDAP_ADMAP,
								(void *) mapdata, _char_free);
		}

		if (retval == PAM_SUCCESS) {
			pHome = calloc(sizeof(int), sizeof(char));
			if (!pHome) {
				retval = PAM_AUTH_ERR;
				return (retval);
			} else {
				*pHome = homeAd;
				retval = pam_set_data(pamh, PAM_ADLDAP_HOMEAD,
									(void *) pHome, _int_free);
				if (PAM_SUCCESS != retval) {
					return (retval);
				}
			}
		}
	} else {
		PAM_LDAP_LOG(LOG_DEBUG,"vf enabled\n");
		/* User's homeVF need to be added into the VF list. */
        if (homeVF != INVALID_VF) {
            vfRoleMap[homeVF] = gid;
        }
  /* If there is no vf list or pam error, add the default VF */

	
              	for (i = 0; i < PAM_ADLDAP_MAXVF - 1; i++) {
                        if (vfRoleMap[i] != INVALID_VF) {
				noVfFlag = 0;
				break;
			} else if (i == PAM_ADLDAP_REAL_MAXVF) {
                               	noVfFlag = 1;
				break;
			} else {
				continue;
			}
		}
		if (!noVfFlag ) {
  			snprintf(cmd_buf, sizeof (cmd_buf),
      				"/fabos/libexec/getDefaultFID");
  			retval = system(cmd_buf);
  			if (retval == -1) {
      			PAM_LDAP_LOG(LOG_ERR, "Getting default FID failed");
	  			retval = PAM_ABORT;
				return (retval);
			}
			default_vf = WEXITSTATUS(retval);
            vfRoleMap[default_vf] = gid;
            homeVF = default_vf;
        }

			if (lf_id_list[0] == 0 &&
				lf_id[0] == 0 &&
				chassis_role[0] == 0 &&
				homeVF == INVALID_VF &&
				vfRoleMap[CHASSIS_ID] == INVALID_VF ) {
				PAM_LDAP_LOG(LOG_DEBUG,"Both server list and ldap mapping NULL\n");
				roleGid = getgroupid(DEFAULT_ROLE);
				vfRoleMap[CHASSIS_ID] = roleGid;
				homeVF = default_vf;
				
		              	for (i = 0; i <= PAM_ADLDAP_REAL_MAXVF; i++) {
					vfRoleMap[i] = roleGid;
				}
			} else {
				strptr = NULL;
				strcpy(buffer,role);
				rolemap = strtok_r(buffer,";", &strptr);	
				while (rolemap != NULL) {
				
					setVfForRoleMap(rolemap,vfRoleMap);
					rolemap = strtok_r(NULL,";", &strptr);
				}
			}	
		
                if (homeVF == INVALID_VF) {
                                if (lf_id) {
                                        PAM_LDAP_LOG(LOG_DEBUG,"set LDAP HOME VF Map value %s \n",lf_id);
                                        homeVF = atoi(lf_id);

                                } else {
                                        PAM_LDAP_LOG(LOG_DEBUG,"set default HOME VF value %d\n",default_vf);
                                        homeVF = default_vf;
                                }
                }

		PAM_LDAP_LOG(LOG_DEBUG,"HOME VF value %d\n",homeVF);

		/*check for Chassis Role
		 *
		 */ 	
		if ( vfRoleMap[CHASSIS_ID] == INVALID_VF) {
			int chassisgid = getgroupid(chassis_role);
			PAM_LDAP_LOG(LOG_DEBUG,"chassis role %s chassis gid %d\n",chassis_role,chassisgid);
			vfRoleMap[CHASSIS_ID] = chassisgid;
		}
		 
		retval = pam_set_data(pamh, PAM_ADLDAP_VFROLEMAP, 
							(void *)vfRoleMap, _char_free);


		/* assign homeVf role to role 
		 */
		struct group    *grp = NULL;
		grp = getgrgid(vfRoleMap[homeVF]);
		if (grp) {
			PAM_LDAP_LOG(LOG_DEBUG,"role %s gid %d \n",grp->gr_name,vfRoleMap[homeVF]);
			strcpy(role,grp->gr_name);
		}

		PAM_LDAP_LOG(LOG_DEBUG, "LDAP account: %s %s %d %d",
				user, role, homeVF, vfRoleMap[homeVF]);

		

		if (retval == PAM_SUCCESS) {
			pHomeVF = malloc(sizeof(int));
			if (!pHomeVF) {
				retval = PAM_AUTH_ERR;
			} else {
				*pHomeVF = homeVF;
				PAM_LDAP_LOG(LOG_DEBUG,"pHomeVF set pam %d\n",*pHomeVF);
				retval = pam_set_data(pamh, PAM_ADLDAP_HOMEVF,
								 	(void *)pHomeVF, _int_free);
			}
		}
	}

	/*
	 * save the login name to ADLDAP server to data
	 */
	char *ouser = calloc( strlen(user) + 1, sizeof(char));
	if (ouser == NULL) {
		retval = PAM_AUTH_ERR;
		return (retval);
	} else {
		strncpy(ouser, user, strlen(user) + 1);
				PAM_LDAP_LOG(LOG_DEBUG,"ouser %s\n",ouser);
		retval = pam_set_data(pamh, PAM_ADLDAP_UNAME, (void *) ouser,
							  _char_free);
		if (PAM_SUCCESS != retval) {
			return (retval);
		}
	}


	/*
	 * change the user name to the one given by role
	 */
	if (retval == PAM_SUCCESS) {
		/*
		 * non-secure mode or non-FCS, convert nonfcsadmin role to admin
		 */
		if (strncmp(role, "nonfcsadmin", SMALL_BUFFER_SIZE) == 0) {
			snprintf(role, sizeof (role), "admin");
		}
		PAM_LDAP_LOG(LOG_DEBUG,"role %s\n",role);
		retval = pam_set_item(pamh, PAM_USER, role);
	}


	/*
	 * add the successful (live) server name to the PAM data
	 */
	if (retval == PAM_SUCCESS) {
		char hostname[SMALL_BUFFER_SIZE];
		char port[SMALL_BUFFER_SIZE];
		char *serv_name = NULL;
		if(NULL == live_server) {
			return retval;
		}
		get_host_port(live_server->hostname, hostname, port);
		serv_name = strdup(hostname);
		if (serv_name != NULL) {
			(void) pam_set_data(pamh, PAM_ADLDAP_LIVESERV,
								(void *)serv_name, _char_free);
		}
	}

	return (retval);
}

/*
 * dbg_init()
 *
 * Description:
 * This function reads the content of file /tmp/pamdebug.conf 
 * to determine the debug print out level.
 *
 * Input(s):
 * None
 *
 * Output(s):
 * None
 *
 * Returns:
 * None
 */
static void
dbg_init(void)
{
	FILE	*fdbg;

	fdbg = fopen("/tmp/pamdebug.conf", "r");
	if (fdbg == NULL) {
		log_level = LOG_EMERG;
		return;
	}
	if (fscanf(fdbg, "%d", &log_level) <= 0) {
		log_level = LOG_EMERG;
	} else if (log_level < 0) {
		log_level = LOG_EMERG;
	}
	(void) fclose(fdbg);
	return;
}

/*
 * cleanup()
 *
 * Description:
 * This function cleans up the memory allocated
 * for storing server related information
 *
 * Input(s):
 * server - List of servers configured for LDAP
 *
 * Output(s):
 * None
 *
 * Returns:
 * None
 */
static void
cleanup(adldap_server_t *server)
{
	adldap_server_t *next;
  
	while (server) {
		next = server->next;
		_pam_drop(server->hostname);
		_pam_drop(server->domain); 
		_pam_drop(server->uidSearchAttr);
		_pam_drop(server->clientDN);
		_pam_drop(server->password);
		_pam_drop(server);
		server = next;
	}
}

/*
 * initialize()
 *
 * Description:
 * This function allocates and open a local port for 
 * communication with the ADLDAP server
 *
 * Input(s):
 * conf - Pointer to the allocated memory to store 
 *			server details
 *
 * Output(s):
 * None
 *
 * Returns:
 * 0 on success
 * Approprite PAM error on failure
 */
static int
initialize(adldap_conf_t *conf)
{
	u_short local_port;
	char hostname[BUFFER_SIZE];
	char domain[BUFFER_SIZE];
	struct addrinfo hints, *res = NULL;
	int err;
	char port_str[PORT_STR_LEN]; 
	char buffer[BUFFER_SIZE];
	char *p;
	FILE *fserver;
	adldap_server_t *server = NULL;
	int timeout;
        int     bindType = BIND_LOGIN_ID;
        int     serverClass = LDAP_SERVER_LOCAL;
        char    uidSearchAttr[SMALL_BUFFER_SIZE];
        char    clientDN[BUFFER_SIZE];
        char    password[SMALL_BUFFER_SIZE];
	int line = 0, columnsRead = 0;

	/*
	 * the first time around, read the configuration file
	 */
	if ((fserver = fopen (conf_file, "r")) == (FILE*)NULL) {
		PAM_LDAP_LOG(LOG_ERR, "Could not open configuration file %s: %s\n",
			conf_file, strerror(errno));
		return PAM_ABORT;
	}

	while (!feof(fserver) &&
		(fgets (buffer, sizeof(buffer), fserver) != (char*) NULL) &&
		(!ferror(fserver))) {
		line++;
		p = buffer;

		/*
		 *  Skip blank lines and whitespace
		 */
		 while (*p &&
			((*p == ' ') || (*p == '\t') ||
			(*p == '\r') || (*p == '\n'))) p++;

		/*
		 *  Nothing, or just a comment.  Ignore the line.
		 */
		if ((!*p) || (*p == '#')) {
			continue;
		}
    
		timeout = 3;
		columnsRead = sscanf(p, "%s %d %s %d %d %s %s %s",
				hostname, &timeout, domain, &serverClass,
				&bindType, uidSearchAttr, clientDN, password);
		if ((columnsRead != 8 ) && (columnsRead != 3)) {
			PAM_LDAP_LOG(LOG_ERR, "ERROR reading %s, line %d: Could not read hostname \n",
					conf_file, line);
			continue; /* invalid line */
		} else {			/* read it in and save the data */
			adldap_server_t *tmp;

			tmp = calloc(sizeof(adldap_server_t), sizeof(char));
			if (NULL == tmp) {
				PAM_LDAP_LOG(LOG_ERR, "mem allocation failed for adldap_server \n");
				return PAM_ABORT;
			}

			if (server) {
				server->next = tmp;
				server = server->next;
			} else {
				conf->server = tmp;
				server= tmp;		/* first time */
			}

			/* sometime later do memory checks here */
			server->hostname = strdup(hostname);
			server->port = 0;
			server->domain = strdup(domain);
			server->uidSearchAttr = (char *)malloc(SMALL_BUFFER_SIZE);
			server->clientDN = (char *)malloc(SMALL_BUFFER_SIZE);
			server->password = (char *)malloc(SMALL_BUFFER_SIZE);
			if (server->uidSearchAttr == NULL ||
				 server->clientDN == NULL ||
				 server->password == NULL) {
				PAM_LDAP_LOG(LOG_ERR, "mem allocation failed for attributes\n");
				return PAM_ABORT;
			}
			if (columnsRead == 8 ) {
				server->bindingMethod = bindType;
				server->serverClass = serverClass;
				strncpy(server->uidSearchAttr, uidSearchAttr, SMALL_BUFFER_SIZE);
				strncpy(server->clientDN, clientDN, SMALL_BUFFER_SIZE);
				strncpy(server->password, password, SMALL_BUFFER_SIZE);
			} else {
                                server->bindingMethod = BIND_LOGIN_ID;
                                server->serverClass = LDAP_SERVER_EXTERNAL;
                                strncpy(server->uidSearchAttr, "userPrincipalName", SMALL_BUFFER_SIZE);
                                strncpy(server->clientDN, "tempOrg.com", SMALL_BUFFER_SIZE);
                                strncpy(server->password, "tempPass", SMALL_BUFFER_SIZE);
			}
			if ((timeout < 1) || (timeout > 60)) {
				server->timeout = 3;
			} else {
				server->timeout = timeout;
			}

			PAM_LDAP_LOG(LOG_INFO, "Hostname - %s, timeout - %d\n", hostname, timeout);
			PAM_LDAP_LOG(LOG_INFO, "Domain - %s\n", domain);

			server->next = NULL;
		}
	}
	fclose(fserver);

	if (!server) {		/* no server found, die a horrible death */
		PAM_LDAP_LOG(LOG_ERR, "No ADLDAP server found in configuration file %s\n",
			conf_file);
		return PAM_AUTHINFO_UNAVAIL;
	}

	/* open a socket.  Dies if it fails */

	conf->sockfd = socket(AF_INET6, SOCK_DGRAM, 0);
	if (conf->sockfd < 0) {
		PAM_LDAP_LOG(LOG_ERR, "Failed to open ADLDAP socket: %s\n", strerror(errno));
		return PAM_AUTHINFO_UNAVAIL;
	}

	/*
	 *  Use our process ID as a local port for ADLDAP.
	 */
	local_port = (getpid() & 0x7fff) + 1024;

	/* zero out hints before initializing */
	memset(&hints, 0, sizeof(hints));
	hints.ai_family = PF_UNSPEC; /* accept any protocol family */
	hints.ai_socktype = SOCK_DGRAM;

	/* 
	 * the IP address portion of the socket address structure will be set to 
	 * INADDR_ANY for an IPv4 address or IN6ADDR_ANY_INIT for an IPv6 address.
	 */
	hints.ai_flags = AI_PASSIVE; 
	hints.ai_protocol = 0;  /* accept any protocol */

	do {
		local_port++;
		snprintf(port_str, PORT_STR_LEN, "%d", htons(local_port));

		if (res) {
			freeaddrinfo(res);	
			res = NULL;
		}

		err = getaddrinfo(NULL, port_str, &hints, &res);
		if (err) {
			PAM_LDAP_LOG(LOG_ERR, "initialize : getaddrinfo returned error = %d\n", err);
			return (PAM_AUTHINFO_UNAVAIL);
		}
	} while ((bind(conf->sockfd, res->ai_addr, res->ai_addrlen) < 0) && 
			(local_port < 64000));

	if (res) {
		freeaddrinfo(res);	
		res = NULL;
	}

	if (local_port >= 64000) {
		close(conf->sockfd);
		PAM_LDAP_LOG(LOG_ERR, "No open port we could bind to.");
		return PAM_AUTHINFO_UNAVAIL;
	}

	return PAM_SUCCESS;
}

#undef PAM_FAIL_CHECK
#define PAM_FAIL_CHECK if (retval != PAM_SUCCESS) {	\
	int *pret = calloc( sizeof(int) , sizeof(char));		\
	if (NULL == pret) {				\
		return retval;				\
	}								\
	*pret = retval;					\
	pam_set_data( pamh, "adldap_setcred_return"	\
	              , (void *) pret, _int_free );	\
	return retval; }

/*
 * pam_sm_authenticate()
 *
 * Description:
 * This function authenticates the user while
 * adhereing to PAM module rules
 *
 * Input(s):
 * pamh		- PAM handle
 * flags	- Flags for pam_authenticate interface
 * argc		- No of arguements passed to the module
 * argv		- Arguements passed to the module
 *
 * Output(s):
 * None
 *
 * Returns:
 * 0 on success
 * Approprite PAM error on failure
 */
PAM_EXTERN int 
pam_sm_authenticate(pam_handle_t *pamh,int flags,int argc,CONST char **argv)
{
	CONST char *user;
	char *authUser = NULL;
	char *password = NULL;
	char *basedn = NULL;
	char *role = NULL;
	char *lf_id_list = NULL;
	char *lf_id = NULL;
	char *chassis_role = NULL;
	char *ad_list = NULL;
	int ctrl, ret, tries, i;
	int retval = PAM_AUTH_ERR, homeVF = -1;
	char hostname[SMALL_BUFFER_SIZE];
	char port[SMALL_BUFFER_SIZE];
	adldap_conf_t config;
	int *vfRoleMap;
	int vf_enabled = 0;
	const char * vf_enabled_str = NULL, *end_ptr;
	 const int       *fips_mode = NULL;
	
	/* Find whether VF is enabled or not from PAM envlist */
	vf_enabled_str = pam_getenv(pamh, "VF_ENABLED");
	if (vf_enabled_str) {
		/* In case vf environmental variable is not available and by
		 * default, it is assumed as non VF environment
		 */
		vf_enabled = strtol(vf_enabled_str , &end_ptr, 10);
	}

	authUser = (char*) calloc(SMALL_BUFFER_SIZE, sizeof(char));
	if (NULL == authUser) {
		retval = PAM_ABORT;
		PAM_FAIL_CHECK;
	}

    basedn = (char*) calloc(SMALL_BUFFER_SIZE, sizeof(char));
    if (NULL == basedn) {
        retval = PAM_ABORT;
        PAM_FAIL_CHECK;
    }

	role = (char*) calloc(SMALL_BUFFER_SIZE, sizeof(char));
	if (NULL == role) {
		retval = PAM_ABORT;
		PAM_FAIL_CHECK;
	}
	lf_id_list = (char*) calloc(SMALL_BUFFER_SIZE, sizeof(char));
	if (NULL == lf_id_list) {
		retval = PAM_ABORT;
		PAM_FAIL_CHECK;
	}
	lf_id = (char*) calloc(SMALL_BUFFER_SIZE, sizeof(char));
	if (NULL == lf_id) {
		retval = PAM_ABORT;
		PAM_FAIL_CHECK;
	}
	chassis_role = (char*) calloc(SMALL_BUFFER_SIZE, sizeof(char));
	if (NULL == chassis_role) {
		retval = PAM_ABORT;
		PAM_FAIL_CHECK;
	}

    ad_list = (char*) calloc(BUFFER_SIZE, sizeof(char));
    if (NULL == ad_list) {
        retval = PAM_ABORT;
        PAM_FAIL_CHECK;
    }

#undef PAM_FAIL_CHECK
#define PAM_FAIL_CHECK if (retval != PAM_SUCCESS) {goto error; }

	/* initialize debug log level */
	dbg_init();

	ctrl = _pam_parse(argc, argv, &config);
	tries = ((ctrl & PAM_RETRY) >> 4) + 1;

	/* grab the user name */
	retval = pam_get_user(pamh, &user, NULL);
	PAM_FAIL_CHECK;

	pam_get_data(pamh, PAM_ADLDAP_FIPSMODE,
                                                   (const void **) &fips_mode);
	PAM_FAIL_CHECK;

	if (fips_mode != NULL)
        	set_fips_mode(*fips_mode);

	/* check that they've entered something, and not too long, either */
	if ((user == NULL) ||
		(strlen(user) > MAXPWNAM)) {
		PAM_LDAP_LOG(LOG_DEBUG, "User name was NULL, or too long");
		retval = PAM_USER_UNKNOWN;
		PAM_FAIL_CHECK;
	}

	strncpy(authUser, user, SMALL_BUFFER_SIZE);
	PAM_LDAP_LOG(LOG_DEBUG, "Got user name %s", user);

	/*
	* Get the IP address of the authentication server
	* Then, open a socket, and bind it to a port
	*/
	retval = initialize(&config);
	PAM_FAIL_CHECK;

	/* 
	* If there's no client id specified, use the service type, to help
	* keep track of which service is doing the authentication.
	*/
	if (!config.client_id) {
		retval = pam_get_item(pamh, PAM_SERVICE, (CONST void **) &config.client_id);
		PAM_FAIL_CHECK;
	}

	/* grab the password (if any) from the previous authentication layer */
	retval = pam_get_item(pamh, PAM_AUTHTOK, (CONST void **) &password);
	PAM_FAIL_CHECK;

	if(password) {
		password = strdup(password);
		PAM_LDAP_LOG(LOG_DEBUG, "Got password %s", password);
	}

	PAM_LDAP_LOG(LOG_DEBUG, "User=%s Password=%s \n", user, password);

	/* no previous password: maybe get one from the user */
	if (!password) {
		if (ctrl & PAM_USE_FIRST_PASS) {
			/* use one pass only, stopping if it fails */
			retval = PAM_AUTH_ERR;	
			PAM_FAIL_CHECK;
		}
	} /* end of password == NULL */

#undef PAM_SERVER_NOT_REACHED_CHECK
#define PAM_SERVER_NOT_REACHED_CHECK if (retval == PAM_AUTHINFO_UNAVAIL) {goto last; }

	/* 
	 * Validate user using other servers in the list 
	 * if the current server is not accessible
	 */ 
	/*This is done the the order in which servers are listed*/
	while(config.server) {
		get_host_port(config.server->hostname, hostname, port);

		PAM_LDAP_LOG(LOG_DEBUG, "Host=%s Port=%s ", hostname, port);
		PAM_LDAP_LOG(LOG_DEBUG, "Timeout =%d", config.server->timeout);
		PAM_LDAP_LOG(LOG_DEBUG, "Domain =%s", config.server->domain);
		PAM_LDAP_LOG(LOG_DEBUG, "bind=%d", config.server->bindingMethod);
		PAM_LDAP_LOG(LOG_DEBUG, "sClas=%d", config.server->serverClass);
		PAM_LDAP_LOG(LOG_DEBUG, "atr=%s", config.server->uidSearchAttr);
		PAM_LDAP_LOG(LOG_DEBUG, "clientDN=%s", config.server->clientDN);
		PAM_LDAP_LOG(LOG_DEBUG, "paswd=%s", config.server->password);

		get_basedn(config.server->domain, basedn);
		PAM_LDAP_LOG(LOG_DEBUG, "Basedn=%s\n", basedn);

		retval = ldw_validate_user(authUser, password, hostname, port,
									basedn, config.server,
									vf_enabled);
		/*
		 * Note: Input "authUser" to ldw_validate_user may get modified inside function if
		 * username provided doesn't contain domain name and UPN based authentication
		 * is being used. This is due to RFE 2991 changes.
		 */
		PAM_LDAP_LOG(LOG_DEBUG, "Return value ldw_validate_user = %d ", retval );

		if(retval == 0) {
			memset(role, 0, SMALL_BUFFER_SIZE);
			retval = ldw_get_user_role(authUser, password, hostname, port,
								basedn, config.server,
								role, SMALL_BUFFER_SIZE, &gid,
								lf_id,chassis_role);
			if (retval != 0) {
				PAM_LDAP_LOG(LOG_DEBUG,
							"Invalid role in the AD server: %d \n",
							retval);
				retval = PAM_AUTH_ERR;
				PAM_FAIL_CHECK;
			}

			if (!vf_enabled) {
				ldw_get_ad_list(authUser, password, hostname, port, basedn,
							config.server, ad_list);
			} else {
                		PAM_LDAP_LOG(LOG_DEBUG,"VF enabled \n");
				vfRoleMap = malloc(PAM_ADLDAP_MAXVF * sizeof(int));
				if (!vfRoleMap) {
					PAM_LDAP_LOG(LOG_DEBUG,
								"Couldn't allocate memory for"
								" vfRoleMap %d \n",
								retval);
					retval = PAM_AUTH_ERR;
					PAM_FAIL_CHECK;
				}

				for (i = 0; i < PAM_ADLDAP_MAXVF; i++) {
					vfRoleMap[i] = INVALID_VF;
				}
				ret = ldw_get_vf_list(authUser, password, hostname, port,
								basedn, config.server,
								vfRoleMap, &homeVF);
			
				PAM_LDAP_LOG(LOG_DEBUG,"\nret %d home vf %d \n",ret, homeVF);	
				
				/*
		 		 * As per the functional spec, if user made a mistake
				 * in specifying VF Role Map it takes the default one.
		 		 */
				if (ret != PAM_SUCCESS) {
					for (i = 0; i < PAM_ADLDAP_MAXVF; i++) {
						vfRoleMap[i] = INVALID_VF;
					}
					homeVF = -1;
				}
			}
			live_server = config.server;

			PAM_LDAP_LOG(LOG_DEBUG, "Switch Role=%s", role);
			PAM_LDAP_LOG(LOG_DEBUG, "Ad List=%s", ad_list);

			goto accept;
			break;
		} else if (retval != -2) {
			PAM_LDAP_LOG(LOG_DEBUG, "Authentication failed \n");
			PAM_LDAP_LOG(LOG_DEBUG, "Return code %d \n", retval);

			if (hostname != NULL) {
				/*
				 * Set Server Name so that pam_fabos_log_auth
				 * can log this error
				 */
				char *serv_name = strdup(hostname);
				if (serv_name != NULL) {
					PAM_LDAP_LOG(LOG_DEBUG, "serv_name = %s\n", serv_name);
					retval = pam_set_data(pamh, PAM_ADLDAP_LIVESERV,
										(void *)serv_name, _char_free);
					if (retval == PAM_SUCCESS) {
						PAM_LDAP_LOG(LOG_DEBUG, "Live Server set to %s\n",
								serv_name);
					} else {
						PAM_LDAP_LOG(LOG_DEBUG,
						"Live Server not set, as pam_set_data failed\n");
					}
				}
			}

			retval = PAM_AUTH_ERR;
			PAM_FAIL_CHECK;
		}

		config.server = config.server->next;
	}
	PAM_LDAP_LOG(LOG_DEBUG, "Server not reachable\n");
	retval = PAM_AUTHINFO_UNAVAIL;
	PAM_SERVER_NOT_REACHED_CHECK;

accept:
	/* Whew! Done the pasword checks, look for an authentication acknowledge. Use role as new user name. */
	retval = role_to_user(pamh, authUser, role, ad_list, vfRoleMap, homeVF, lf_id_list, lf_id, chassis_role);
	if (retval != PAM_SUCCESS) {
		retval = PAM_AUTH_ERR;

error:
		PAM_LDAP_LOG(LOG_DEBUG, "pam_sm_authenticate: %d: returned error %d: \n", __LINE__, retval);
		/* If there was a password pass it to the next layer */
		if (password && *password) {
			pam_set_item(pamh, PAM_AUTHTOK, password);
		}
	}

last: 
	if (ctrl & PAM_DEBUG_ARG) {
		PAM_LDAP_LOG(LOG_DEBUG, "authentication %s", 
			retval==PAM_SUCCESS ? "succeeded":"failed" );
	}
 
	close(config.sockfd);
	cleanup(config.server);
#ifdef DEBUG_ZEROIZATION
	FILE *fp = fopen("/dev/console", "w+");
	        fprintf(fp, "ZeroizationLog: In pam_sm_authenticate(): Before calling _pam_forget(password), password = %s\n", password);
#endif
	int passlen =0;
	if (password != NULL) {
		passlen = strlen(password);
		memset(password, 0, passlen);
	}
   #ifdef MEMCMP_ZEROIZATION
	pam_memcmp_zero(password, passlen);
   #endif
#ifdef DEBUG_ZEROIZATION
	fprintf(fp, "ZeroizationLog: In pam_sm_authenticate(): After calling _pam_forget(password), password = %s\n", password);
	fclose(fp);
#endif
	free(password);
	password = NULL;

	{
		int *pret = calloc( sizeof(int) , sizeof(char));
		if (NULL == pret) {
			return retval;
		}

		*pret = retval;

		pam_set_data( pamh, "adldap_setcred_return", (void *) pret, _int_free );
	}

	/*Free the memory allocated to char pointers*/	
	if (authUser){
		free(authUser);
		authUser = NULL;	
	}
	if (basedn){
		free(basedn);
		basedn = NULL;
	}
	if (role){
		free(role);
		role = NULL;
	}
	if (ad_list){
		free(ad_list);
		ad_list = NULL;
	}
	return (retval);
}

/*
 * pam_sm_setcred()
 *
 * Description:
 * This function sets credentials for PAM module
 *
 * Input(s):
 * pamh		- PAM handle
 * flags	- Flags for pam_authenticate interface
 * argc		- No of arguements passed to the module
 * argv		- Arguements passed to the module
 *
 * Output(s):
 * None
 *
 * Returns:
 * 0 on success
 * Approprite PAM error on failure
 */
PAM_EXTERN int
pam_sm_setcred(pam_handle_t *pamh,int flags,int argc,CONST char **argv)
{
	int retval, *pret;

	retval = PAM_SUCCESS;
	pret = &retval;
	pam_get_data( pamh, "adldap_setcred_return", (CONST void **) &pret );
	return *pret;
}

#undef PAM_FAIL_CHECK
#define PAM_FAIL_CHECK if (retval != PAM_SUCCESS) { return PAM_SESSION_ERR; }

/*
 * pam_open_session()
 *
 * Description:
 * This function sets environmental variable for
 * open session
 *
 * Input(s):
 * pamh		- PAM handle
 * flags	- Flags for pam_authenticate interface
 * argc		- No of arguements passed to the module
 * argv		- Arguements passed to the module
 *
 * Output(s):
 * None
 *
 * Returns:
 * 0 on success
 * Approprite PAM error on failure
 */
PAM_EXTERN int
pam_sm_open_session(pam_handle_t *pamh, int flags,
		    int argc, CONST char **argv)
{
	int retval = PAM_AUTH_ERR;
  	CONST char *user = NULL;

	/*
	 * If public key authentication is enabled and authentication mode is 
	 * ldap only, pam_sm_authentication from ldap will return failure.
	 * However, ssh authentication process will continue if it finds a 
	 * matching public/private key pair. It then checks pam_open_session
	 * and if we return SUCCESS from here, session gets established
	 * wrongly. We, therefore, check for the username here which will be
	 * set as part of pam_sm_authenticate only if authentication was 
	 * successful. If we don't find PAM_ADLDAP_UNAME means authentication
	 * was not successful from LDAP side so we return failure from session
	 * also.
	 */
	retval = pam_get_data(pamh, PAM_ADLDAP_UNAME, (CONST void **) &user);
	PAM_FAIL_CHECK;
	return PAM_SUCCESS;
	/*
	return pam_private_session(pamh, flags, argc, argv, PW_STATUS_START);
	*/
}

/*
 * pam_close_session()
 *
 * Description:
 * This function sets environmental variable for
 * close session
 *
 * Input(s):
 * pamh		- PAM handle
 * flags	- Flags for pam_authenticate interface
 * argc		- No of arguements passed to the module
 * argv		- Arguements passed to the module
 *
 * Output(s):
 * None
 *
 * Returns:
 * 0 on success
 * Approprite PAM error on failure
 */
PAM_EXTERN int
pam_sm_close_session(pam_handle_t *pamh, int flags,
		     int argc, CONST char **argv)
{
	return PAM_SUCCESS;
	/*
	return pam_private_session(pamh, flags, argc, argv, PW_STATUS_STOP);
	*/
}

#ifdef PAM_STATIC
/* static module data */
struct pam_module _pam_adldap_modstruct = {
	"pam_adldap_auth",
	pam_sm_authenticate,
	pam_sm_setcred,
	NULL,
	pam_sm_open_session,
	pam_sm_close_session,
	NULL,
};
#endif
