/*
 * $Id: pam_radius_auth.c,v 1.27 2002/05/06/07 14:18:34 aland Exp $
 * pam_radius_auth
 *      Authenticate a user via a RADIUS session
 *
 * 0.9.0 - Didn't compile quite right.
 * 0.9.1 - Hands off passwords properly.  Solaris still isn't completely happy
 * 0.9.2 - Solaris now does challenge-response.  Added configuration file
 *         handling, and skip_passwd field
 * 1.0.0 - Added handling of port name/number, and continue on select
 * 1.1.0 - more options, password change requests work now, too.
 * 1.1.1 - Added client_id=foo (NAS-Identifier), defaulting to PAM_SERVICE
 * 1.1.2 - multi-server capability.
 * 1.2.0 - ugly merger of pam_radius.c to get full RADIUS capability
 * 1.3.0 - added my own accounting code.  Simple, clean, and neat.
 * 1.3.1 - Supports accounting port (oops!), and do accounting authentication
 * 1.3.2 - added support again for 'skip_passwd' control flag.
 * 1.3.10 - ALWAYS add Password attribute, to make packets RFC compliant.
 * 1.3.11 - Bug fixes by Jon Nelson <jnelson@securepipe.com>
 * 1.3.12 - miscellanous bug fixes.  Don't add password to accounting
 *          requests; log more errors; add NAS-Port and NAS-Port-Type
 *          attributes to ALL packets.  Some patches based on input from
 *          Grzegorz Paszka <Grzegorz.Paszka@pik-net.pl>
 * 1.3.13 - Always update the configuration file, even if we're given
 *          no options.  Patch from Jon Nelson <jnelson@securepipe.com>
 * 1.3.14 - Don't use PATH_MAX, so it builds on GNU Hurd.
 * 1.3.15 - Implement retry option, miscellanous bug fixes.
 *
 * See end for Copyright information */

#define PAM_SM_AUTH
#define PAM_SM_PASSWORD
#define PAM_SM_SESSION

#define PORT_STR_LEN	6
#define MAX_LENGTH		256

#include <limits.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <netdb.h>
#include <grp.h>
#include <time.h>

#ifdef sun
#include <security/pam_appl.h>
#endif
#include <security/pam_modules.h>
#include <unistd.h>
#include <sys/wait.h>
#include "pam_radius_auth.h"
#include "../pam_unix/brocade_hashing.h"

# define        h_addr  h_addr_list[0] /* Address, for backward compatibility.*/


/* internal data */
static CONST char *pam_module_name = "pam_radius_auth";
static char conf_file[BUFFER_SIZE]; /* configuration file */

/* we need to save these from open_session to close_session, since
 * when close_session will be called we won't be root anymore and
 * won't be able to access again the radius server configuration file
 * -- cristiang */
static radius_server_t *live_server = NULL;
static time_t session_time;
static int log_level = LOG_EMERG;
static pam_handle_t *pamhandle = NULL;


static int
do_peap_authentication(radius_conf_t * config, CONST char * user, char * password,
        AUTH_HDR *response);
static int rad_converse(pam_handle_t *pamh, int msg_style, char *message, char **password);
void pam_memcmp_zero(void *buf, int len);

/* logging */
static void _pam_log(int err, CONST char *format, ...)
{
    va_list args;
    char buffer[BUFFER_SIZE];

    va_start(args, format);
    vsprintf(buffer, format, args);
    /* don't do openlog or closelog, but put our name in to be friendly */
    syslog(err, "%s: %s", pam_module_name, buffer);
    if (err <= log_level) {
	    fprintf(stderr, "%s\n", buffer);
    }
    va_end(args);
}

static void _pam_log2(const char *func, int line, int err,
					  CONST char *format, ...)
{
	va_list args;
	char buffer[BUFFER_SIZE];

	va_start(args, format);
	vsprintf(buffer, format, args);
	if (err <= log_level) {
		fprintf(stderr, "%s(%d): %s\n", func, line, buffer);
	}
	va_end(args);
}

#define	PAM_LOG(err, format...)											\
do {																	\
	_pam_log2(__FUNCTION__, __LINE__, err, format);						\
} while (0)


/* argument parsing */
static int _pam_parse(int argc, CONST char **argv, radius_conf_t *conf)
{
  int ctrl=0;

  memset(conf, 0, sizeof(radius_conf_t)); /* ensure it's initialized */

  strcpy(conf_file, CONF_FILE);

  /*
   *  If either is not there, then we can't parse anything.
   */
  if ((argc == 0) || (argv == NULL)) {
    return ctrl;
  }

  /* step through arguments */
  for (ctrl=0; argc-- > 0; ++argv) {

    /* generic options */
    if (!strncmp(*argv,"conf=",5)) {
      strcpy(conf_file,*argv+5);

    } else if (!strcmp(*argv, "use_first_pass")) {
      ctrl |= PAM_USE_FIRST_PASS;

    } else if (!strcmp(*argv, "try_first_pass")) {
      ctrl |= PAM_TRY_FIRST_PASS;

    } else if (!strcmp(*argv, "skip_passwd")) {
      ctrl |= PAM_SKIP_PASSWD;

    } else if (!strncmp(*argv,"retry=",6)) {
      int i = atoi(*argv+6);
      i &= 0x03;		/* keep the low 3 bits only */
      ctrl |= (i << 4);

    } else if (!strncmp(*argv, "client_id=", 10)) {
      if (conf->client_id) {
	_pam_log(LOG_WARNING, "ignoring duplicate '%s'", *argv);
      } else {
	conf->client_id = (char *) *argv+10; /* point to the client-id */
      }
    } else if (!strcmp(*argv, "accounting_bug")) {
      conf->accounting_bug = TRUE;

    } else if (!strcmp(*argv, "debug")) {
      ctrl |= PAM_DEBUG_ARG;
      conf->debug = 1;

    } else {
      _pam_log(LOG_WARNING, "unrecognized option '%s'", *argv);
    }
  }

  return ctrl;
}

/* Callback function used to free the saved return value for pam_setcred. */
void _int_free( pam_handle_t * pamh, void *x, int error_status )
{
    free(x);
}

/* Callback funtion to free the saved RADIUS user name */
void _char_free(pam_handle_t *pamh, void *x, int error_status)
{
    free(x);
}

/*************************************************************************
 * SMALL HELPER FUNCTIONS
 *************************************************************************/

/*
 * Return an IP address in host long notation from
 * one supplied in standard dot notation.
 */
static UINT4 ipstr2long(char *ip_str) {
  char	buf[6];
  char	*ptr;
  int	i;
  int	count;
  UINT4	ipaddr;
  int	cur_byte;

  ipaddr = (UINT4)0;

  for(i = 0;i < 4;i++) {
    ptr = buf;
    count = 0;
    *ptr = '\0';

    while(*ip_str != '.' && *ip_str != '\0' && count < 4) {
      if(!isdigit(*ip_str)) {
	return((UINT4)0);
      }
      *ptr++ = *ip_str++;
      count++;
    }

    if(count >= 4 || count == 0) {
      return((UINT4)0);
    }

    *ptr = '\0';
    cur_byte = atoi(buf);
    if(cur_byte < 0 || cur_byte > 255) {
      return ((UINT4)0);
    }

    ip_str++;
    ipaddr = ipaddr << 8 | (UINT4)cur_byte;
  }
  return(ipaddr);
}

/*
 * Check for valid IP address in standard dot notation.
 */
static int good_ipaddr(char *addr) {
  int	dot_count;
  int	digit_count;

  dot_count = 0;
  digit_count = 0;
  while(*addr != '\0' && *addr != ' ') {
    if(*addr == '.') {
      dot_count++;
      digit_count = 0;
    } else if(!isdigit(*addr)) {
      dot_count = 5;
    } else {
      digit_count++;
      if(digit_count > 3) {
	dot_count = 5;
      }
    }
    addr++;
  }
  if(dot_count != 3) {
    return(-1);
  } else {
    return(0);
  }
}

/*
 * Return an IP address in host long notation from a host
 * name or address in dot notation.
 */
static UINT4 get_ipaddr(char *host) {
  struct hostent *hp;

  if(good_ipaddr(host) == 0) {
    return(ipstr2long(host));

  } else if((hp = gethostbyname(host)) == (struct hostent *)NULL) {
    return((UINT4)0);
  }

  return(ntohl(*(UINT4 *)hp->h_addr));
}

/*
 * take server->hostname, without affecting the input pointer extract
 * hostname without port and return.
 */
static char *
gethostname_from_server(const radius_server_t *server)
{
	char *p = NULL, *temp = NULL;
	char *retPtr = NULL;

	if (NULL != server && NULL != server->hostname) {
		if ((p = strrchr(server->hostname, ':')) != NULL) {
			temp = p;
			*(p++) = '\0';		/* split the port off from the host name */
		}
		retPtr = strdup(server->hostname);
		if(temp)
			*(temp++) = ':';    /* restore server back */
	}
	
	return retPtr;
}

/*
 * take server->hostname, and convert it to server->ip and server->port
 */
static int
host2server(radius_server_t *server)
{
  char *p;

  /*
   * Port number is placed at the end of host name followed by last ":"
   * If the port number is not appended to the host name there is
   * a possibility that port number is taken from the last part of IPV6
   * address itself which is an error.
   */
  if ((p = strrchr(server->hostname, ':')) != NULL) {
    *(p++) = '\0';		/* split the port off from the host name */
  }

  /*
   *  If the server port hasn't already been defined, go get it.
   */
  if (!server->port) {
    if (p && isdigit(*p)) {	/* the port looks like it's a number */
      unsigned int i = atoi(p) & 0xffff;

      if (!server->accounting) {
	server->port = htons((u_short) i);
      } else {
	server->port = htons((u_short) (i + 1));
      }
    } else {			/* the port looks like it's a name */
      struct servent *svp;

      if (p) {			/* maybe it's not "radius" */
	svp = getservbyname (p, "udp");
	*(--p) = ':';		/* be sure to put the delimiter back */
      } else {
	if (!server->accounting) {
	  svp = getservbyname ("radius", "udp");
	} else {
	  svp = getservbyname ("radacct", "udp");
	}
      }

      if (svp == (struct servent *) 0) {
	return PAM_AUTHINFO_UNAVAIL;
      }

      server->port = svp->s_port;
    }
  }

  return PAM_SUCCESS;
}

/*
 * Do XOR of two buffers.
 */
static unsigned char *
xor(unsigned char *p, unsigned char *q, int length)
{
  int i;
  unsigned char *retval = p;

  for (i = 0; i < length; i++) {
    *(p++) ^= *(q++);
  }
  return retval;
}

/**************************************************************************
 * MID-LEVEL RADIUS CODE
 **************************************************************************/

/*
 * get a pseudo-random vector.
 */
static void
get_random_vector(unsigned char *vector)
{
#ifdef linux
  int fd = open("/dev/urandom",O_RDONLY); /* Linux: get *real* random numbers */
  int total = 0;
  if (fd >= 0) {
    while (total < AUTH_VECTOR_LEN) {
      int bytes = read(fd, vector + total, AUTH_VECTOR_LEN - total);
      if (bytes <= 0)
	break;			/* oops! Error */
      total += bytes;
    }
    close(fd);
  }

  if (total != AUTH_VECTOR_LEN)
#endif
    {				/* do this *always* on other platforms */
      MD5_CTX my_md5;
      struct timeval tv;
      struct timezone tz;
      static unsigned int session = 0; /* make the number harder to guess */

      /* Use the time of day with the best resolution the system can
	 give us -- often close to microsecond accuracy. */
      gettimeofday(&tv,&tz);

      if (session == 0) {
	session = getppid();	/* (possibly) hard to guess information */
      }

      tv.tv_sec ^= getpid() * session++;

      /* Hash things to get maybe cryptographically strong pseudo-random numbers */
      MD5Init(&my_md5);
      MD5Update(&my_md5, (unsigned char *) &tv, sizeof(tv));
      MD5Update(&my_md5, (unsigned char *) &tz, sizeof(tz));
      MD5Final(vector, &my_md5);	      /* set the final vector */
    }
}

/*
 * RFC 2139 says to do generate the accounting request vector this way.
 * However, the Livingston 1.16 server doesn't check it.  The Cistron
 * server (http://home.cistron.nl/~miquels/radius/) does, and this code
 * seems to work with it.  It also works with Funk's Steel-Belted RADIUS.
 */
static void
get_accounting_vector(AUTH_HDR *request, radius_server_t *server)
{
  MD5_CTX my_md5;
  int secretlen = strlen(server->secret);
  int len = ntohs(request->length);

  memset(request->vector, 0, AUTH_VECTOR_LEN);
  MD5Init(&my_md5);
  memcpy(((char *)request) + len, server->secret, secretlen);

  MD5Update(&my_md5, (char *)request, len + secretlen);
  MD5Final(request->vector, &my_md5);      /* set the final vector */
}

/*
 * Verify the response from the server
 */
static int
verify_packet(char *secret, AUTH_HDR *response, AUTH_HDR *request)
{
  MD5_CTX my_md5;
  unsigned char	calculated[AUTH_VECTOR_LEN];
  unsigned char	reply[AUTH_VECTOR_LEN];

  /*
   * We could dispense with the memcpy, and do MD5's of the packet
   * + vector piece by piece.  This is easier understand, and maybe faster.
   */
  memcpy(reply, response->vector, AUTH_VECTOR_LEN); /* save the reply */
  memcpy(response->vector, request->vector, AUTH_VECTOR_LEN); /* sent vector */

  /* MD5(response packet header + vector + response packet data + secret) */
  MD5Init(&my_md5);
  MD5Update(&my_md5, (unsigned char *) response, ntohs(response->length));

  /*
   * This next bit is necessary because of a bug in the original Livingston
   * RADIUS server.  The authentication vector is *supposed* to be MD5'd
   * with the old password (as the secret) for password changes.
   * However, the old password isn't used.  The "authentication" vector
   * for the server reply packet is simply the MD5 of the reply packet.
   * Odd, the code is 99% there, but the old password is never copied
   * to the secret!
   */
  if (*secret) {
    MD5Update(&my_md5, secret, strlen(secret));
  }

  MD5Final(calculated, &my_md5);      /* set the final vector */

  /* Did he use the same random vector + shared secret? */
  if (memcmp(calculated, reply, AUTH_VECTOR_LEN) != 0) {
    return FALSE;
  }
  return TRUE;
}

/*
 * Find an attribute in a RADIUS packet.  Note that the packet length
 * is *always* kept in network byte order.
 */
static attribute_t *
find_attribute(AUTH_HDR *response, unsigned char type)
{
  attribute_t *attr = (attribute_t *) &response->data;

  int len = ntohs(response->length) - AUTH_HDR_LEN;

  while (attr->attribute != type) {
    if ((len -= attr->length) <= 0) {
      return NULL;		/* not found */
    }
    attr = (attribute_t *) ((char *) attr + attr->length);
  }

  return attr;
}

/*
 * Find a given VSA in a RADIUS packet. This ensures if there are multiple
 * VSA in a response packet, the first VSA with the specified vendor
 * type will be found.
 */
static attribute_t *
find_vsa(AUTH_HDR *response, unsigned int vendor_id, unsigned char vendor_type)
{
	attribute_t * attr = (attribute_t *) &response->data;

	int len = ntohs(response->length) - AUTH_HDR_LEN;

	PAM_LOG(LOG_DEBUG, "looking for vendor id: %d, type: %d in len: %d\n",
			vendor_id, vendor_type, len);

	while (len > 0) {
		PAM_LOG(LOG_DEBUG, "attribute: %d, len: %d\n", attr->attribute,
				attr->length);
		if (attr->attribute == PW_VENDOR_SPECIFIC) {
			unsigned int	id;

			id = *(unsigned int *)attr->data;
			id = ntohl(id);
			PAM_LOG(LOG_DEBUG, "found VSA. id: %d\n", id);
			if (id == vendor_id) {
				unsigned char	type = *(attr->data + 4);

				PAM_LOG(LOG_DEBUG, "vendor id match. type: %d\n", type);
				if (type == vendor_type) {
					return attr;
				}
			}
		}
		len -= attr->length;
		attr = (attribute_t *) ((char *) attr + attr->length);
	}

	return NULL;
}

/*
 * Get group id of the role.
 */
int
getgroupid(const char *role_name)
{
	struct group    *grp = NULL;

	if (!role_name) {
		return (PAM_RADIUS_INVALID_ROLE);
	}

	if (strcmp(role_name, "root") == 0) {
		return (PAM_RADIUS_INVALID_ROLE);
	}

	grp = getgrnam(role_name);
	if (!grp) {
		return (PAM_RADIUS_INVALID_ROLE);
	}

	return (grp->gr_gid);
}

/*
 * This function converts from string to integer. In case of error
 * or NULL input, it returns -1.
 */
static int
convertToInt(const char *str)
{
	int ret = -1;
	char *end_ptr;

	if (str == NULL) {
		return (ret);
	}

	errno = 0;
	ret = strtol(str, &end_ptr, 10);

	if ((*(end_ptr) != '\0') || ((errno) != 0))  {
		return (-1);
	}

	return (ret);
}

/*
 * int inform_password_expiry()
 *
 * Description:
 * This function finds the days left for the password expiration and
 * inform the user, if days remaining are with in the warning period.
 *
 * Input(s):
 * 	text - This text needs to be communicated.
 *	pamh - Pam Handle
 *
 * Output(s):
 *
 * Returns:
 *   SUCCESS if successful; error otherwise.
 */
static int inform_passwd_expiration(pam_handle_t *pamh,	char *text)
{
	int retval = -1;
	struct pam_conv *conv;
	struct pam_message *pmsg[1], msg[1];
	struct pam_response *resp;

	pmsg[0] = &msg[0];
	msg[0].msg = text;
	msg[0].msg_style = PAM_TEXT_INFO;

	PAM_LOG(LOG_DEBUG, "Enter");
	resp = NULL;
	retval = pam_get_item(pamh, PAM_CONV, (const void **) &conv);
	if (retval == PAM_SUCCESS) {
		retval = conv->conv(1, (const struct pam_message **)pmsg,
							&resp, conv->appdata_ptr);

		if (retval != PAM_SUCCESS) {
			PAM_LOG(LOG_DEBUG, "Couldn't inform user");
			return (PAM_RADIUS_COMM_FAILED);
		}
	} else if (retval != PAM_CONV_AGAIN) {
		PAM_LOG(LOG_DEBUG, "Couldn't inform user");
		return (PAM_RADIUS_COMM_FAILED);
	}

	return (retval);
}

/*
 * int warn_password_expiry()
 *
 * Description:
 * This function finds the days left for the password expiration and
 * inform the user, if days remaining are with in the warning period.
 *
 * Input(s):
 *   response - Radius response that contains VSAs
 *
 * Output(s):
 *
 * Returns:
 *   SUCCESS if successful; error otherwise.
 */
static int
warn_passwd_expiry(pam_handle_t *pamh, const AUTH_HDR *response)
{
    attribute_t	*brcd_passwd_date = NULL, *brcd_warn_days = NULL;
	int		type1 = BRCD_ATTR_PASSWD_EXPIREDATE;
	int		type2 = BRCD_ATTR_WARNPERIOD;
    int	ret = 0, daysleft, warn_days, month, day, year, len, timeleft;
	struct tm 		*ptm = NULL;
	struct tm		expiretm;
	time_t 	expiretime, now, currtime;
	char	text[80] = {0}, tmpbuf[256] = {0}, *sptr;
	char *smonth = NULL, *sday = NULL, *syear = NULL, *swarn = NULL;
	const char *pd;
	int maxDays;

	if (!pamh || !response) {
		return (PAM_AUTH_ERR);
	}

	PAM_LOG(LOG_DEBUG, "Enter");
	brcd_passwd_date = find_vsa(response, BRCD_VENDOR_CODE, type1);
	brcd_warn_days = find_vsa(response, BRCD_VENDOR_CODE, type2);

	if (brcd_passwd_date) {
		/* Customer should enter both expiry date and warn period.*/
		if (!brcd_warn_days) {
			PAM_LOG(LOG_DEBUG, "User should enter warn period");
			return (PAM_RADIUS_IMPROPER_SYNTAX);
		}
		pd = brcd_warn_days->data + sizeof (int) +
			sizeof (unsigned char);

		/*
		 * Vendor length includes the Vendor type, length and data
		 * fields.
		 */
		len = *pd++;
		memcpy(tmpbuf, pd, len-2);
		if (tmpbuf) {
			warn_days = convertToInt(tmpbuf);
		}
		if (warn_days == -1) {
			PAM_LOG(LOG_DEBUG, "Wrong syntax for warn days");
			return (PAM_RADIUS_IMPROPER_SYNTAX);
		}

		pd = brcd_passwd_date->data + sizeof (int) +
								sizeof (unsigned char);
		/*
		 * Vendor length includes the Vendor type, length and data
		 * fields. It contains the sample date string 12/08/2008 or
		 * 12/8/2008. Here 12 is the month,8 is the date and 2008
		 * is the year.
		 */
		len = *pd++;
		memcpy(tmpbuf, pd, len-2);

		smonth = strtok_r(tmpbuf, "/", &sptr);
		if (!smonth) {
			PAM_LOG(LOG_DEBUG, "Wrong syntax for month");
			return (PAM_RADIUS_IMPROPER_SYNTAX);
		}

		/* Takes care about the day preceding with '0'. */
		if (*smonth == '0') {
			month = convertToInt((smonth + 1));
		} else {
			month = convertToInt(smonth);
		}
		if ((month < JAN) || (month > DEC)) {
			PAM_LOG(LOG_DEBUG, "Wrong month");
			return (PAM_RADIUS_IMPROPER_SYNTAX);
		}

		sday = strtok_r(NULL, "/", &sptr);
		if (!sday) {
			PAM_LOG(LOG_DEBUG, "Wrong syntax for day");
			return (PAM_RADIUS_IMPROPER_SYNTAX);
		}

		/* Takes care about the month preceding with '0'. */
		if (*sday == '0') {
			day = convertToInt((sday + 1));
		} else {
			day = convertToInt(sday);
		}
		if ((month == JAN) || (month == MAR) || (month == MAY) ||
			(month == JUL) ||	(month == AUG) || (month == OCT) ||
			(month == DEC)) {
			maxDays = THIRTYONE;
		} else if ((month == APR) || (month == JUN) || (month == SEP) ||
			(month == NOV)) {
			maxDays = THIRTY;
		} else {
			/* Only Feb is left. */
			maxDays = TWENTYEIGHT;
		}
		if ((day < 1) || (day > maxDays)) {
			PAM_LOG(LOG_DEBUG, "Wrong day");
			return (PAM_RADIUS_IMPROPER_SYNTAX);
		}

		syear = strtok_r(NULL, "/", &sptr);
		if (!syear) {
			PAM_LOG(LOG_DEBUG, "Wrong syntax for year");
			return (PAM_RADIUS_IMPROPER_SYNTAX);
		}
		year = convertToInt(syear);
		if (year < 0) {
			PAM_LOG(LOG_DEBUG, "Wrong year");
			return (PAM_RADIUS_IMPROPER_SYNTAX);
		}

		PAM_LOG(LOG_DEBUG, "year %d, day:%d, month:%d",
				year, day, month);
		/* Get the Current UTC time. */
		time(&now);
		ptm =  gmtime(&now);
		currtime = mktime(ptm);

		/* Assign the user's password expiry date. Here it assumes
		 * password expires at the end of the day 23:59:59(HH:MM:SS).
		 */
		expiretm.tm_sec = 59;
		expiretm.tm_min = 59;
		expiretm.tm_hour = 23;
		expiretm.tm_mon = month - 1;
		expiretm.tm_mday = day;
		expiretm.tm_year = year - 1900;
		expiretime = mktime(&expiretm);

		/* Calculates the days left for the password. */
		daysleft = (difftime(expiretime, currtime))/(24 * 60 * 60);
		timeleft = (difftime(expiretime, currtime));

		PAM_LOG(LOG_DEBUG, "days left:%d, warn period:%d", daysleft,
				warn_days);
		/* If the days left are less than or equal to warn days,
		 * give the warning to the user.
		 */
		if (daysleft <= warn_days) {
			if (daysleft > 0) {
				sprintf(text, "Warning: your password will expire"
				" in %d days\n", daysleft);
				ret = inform_passwd_expiration(pamh, text);
			} else if (daysleft == 0) {
                                if (timeleft <= 0) {
                                        PAM_LOG(LOG_DEBUG, "Warning: your password already"
                                                        " expired", text);
                                        return (PAM_RADIUS_PASSWORD_EXPIRED);
                                } else {
                                        sprintf(text, "Warning: your password expires today\n");
                                        ret = inform_passwd_expiration(pamh, text);
                                }
			} else {
				/* Right now we are not informing the user about passwd
				 * expiration. */
				PAM_LOG(LOG_DEBUG, "Warning: your password already"
						" expired", text);
				return (PAM_RADIUS_PASSWORD_EXPIRED);
			}
			PAM_LOG(LOG_DEBUG, "text:%s", text);
		}
	}

	return (ret);
}

/*
 * int parse_home_vf()
 *
 * Description:
 * This function parses the home VF from the string.
 *
 * Input(s):
 *   data - This is the HomeLF string.Example:"HomeLF=1"
 *
 * Output(s):
 *   homeVF - Gets the homeVF from the string and assign it.
 *
 * Returns:
 *   SUCCESS if successful; error otherwise.
 */
static int
parse_home_vf(char *data, int *homeVF)
{
	char *sptr;
	char *homevf = NULL, *temp = NULL;
	int ret = -1;

	if (!data || !homeVF) {
		return (PAM_AUTH_ERR);
	}

	PAM_LOG(LOG_DEBUG, "Data:%s", data);
	temp = strtok_r(data, "=", &sptr);
	/* Check for whether it is homeLF or lf role map. This
	 * needs to be in this while loop because they can
	 * appeared multiple times.
	*/
	if (strncasecmp(temp, BRCD_ATTR_HOMEVF,
					strlen(BRCD_ATTR_HOMEVF)) == 0) {
		homevf = strtok_r(NULL, ";", &sptr);
		if (homevf == NULL) {
			return (PAM_RADIUS_IMPROPER_SYNTAX);
		}
		ret = convertToInt(homevf);
		if ((ret > PAM_RADIUS_MAXVF) || (ret < 1)) {
			return (PAM_RADIUS_IMPROPER_SYNTAX);
		}
		/* Only first homeVF values is taken. */
		if (*homeVF == INVALID_VF) {
			/* If once it gets the homeVF, it breaks. It doesn't
			 * consider if user specifies homeVF multiple times. */
			*homeVF = ret;
		}
			PAM_LOG(LOG_DEBUG, "Home LF:%d", *homeVF);
	}

	return (PAM_SUCCESS);
}

/*
 * int parse_chassis_role()
 *
 * Description:
 * This function parses the chassis role from the VSA.
 *
 * Input(s):
 *   data - This is the Chassis role. Example:"ChassisRole=admin"
 *
 * Output(s):
 *   chassis_gid -Gets the chassis role's gid from thevsa and assign it.
 *
 * Returns:
 *   SUCCESS if successful; error otherwise.
 */
static int
parse_chassis_role(char *data, int *chassis_gid)
{
	char *sptr;
	char *chassis_role = NULL, *temp = NULL;
	int ret = -1;

	if (!data || !chassis_gid) {
		return (PAM_AUTH_ERR);
	}

	PAM_LOG(LOG_DEBUG, "Data:%s", data);
	temp = strtok_r(data, "=", &sptr);
	/* Check for whether it is homeLF or LF role map. This
	 * needs to be in this while loop because they can
	 * appeared multiple times.
	*/
	if (strncasecmp(temp, BRCD_ATTR_CHASSISROLE,
					strlen(BRCD_ATTR_CHASSISROLE)) == 0) {
		chassis_role = strtok_r(NULL, ";", &sptr);
		if (chassis_role == NULL) {
			return (PAM_RADIUS_IMPROPER_SYNTAX);
		}
		ret = getgroupid(chassis_role);
		/* Only first chassis role values is taken. */
		if (*chassis_gid == INVALID_VF) {
			/* If once it gets the homeVF, it breaks. It doesn't
			 * consider if user specifies homeVF multiple times. */
			*chassis_gid = ret;
		}
		PAM_LOG(LOG_DEBUG, "Chassis Role's GID:%d", *chassis_gid);
	}

	return (PAM_SUCCESS);
}

/*
 * int parse_vf_list()
 *
 * Description:
 * This function parses the LF Role list and assign it to VF Role
 * Map. It gets input as role:LFList.
 * Example string user:11-13,15,17,19;
 *
 * Input(s):
 *   data - This is the LFRoleList string.
 *
 * Output(s):
 *   vfRoleMap - Gets the vf and role list from the string and assign
 * it to the VF Role Map.
 *
 * Returns:
 *   SUCCESS if successful; error otherwise.
 */
static int
parse_vf_list(char *data, int *vfRoleMap)
{
	char	*sptr, *range_str = NULL, *indexstr = NULL;
	char	*role = NULL;
	int     range_val1, range_val2, gid, i = 0, index;

	if (!data || !vfRoleMap) {
		return (PAM_AUTH_ERR);
	}

	PAM_LOG(LOG_DEBUG, "data:%s", data);
	/* Now it is VF Role Map, so process the role and vfs
	 * associated  with the role.
	 */
	role = strtok_r(data, ":", &sptr);
	if (role == NULL) {
		return (PAM_RADIUS_IMPROPER_SYNTAX);
	}

	/* convert mix case role string to lower case */
	while (role[i] != '\0') {
		role[i] = tolower(role[i]);
		i++;
	}
	PAM_LOG(LOG_DEBUG, "vf role:%s", role);

	indexstr = strtok_r(NULL, ",", &sptr);
	/* Convert to GID */
	gid = getgroupid(role);
	if (gid == PAM_RADIUS_INVALID_ROLE) {
		return (gid);
	}

	/* Here indexstr is either a single VF or range of VFs. */
	while (indexstr != NULL) {
		errno = 0;
		/* If the range of VFs specified, below part takes care
		 * about it. It validates the range of strings.
		 */
		if ((range_str = strchr(indexstr, '-')) != NULL) {
			*range_str = '\0';
			range_str++;
			if ((strchr(range_str, '-')) != NULL) {
				return (PAM_RADIUS_IMPROPER_SYNTAX);
			}
			if ((*range_str == '\0') || (*indexstr == '\0')) {
				return (PAM_RADIUS_IMPROPER_SYNTAX);
			}

			range_val1 = convertToInt(indexstr);
			if ((range_val1 > PAM_RADIUS_MAXVF) || (range_val1 < 0)) {
					return (PAM_RADIUS_IMPROPER_SYNTAX);
			}

			range_val2 = convertToInt(range_str);
			if ((range_val2 > PAM_RADIUS_MAXVF) || (range_val2 < 0)) {
					return (PAM_RADIUS_IMPROPER_SYNTAX);
			}

			PAM_LOG(LOG_DEBUG, "vf range_str1:%d\n", range_val1);
			PAM_LOG(LOG_DEBUG, "vf range_str2:%d\n", range_val2);

			for (i = range_val1; i <= range_val2; i++) {
				/* If this vf is already assigned earlier, only
				 * first value has to be taken into account. */
				if (vfRoleMap[i] == INVALID_VF) {
					vfRoleMap[i] = gid;
				}
			}
		} else {
			index = convertToInt(indexstr);
			if ((index > PAM_RADIUS_MAXVF) || (index < 0)) {
				return (PAM_RADIUS_IMPROPER_SYNTAX);
			}
			PAM_LOG(LOG_DEBUG, "vf index:%d\n", index);

			/* If this vf is already assigned earlier, only
			 * first value has to be taken into account. */
			if (vfRoleMap[index] == INVALID_VF) {
				vfRoleMap[index] = gid;
			}
		}
		/* Comma separates multiple vfs. so now check for next
			vf after comma .*/
		indexstr = strtok_r(NULL, ",", &sptr);
	}

	return (PAM_SUCCESS);
}

/*
 * int brcd_vfav_get()
 *
 * Description:
 * Retrieve attribute value for the attribute and parses the string
 * based on semicolon and send the substring to parse_home_vf and
 * parse_vf_list.
 * Example VF role map string: "HomeLF=1;LFRoleList=security
 * admin:2,4-8,10;user:11-13,15,17,19;admin:22-25,29,31"
 *
 * Input(s):
 * pattr - ontains Radius attributes and VSAs.
 *
 * Output(s):
 * vfRoleMap - Gets the vf and role list from the string and assign
 * it to the VF Role Map.
 *   homeVF - Gets the homeVF from the string and assign it.
 *
 * Returns:
 *   SUCCESS if successful; error otherwise.
 */
static int
brcd_vfav_get(const attribute_t *pattr,	int *vfRoleMap, int *homeVF)
{
	const char *pd;
	int		len, ret = -1, isContextList = 0;
	int		chassis_gid = INVALID_CHASSIS;
	char	tmpbuf[256] = {0}, *sptr = NULL;
	u_char	*avdata = NULL;

	if (!pattr || !vfRoleMap || !homeVF) {
		return (PAM_AUTH_ERR);
	}
	PAM_LOG(LOG_DEBUG, "brcd_vfav_get enter");
	pd = pattr->data + sizeof (int) + sizeof (unsigned char);

	/*
	 * Vendor length includes the Vendor type, length and data fields
	 */
	len = *pd++;
	memcpy(tmpbuf, pd, len-2);

	/*
	 * There are 3 parses for the string. First it parses for
	 * HomeLF.
	 */
	avdata = strtok_r(tmpbuf, ";", &sptr);
	while (avdata != NULL) {
		ret = parse_home_vf(avdata, homeVF);
		if (ret != PAM_SUCCESS) {
			PAM_LOG(LOG_DEBUG, "HomeLF error");
			return (ret);
		}
		if (*homeVF != INVALID_VF) {
			break;
		}
		avdata = strtok_r(NULL, ";", &sptr);
	}

	sptr = NULL;
	memcpy(tmpbuf, pd, len-2);
	/*
	 * Now it parses for chassis role. We are interested only first definition.
	 * If there are multiple instances, ignore them.
	 */
	avdata = strtok_r(tmpbuf, ";", &sptr);
	while ((avdata != NULL) && (vfRoleMap[CHASSIS_ID] == INVALID_CHASSIS)) {
		ret = parse_chassis_role(avdata, &chassis_gid);
		if (ret != PAM_SUCCESS) {
			PAM_LOG(LOG_DEBUG, "Chassis role error");
			return (ret);
		}
		if (chassis_gid != INVALID_CHASSIS) {
			vfRoleMap[CHASSIS_ID] = chassis_gid;
			break;
		}
		avdata = strtok_r(NULL, ";", &sptr);
	}

	/*
	 * Now it parses for LFRoleList.
	 */
	memcpy(tmpbuf, pd, len-2);

	avdata = strtok_r(tmpbuf, ";", &sptr);
	while ((avdata != NULL)) {
		/* Searches for "=" and if it is contextRoleList strips
		 * "LFRoleList=" part. */
		if (strchr(avdata, '=') != NULL) {
			if(strncasecmp(avdata, BRCD_ATTR_VFROLEMAP,
				strlen(BRCD_ATTR_VFROLEMAP)) == 0) {
				avdata = avdata + strlen(BRCD_ATTR_VFROLEMAP) + 1;
				/* This flag is useful to specify role list only after
				 * context role list or after other role list. */
				isContextList = 1;
			} else if(!strncasecmp(avdata, BRCD_ATTR_HOMEVF,
				strlen(BRCD_ATTR_HOMEVF)) || !strncasecmp(avdata,
				BRCD_ATTR_CHASSISROLE, strlen(BRCD_ATTR_CHASSISROLE)) || 
					!strncasecmp(avdata, BRCD_ATTR_ADLIST, 
						strlen(BRCD_ATTR_ADLIST)) || 
					!strncasecmp(avdata, BRCD_ATTR_HOMEAD, 
						strlen(BRCD_ATTR_HOMEAD))) {
				/*Skip if it is HomeLF or ChassisRole attribute*/
				isContextList = 0;
				avdata = strtok_r(NULL, ";", &sptr);
				continue;
			} else {
				/* Return error if it is neither. */
				return (PAM_RADIUS_IMPROPER_SYNTAX);
			}
		} else {
			/* If role list doesn't follow "LFRoleList" string or
			 * other role list, it returns error.
			 */
			if (!isContextList) {
				return (PAM_RADIUS_IMPROPER_SYNTAX);
			}
		}
		ret = parse_vf_list(avdata, vfRoleMap);
		if (ret != PAM_SUCCESS) {
			return (ret);
		}

		avdata = strtok_r(NULL, ";", &sptr);
	}

	return (PAM_SUCCESS);
}

/*
 * Retrieve attribute value for the specified name. It copies the value
 * whose name matches the input attr_name. Subsequent values with matching
 * name in the the same attribute data field are ignored. Funtion returns
 * -1 for error or not found, 0 for found.
 */
static int
brcd_av_get(const attribute_t *pattr, const char *attr_name, char *buf,
			int buflen)
{
	const u_char *pd;
	int	len, ret = -1;
	char	*pa, *pb, tmpbuf[256] = {0};

	if (!pattr || !attr_name || !buf) {
		return (ret);
	}

	pd = pattr->data + sizeof (int) + sizeof (unsigned char);

	/*
	 * Vendor length includes the Vendor type, length and data fields
	 */
	len = *pd++;
	memcpy(tmpbuf, pd, len-2);

	/* search for the key=value sequence */
	pa = pb = tmpbuf;
	while (*pa) {
		while (*pb && (*pb != '=')) pb++;
		if (*pb) {
			*pb = '\0';
			PAM_LOG(LOG_DEBUG, "got key: %s", pa);
			if (strcmp(pa, attr_name) == 0) {
				/* find key, looking for value */
				*pb++ = '=';
				pa = pb;
				while (*pb && (*pb != ';')) pb++;
				if (pa < pb) {
					/* there is value field */
					if (*pb) {
						*pb = '\0';
					}
					PAM_LOG(LOG_DEBUG, "got value: %d", pa);
					strncpy(buf, pa, buflen);
					ret = 0;
					break;
				}
			} else {
				/* key not match, skip */
				*pb++ = '=';
				pa = pb;
				while (*pb && (*pb != ';')) pb++;
				if (*pb)	pb++;
				pa = pb;
			}
		} else {
			/* reach the end */
			pa = pb;
		}
	}
	return (ret);
}

#define	ACCT_MAX_AD	256
#define BM_NBBY 8                                       /* # bits per byte */
#define BM_NBPF (sizeof (unsigned long) * BM_NBBY)      /* # bits per field */
#define BM_SetBitN(n, p)        ((p)[(n)/BM_NBPF] |= \
                         ((unsigned)(1 << ((n) % BM_NBPF))))
#define	BM_isBitNSet(n, p)		((p))[(n)/BM_NBPF] & \
						((unsigned)(1 << ((n) % BM_NBPF)))
typedef unsigned long admap_t[ACCT_MAX_AD>>5];

/*
 * Upon receiving Access-Accept response from the server, get the AD member
 * list from the Vendor-Specific attribute. Return the PAM error code.
 */
static int
get_ad_list(const AUTH_HDR *response, char *adbuf, int buflen,
			char *homebuf, int hbuflen)
{
    attribute_t	*brcd_vsa;
	char	val_buf[256] = { 0 };
	int		type = BRCD_TYPE_AV1;
    int		ret = -1;

	if (!response) {
		return (PAM_AUTH_ERR);
	}

	memset(adbuf, 0, buflen);
	memset(homebuf, 0, hbuflen);
	do {
		brcd_vsa = find_vsa(response, BRCD_VENDOR_CODE, type);
		if (brcd_vsa) {
			ret = brcd_av_get(brcd_vsa, BRCD_ATTR_ADLIST, val_buf,
							  sizeof (val_buf));
			if (!ret) {
				if (adbuf[0] && !adbuf[buflen - 1] &&
					adbuf[strlen(adbuf) - 1] != ',') {
					adbuf[strlen(adbuf)] = ',';
				}
				strncat(adbuf, val_buf, buflen - strlen(adbuf));
			}
			/* retrieve the first valid HomeAd value */
			if (!homebuf[0]) {
				ret = brcd_av_get(brcd_vsa, BRCD_ATTR_HOMEAD, val_buf,
								  sizeof (val_buf));
				if (!ret) {
					char	*p = NULL;

					PAM_LOG(LOG_DEBUG, "Got home AD: %s", val_buf);
					strncat(homebuf, val_buf, hbuflen);
				}
			}
		}
		type++;
	} while (type <= BRCD_TYPE_AV4);

	if (adbuf[0]) {
		PAM_LOG(LOG_DEBUG, "Got AD List: %s", adbuf);
	}

	return (PAM_SUCCESS);
}

/*
 * Upon receiving Access-Accept response from the server, get the VF
 * Role Map from the Vendor-Specific attribute.
 * Return the PAM error code.
 */
static int
get_vf_list(const AUTH_HDR *response, int *vfRoleMap, int *homeVF)
{
    attribute_t	*brcd_vsa;
	int		type = BRCD_TYPE_AV1;
    int		ret = -1;

	if (!response || !vfRoleMap || !homeVF) {
		return (PAM_AUTH_ERR);
	}

	do {
		brcd_vsa = find_vsa(response, BRCD_VENDOR_CODE, type);

		if (brcd_vsa) {
			/* retrieve vfRoleMap and  the first valid HomeVF value */
			ret = brcd_vfav_get(brcd_vsa, vfRoleMap, homeVF);
			if (ret != PAM_SUCCESS) {
				return (ret);
			}
		}
		type++;
	} while (type <= BRCD_TYPE_AV4);

	return (PAM_SUCCESS);
}

/*
 * Upon receiving Access-Accept response from the server, find out the
 * assigned role, ADlist and homeAD, or VF Role Map and homeVF from the
 * Vendor-Specific attributes. Return the PAM error code.
 */
static int
role_to_user(pam_handle_t *pamh, const AUTH_HDR *response, const char *user, int default_vf)
{
	attribute_t	*brcd_vsa;
	char	role[MAXPASS];
	int		retval;
	const char * vf_enabled_str = NULL;
	int vf_enabled = 0;
	char *end_ptr;

    if (!response || !pamh || !user) {
		return (PAM_AUTH_ERR);
    }

	/* Find whether VF is enabled or not from PAM envlist */
	vf_enabled_str = pam_getenv(pamh, "VF_ENABLED");
	if (vf_enabled_str) {
		/* In case vf environmental variable is not available and by
		 * default, it is assumed as non VF environment
		 */
		vf_enabled = strtol(vf_enabled_str , &end_ptr, 10);
	}
	PAM_LOG(LOG_DEBUG, "VF_Enabled:%d", vf_enabled);

    brcd_vsa = find_vsa(response, BRCD_VENDOR_CODE, BRCD_TYPE_AUTH_ROLE);
    if (brcd_vsa) {
		int		len;
		int 	i = 0;
		unsigned char	*pd;

		PAM_LOG(LOG_DEBUG, "found BRCD VSA");
		/*
		 * data format for Vendor-Specific attributes
		 * Vendor ID			4
		 * Vendor type 			1
		 * Vendor length			1
		 * Attribute-Specific data	m
		 */
		pd = brcd_vsa->data + sizeof (int) + sizeof (unsigned char);

		/*
		 * Vendor length includes the Vendor type, length and data fields
		 */
		len = *pd++;

		/* make sure response packet payload doesn't overflow our buffer */
		if (len - 2 > MAXPASS - 1) {
			memcpy(role, pd, MAXPASS - 1);
			role[MAXPASS - 1] = '\0';
		} else {
			memcpy(role, pd, len - 2);
			role[len - 2] = '\0';
		}
        /* convert mix case role string to lower case */
        while (role[i] != '\0') {
            role[i] = tolower(role[i]);
            i++;
        }
    } else {
		PAM_LOG(LOG_DEBUG, "BRCD VSA not found, use default role");
		rad_converse(pamh, PAM_ERROR_MSG, "Switch role not specified, use "
					 "default.", NULL);
		/* TODO: Incase of VF, it should take the default VF's role. */
		strncpy(role, BRCD_DEFAULT_ROLE, sizeof (role));
    }
    PAM_LOG(LOG_DEBUG, "role buffer: %s", role);

    /* make sure the selected role is allowed by switch security policy. */
    {
		char		cmd_buf[256+MAXPASS];
		const char	*pinst;
		int 		roleValid;
		/* get the switch instance from PAM envlist */
		pinst = pam_getenv(pamh, "FABOS_SWITCHNO");
		if (pinst == NULL) {
			pinst = "0";
		}

		snprintf(cmd_buf, sizeof (cmd_buf),
				"/fabos/libexec/secRoleCheck %s > /dev/null 2>&1",
				role);
		retval = system(cmd_buf);
		if ((retval != 0) && (errno != ECHILD)) {
			PAM_LOG(LOG_DEBUG,"secRoleCheck command failed");
		}
  		roleValid = WEXITSTATUS(retval);
		PAM_LOG(LOG_DEBUG, "check role %s returns %d",
				role, roleValid);
		if (roleValid) {
			PAM_LOG(LOG_DEBUG,"Invalid Role");
                        rad_converse(pamh, PAM_ERROR_MSG, "User specified role invalid, use"
		      			 " default.", NULL);
		        /* TODO: Incase of VF, it should take the default VF's role. */
		        strncpy(role, BRCD_DEFAULT_ROLE, sizeof (role));
		}
    }

	if (!vf_enabled) {
		char	adbuf[BUFFER_SIZE*2] = {0};
		char	homebuf[256] = {0};
		char	cmdbuf[BUFFER_SIZE*3];
		admap_t	admap = {0};
		unsigned long *mapdata;
		int		*pHome;
		int		homeAd = -1;
		FILE	*pfile;

		/* Fix for defect 639516 - Remote user with root role is not allowed */
		if (strcasecmp(role, "root") == 0) {
			return (PAM_AUTH_ERR);
		}

		retval = 0;
		get_ad_list(response, adbuf, sizeof (adbuf), homebuf,
					sizeof (homebuf));
		PAM_LOG(LOG_DEBUG, "radius account: %s %s %s %s", user, role, adbuf,
				homebuf);

		/* use popen to parse, convert & validate AD input buffer */
		if (adbuf[0]) {
			if (homebuf[0]) {
				snprintf(cmdbuf, sizeof (cmdbuf),
						 "/fabos/libexec/secADParse -h \"%s\" -a \"%s\" "
						 "2> /dev/null", homebuf, adbuf);
			} else {
				snprintf(cmdbuf, sizeof (cmdbuf),
						 "/fabos/libexec/secADParse -a \"%s\" 2> /dev/null",
						 adbuf);
			}

			PAM_LOG(LOG_DEBUG, "cmd: %s", cmdbuf);
			pfile = popen(cmdbuf, "r");
			if (!pfile) {
				PAM_LOG(LOG_WARNING, "popen failed %d (%s)", errno,
						strerror(errno));
				retval = -1;
			} else {
				fscanf(pfile, "%lx:%lx:%lx:%lx:%lx:%lx:%lx:%lx", &admap[0],
					   &admap[1], &admap[2], &admap[3], &admap[4], &admap[5],
					   &admap[6], &admap[7]);
				fscanf(pfile, "%d", &homeAd);
				retval = pclose(pfile);
				/*
				* If SIGCHLD is set to SIG_IGN, pclose will
				* always return -1 and errno will be set to
				* ECHILD. So to ignore this case, if errno is
				* ECHILD always set retval to 0
				*/
				if ((retval == -1) && (errno == ECHILD))
					retval = 0;
				PAM_LOG(LOG_DEBUG, "admap: %lx:%lx:%lx:%lx:%lx:%lx:%lx:%lx, "
						"homeAD: %d", admap[0], admap[1], admap[2], admap[3],
						admap[4], admap[5], admap[6], admap[7], homeAd);
			}
		} else {
			/* empty AD member list, default to AD 0 */
			BM_SetBitN(0, admap);
			homeAd = 0;
			retval = PAM_SUCCESS;
		}

		/* if there is an error, just set AD bitmap to all zero */
		if (retval) {
			memset(admap, 0, sizeof (admap));
			retval = PAM_SUCCESS;
		}

		/* set AD information to PAM data for other modules to use */
		mapdata = malloc(sizeof (admap));
		if (!mapdata) {
			retval = PAM_AUTH_ERR;
		} else {
			memcpy(mapdata, admap, sizeof (admap));
			retval = pam_set_data(pamh, PAM_RADIUS_ADMAP, (void *) mapdata,
								  _char_free);
		}
		if (retval == PAM_SUCCESS) {
			pHome = malloc(sizeof(int));
			if (!pHome) {
				retval = PAM_AUTH_ERR;
			} else {
				*pHome = homeAd;
				retval = pam_set_data(pamh, PAM_RADIUS_HOMEAD, (void *) pHome,
									  _int_free);
			}
		}
	} else {
		int		*pHomeVF;
		int		homeVF = INVALID_VF;
		int		*vfRoleMap, i, gid, novf_flag = 0;

		retval = 0;

		/* Allocate memory for VF Role Map and initialize it to
			INVALID_VF */
		vfRoleMap = malloc((PAM_RADIUS_MAXVF) * sizeof (int));
		if (!vfRoleMap) {
			return (PAM_AUTH_ERR);
		}
		for (i = 0; i < PAM_RADIUS_MAXVF; i++) {
			vfRoleMap[i] = INVALID_VF;
		}

		/* Get the VF list from the VSA */
		retval = get_vf_list(response, vfRoleMap,  &homeVF);

		/*
		 * As per the functional spec, if user made a mistake in
		 * specifying VF Role Map it takes the default one.
		 */
		if (retval != PAM_SUCCESS) {
			for (i = 0; i < PAM_RADIUS_MAXVF; i++) {
				vfRoleMap[i] = INVALID_VF;
			}
			homeVF = INVALID_VF;
			novf_flag = 1;
		}

		gid = getgroupid(role);
		if (gid == PAM_RADIUS_INVALID_ROLE) {
			return (PAM_AUTH_ERR);
		}

		/* User's homeVF need to be added into the VF list. */
		if (homeVF != INVALID_VF) {
			vfRoleMap[homeVF] = gid;
		}

		/* If there is no vf list or pam error, add the default VF */
		if (!novf_flag) {
			for (i = 0; i < (PAM_RADIUS_MAXVF - 1); i++) { /* DEFECT 237446 */
				if (vfRoleMap[i] != INVALID_VF) {
					break;
				}
			}

			/* It reaches the end of loop. No vf list is there */
			if (i == (PAM_RADIUS_MAXVF - 1)) {   /* DEFECT 237446 */
				novf_flag = 1;
			}
		}
		if (novf_flag) {
			PAM_LOG(LOG_DEBUG, "Radius default vf %d was set"
				" with role %s", default_vf, role);
			vfRoleMap[default_vf] = gid;
			homeVF = default_vf;
		}

		retval = pam_set_data(pamh, PAM_RADIUS_VFROLEMAP,
							(void *)vfRoleMap, _char_free);
		PAM_LOG(LOG_DEBUG, "Radius account: %s %s %d",
				user, role, homeVF);

		if (retval == PAM_SUCCESS) {
			pHomeVF = malloc(sizeof(int));
			if (!pHomeVF) {
				retval = PAM_AUTH_ERR;
			} else {
				*pHomeVF = homeVF;
				retval = pam_set_data(pamh, PAM_RADIUS_HOMEVF,
								 	(void *)pHomeVF, _int_free);
			}
		}
	}

	/* Inform the user about password expiry days. If password
	 * is expired return error. */
	retval = warn_passwd_expiry(pamh, response);
	if ((retval == PAM_RADIUS_PASSWORD_EXPIRED) || (retval ==
		PAM_AUTH_ERR)) {
		return (PAM_AUTH_ERR);
	}

    /* save the login name to RADIUS server to data */
    {
		char *ouser = malloc( strlen(user) + 1);
		if (ouser == NULL) {
			retval = PAM_AUTH_ERR;
		} else {
			strcpy(ouser, user);
			retval = pam_set_data(pamh, PAM_RADIUS_UNAME, (void *) ouser,
								  _char_free);
		}
    }

    /* change the user name to the one given by role */
    if (retval == PAM_SUCCESS) {
		/* non-secure mode or non-FCS, convert nonfcsadmin role to admin */
		if (strcmp(role, "nonfcsadmin") == 0) {
			snprintf(role, sizeof (role), "admin");
		}
		retval = pam_set_item(pamh, PAM_USER, role);
	}

	/* add the successful (live) server name to the PAM data */
	if (retval == PAM_SUCCESS) {
		if(NULL == live_server) {
			return retval;
		}
		char	*serv_name = strdup(live_server->hostname);
		if (serv_name != NULL) {
			(void) pam_set_data(pamh, PAM_RADIUS_LIVESERV,
								(void *)serv_name, _char_free);
		}
	}

    return (retval);
}

/*
 * read the content of file /tmp/pamdebug.conf to determine the debug
 * print out level.
 */
static void
dbg_init(void)
{
	FILE	*fdbg;

	fdbg = fopen("/tmp/pamdebug.conf", "r");
	if (fdbg == NULL) {
		log_level = LOG_EMERG;
		return;
	}
	if (fscanf(fdbg, "%d", &log_level) <= 0) {
		log_level = LOG_EMERG;
	} else if (log_level < 0) {
		log_level = LOG_EMERG;
	}
	(void) fclose(fdbg);
	return;
}

/*
 * Add an attribute to a RADIUS packet.
 */
static void
add_attribute(AUTH_HDR *request, unsigned char type, CONST unsigned char *data, int length)
{
  attribute_t *p;

  p = (attribute_t *) ((unsigned char *)request + ntohs(request->length));
  p->attribute = type;
  p->length = length + 2;		/* the total size of the attribute */
  request->length = htons(ntohs(request->length) + p->length);
  memcpy(p->data, data, length);
}

/*
 * Add an integer attribute to a RADIUS packet.
 */
static void
add_int_attribute(AUTH_HDR *request, unsigned char type, int data)
{
  int value = htonl(data);

  add_attribute(request, type, (unsigned char *) &value, sizeof(int));
}

/*
 * Remove an attribute from a RADIUS packet.
 */
static void
remove_attribute(AUTH_HDR *request, unsigned char type)
{
  attribute_t	*p;
  int			length;

  p = find_attribute(request, type);
  if (!p) {
	_pam_log(LOG_DEBUG, "attribute type: %u not found", type);
	return;
  }

  length = ((unsigned char *)p - (unsigned char *)request) + p->length;
  _pam_log(LOG_DEBUG, "length with attr: %d, attr_len: %d", length, p->length);

  length = ntohs(request->length) - length;
  _pam_log(LOG_DEBUG, "total length: %d, length left: %d",
		   ntohs(request->length), length);

  request->length = htons(ntohs(request->length) - p->length);
  _pam_log(LOG_DEBUG, "new total length: %d", ntohs(request->length));

  if (length > 0) {
	_pam_log(LOG_DEBUG, "copy %d bytes from %p to %p", length,
			 ((unsigned char*)p) + p->length, p);
	memmove(p, ((unsigned char *)p) + p->length, length);
  } else {
	_pam_log(LOG_DEBUG, "set %d bytes from %p to 0", p->length, p);
	memset(p, 0, p->length);
  }
}

/*
 * Add a RADIUS password attribute to the packet.  Some magic is done here.
 *
 * If it's an PW_OLD_PASSWORD attribute, it's encrypted using the encrypted
 * PW_PASSWORD attribute as the initialization vector.
 *
 * If the password attribute already exists, it's over-written.  This allows
 * us to simply call add_password to update the password for different
 * servers.
 */
static void
add_password(AUTH_HDR *request, unsigned char type, CONST char *password, char *secret)
{
  MD5_CTX md5_secret, my_md5;
  unsigned char misc[AUTH_VECTOR_LEN];
  int i;
  int length = strlen(password);
  char hashed[256 + AUTH_PASS_LEN]; /* can't be longer than this */
  char *vector;
  attribute_t *attr;

  if (length > MAXPASS) {	/* shorten the password for now */
    length = MAXPASS;
  }

  if (length == 0) {
    length = AUTH_PASS_LEN;	/* 0 maps to 16 */
  } if ((length & (AUTH_PASS_LEN - 1)) != 0) {
    length += (AUTH_PASS_LEN - 1); /* round it up */
    length &= ~(AUTH_PASS_LEN - 1); /* chop it off */
  }                             /* 16*N maps to itself */

  memset(hashed, '\0', length);
  memcpy(hashed, password, length);

  /* remove CHAP password from previous config */
  remove_attribute(request, PW_CHAP_PASSWORD);

  attr = find_attribute(request, PW_PASSWORD);

  if (type == PW_PASSWORD) {
    vector = request->vector;
  } else {
    vector = attr->data;	/* attr CANNOT be NULL here. */
  }

  /* ************************************************************ */
  /* encrypt the password */
  /* password : e[0] = p[0] ^ MD5(secret + vector) */
  MD5Init(&md5_secret);
  MD5Update(&md5_secret, secret, strlen(secret));
  my_md5 = md5_secret;		/* so we won't re-do the hash later */
  MD5Update(&my_md5, vector, AUTH_VECTOR_LEN);
  MD5Final(misc, &my_md5);      /* set the final vector */
  xor(hashed, misc, AUTH_PASS_LEN);

  /* For each step through, e[i] = p[i] ^ MD5(secret + e[i-1]) */
  for (i = 1; i < (length >> 4); i++) {
    my_md5 = md5_secret;	/* grab old value of the hash */
    MD5Update(&my_md5, &hashed[(i-1) * AUTH_PASS_LEN], AUTH_PASS_LEN);
    MD5Final(misc, &my_md5);      /* set the final vector */
    xor(&hashed[i * AUTH_PASS_LEN], misc, AUTH_PASS_LEN);
  }

  if (type == PW_OLD_PASSWORD) {
    attr = find_attribute(request, PW_OLD_PASSWORD);
  }

  if (!attr) {
    add_attribute(request, type, hashed, length);
  } else {
    memcpy(attr->data, hashed, length); /* overwrite the packet */
  }
}

/*
 * int add_chap_password()
 *
 * Description:
 *   Add a RADIUS CHAP-PASSWORD attribute to the packet.
 *
 *   This uses the input password as the secret to construct the
 *   CHAP response. In this model, the NAS generates random number as
 *   CHAP challenge, compute the response and send as CHAP-PASSWORD
 *   in the Access-Request
 *
 * Input(s):
 *   request - pointer to RADIUS request header
 *   chap_id - CHAP ID
 *   password - password string to do CHAP
 *
 * Output(s):
 *   request - pointer to RADIUS request header
 *
 * Return:
 *  0 success; -1 failure.
 */
static int
add_chap_password(AUTH_HDR *request, u_char chap_id, CONST char *password)
{
  unsigned char	resp[CHAP_VALUE_LENGTH+1];
  unsigned char	string[MAXPASS+AUTH_VECTOR_LEN+1], *ptr = string;
  int	length;
  MD5_CTX	md5_context;
  attribute_t	*attr;

  length = strlen(password);
  if (length <= 0 || length > MAXPASS) {
    _pam_log(LOG_ERR, "invalid password length: %d.", length);
    return (-1);
  }

  *ptr++ = chap_id;
  memcpy(ptr, password, length);
  ptr += length;
  memcpy(ptr, request->vector, AUTH_VECTOR_LEN);
  ptr += AUTH_VECTOR_LEN;

  resp[0] = chap_id;
  MD5Init(&md5_context);
  MD5Update(&md5_context, string, ptr - string);
  MD5Final(resp+1, &md5_context);

  /* remove password from previous config */
  remove_attribute(request, PW_PASSWORD);

  attr = find_attribute(request, PW_CHAP_PASSWORD);
  if (!attr) {
    add_attribute(request, PW_CHAP_PASSWORD, resp, AUTH_VECTOR_LEN+1);
  } else {
    memcpy(attr->data, resp, AUTH_VECTOR_LEN+1);
  }

  return (0);
}

static void
cleanup(radius_server_t *server)
{
  radius_server_t *next;
#ifdef DEBUG_ZEROIZATION
  FILE *fp = fopen("/dev/console", "w+");
#endif

  while (server) {
    next = server->next;
    _pam_drop(server->hostname);
    #ifdef DEBUG_ZEROIZATION
    fprintf(fp, "ZeroizationLog: In cleanup(): Before memset(server->secret), secret = %s\n", server->secret);
    #endif
     int secretlen = 0;
     if (server->secret != NULL) {
     	secretlen = strlen(server->secret);
     	memset(server->secret, 0, secretlen);
     }
   #ifdef MEMCMP_ZEROIZATION
/* Add code for memcheck for CC requirement*/
	pam_memcmp_zero(server->secret, secretlen);
   #endif
    #ifdef DEBUG_ZEROIZATION
    fprintf(fp, "ZeroizationLog: In cleanup(): After memset(server->secret), secret = %s\n", server->secret);
    fclose(fp);
    #endif
    if (server->secret != NULL) {
	    free(server->secret);
	    server->secret = NULL;
    }
    _pam_drop(server);
    server = next;
  }
}

/*
 * allocate and open a local port for communication with the RADIUS
 * server
 */
static int
initialize(radius_conf_t *conf, int accounting)
{
  struct sockaddr salocal;
  u_short local_port;
  char hostname[BUFFER_SIZE];
  char secret[MAX_LENGTH] = {0};
  char encrypted_secret[MAX_LENGTH] = {0};
  char *psecret = NULL;
  char *decryptedsecret = NULL;
  struct addrinfo hints, *res = NULL;
  int err;
  int ret = 0;
  char port_str[PORT_STR_LEN];

  char buffer[BUFFER_SIZE];
  char *p;
  FILE *fserver;
  radius_server_t *server = NULL;
  struct sockaddr_in * s_in;
  int timeout;
  int encryption_lvl = 0;
  int pap = 0;
  int line = 0;

  /* the first time around, read the configuration file */
  if ((fserver = fopen (conf_file, "r")) == (FILE*)NULL) {
    _pam_log(LOG_ERR, "Could not open configuration file %s: %s\n",
	    conf_file, strerror(errno));
    return PAM_ABORT;
  }

  while (!feof(fserver) &&
	 (fgets (buffer, sizeof(buffer), fserver) != (char*) NULL) &&
	 (!ferror(fserver))) {
    line++;
    p = buffer;

    /*
     *  Skip blank lines and whitespace
     */
    while (*p &&
	   ((*p == ' ') || (*p == '\t') ||
	    (*p == '\r') || (*p == '\n'))) p++;

    /*
     *  Nothing, or just a comment.  Ignore the line.
     */
    if ((!*p) || (*p == '#')) {
      continue;
    }

    timeout = 3;
    if (sscanf(p, "%s %s %d %d %s %d", hostname, secret, &timeout, &pap,
		encrypted_secret, &encryption_lvl) < 2) {
      _pam_log(LOG_ERR, "ERROR reading %s, line %d: Could not read hostname or secret\n",
	       conf_file, line);
      continue; /* invalid line */
    } else {			/* read it in and save the data */
	  _pam_log(LOG_DEBUG,
      "hostname %s, sec %s, timeout %d, pap %d, encr_sec %s, encr_lvl %d\n",
	  hostname, secret, timeout, pap, encrypted_secret, encryption_lvl);
      radius_server_t *tmp;

      tmp = malloc(sizeof(radius_server_t));
      if (server) {
	server->next = tmp;
	server = server->next;
      } else {

	conf->server = tmp;
	server= tmp;		/* first time */
      }

	  /* Changes to remove plaintext secret from RADIUS config file */
	  if (encryption_lvl) {
		  _pam_log(LOG_DEBUG, "Decryption to be done for secret\n");

		  psecret = (char *)malloc(MAX_LENGTH);
		  if (NULL == psecret) {
			  free(tmp);
			  fclose(fserver);
			  memset(secret, 0, MAX_LENGTH);
			  memset(encrypted_secret, 0, MAX_LENGTH);
			  return (PAM_AUTHINFO_UNAVAIL);
		  }
		  memset(psecret, 0, MAX_LENGTH);
		  memcpy(psecret, encrypted_secret, MAX_LENGTH);

		  decryptedsecret = (char *)malloc(MAX_LENGTH);
		  if (decryptedsecret == NULL) {
			  memset(psecret, 0, MAX_LENGTH);
			  free(psecret);
			  free(tmp);
			  fclose(fserver);
			  memset(secret, 0, MAX_LENGTH);
			  memset(encrypted_secret, 0, MAX_LENGTH);
			  return (PAM_AUTHINFO_UNAVAIL);
		  }
		  memset(decryptedsecret, 0, MAX_LENGTH);

		  ret =  decryptPassword(psecret, decryptedsecret);
		  if (ret < 0) {
			  _pam_log(LOG_ERR, "decryptPassword failed in "
			  "configuration file %s psecret %s\n",
			  conf_file, psecret);
			  memset(psecret, 0, MAX_LENGTH);
			  memset(decryptedsecret, 0, MAX_LENGTH);
			  free(psecret);
			  free(decryptedsecret);
			  free(tmp);
			  fclose(fserver);
			  memset(secret, 0, MAX_LENGTH);
			  memset(encrypted_secret, 0, MAX_LENGTH);
			  return (PAM_AUTHINFO_UNAVAIL);
		  }

		  /* Copy decrypted secret to secret for further processing */
		  memcpy(secret, decryptedsecret, MAX_LENGTH);
		  memset(psecret, 0, MAX_LENGTH);
		  memset(decryptedsecret, 0, MAX_LENGTH);
		  free(psecret);
		  free(decryptedsecret);
	  }

      /* sometime later do memory checks here */
      server->hostname = strdup(hostname);
      server->secret = strdup(secret);
      server->accounting = accounting;
      server->pap = pap;
      server->port = 0;

      if ((timeout < 1) || (timeout > 60)) {
	    server->timeout = 3;
      } else {
	    server->timeout = timeout;
      }
     /* retries are 0 by default and are not configurable so maintaining the same value in server configuration */
     server->retries = 0;
      server->next = NULL;
    }
  }
  fclose(fserver);

  if (!server) {		/* no server found, die a horrible death */
    _pam_log(LOG_ERR, "No RADIUS server found in configuration file %s\n",
	     conf_file);
	memset(secret, 0, MAX_LENGTH);
	memset(encrypted_secret, 0, MAX_LENGTH);
    return PAM_AUTHINFO_UNAVAIL;
  }

  /* open a socket.  Dies if it fails */

  conf->sockfd = socket(AF_INET6, SOCK_DGRAM, 0);
  if (conf->sockfd < 0) {
    _pam_log(LOG_ERR, "Failed to open RADIUS socket: %s\n", strerror(errno));
	memset(secret, 0, MAX_LENGTH);
	memset(encrypted_secret, 0, MAX_LENGTH);
    return PAM_AUTHINFO_UNAVAIL;
  }

  /*
   *  Use our process ID as a local port for RADIUS.
   */
  local_port = (getpid() & 0x7fff) + 1024;

  /* zero out hints before initializing */
  memset(&hints, 0, sizeof(hints));
  hints.ai_family = PF_UNSPEC; /* accept any protocol family */
  hints.ai_socktype = SOCK_DGRAM;
  /*
   * the IP address portion of the socket address structure will be set to
   * INADDR_ANY for an IPv4 address or IN6ADDR_ANY_INIT for an IPv6 address.
   */
  hints.ai_flags = AI_PASSIVE;
  hints.ai_protocol = 0;  /* accept any protocol */

  do {
    local_port++;
    snprintf(port_str, PORT_STR_LEN, "%d", htons(local_port));

    if (res) {
      freeaddrinfo(res);
      res = NULL;
    }

    err = getaddrinfo(NULL, port_str, &hints, &res);
    if (err) {
      _pam_log(LOG_ERR, "initialize : getaddrinfo returned error = %d\n", err);
  	  memset(secret, 0, MAX_LENGTH);
	  memset(encrypted_secret, 0, MAX_LENGTH);
      return (PAM_AUTHINFO_UNAVAIL);
    }
  } while ((bind(conf->sockfd, res->ai_addr, res->ai_addrlen) < 0) &&
          (local_port < 64000));

  if (res) {
    freeaddrinfo(res);
    res = NULL;
  }

  if (local_port >= 64000) {
    close(conf->sockfd);
    _pam_log(LOG_ERR, "No open port we could bind to.");
	memset(secret, 0, MAX_LENGTH);
	memset(encrypted_secret, 0, MAX_LENGTH);
    return PAM_AUTHINFO_UNAVAIL;
  }

  memset(secret, 0, MAX_LENGTH);
  memset(encrypted_secret, 0, MAX_LENGTH);
  return PAM_SUCCESS;
}

/*
 * Helper function for building a radius packet.
 * It initializes *some* of the header, and adds common attributes.
 */
static void
build_radius_packet(AUTH_HDR *request, CONST char *user, CONST char *password,
		    radius_conf_t *conf)
{
  char hostname[256];
  UINT4 ipaddr;



  hostname[0] = '\0';
  gethostname(hostname, sizeof(hostname) - 1);

  request->length = htons(AUTH_HDR_LEN);

  if (password) {		/* make a random authentication req vector */
    get_random_vector(request->vector);
  }

  add_attribute(request, PW_USER_NAME, (char *) user, strlen(user));

  /*
   *  Add a password, if given.
   */
  if (password) {
    if (request->code == PW_AUTHENTICATION_REQUEST &&
	(conf->server->pap == 0)) {
      add_chap_password(request, CHAP_MD5, password);
    } else {
      add_password(request, PW_PASSWORD, password, conf->server->secret);
    }

    /*
     *  Add a NULL password to non-accounting requests.
     */
  } else if (request->code != PW_ACCOUNTING_REQUEST) {
    add_password(request, PW_PASSWORD, "", conf->server->secret);
  }

  /* the packet is from localhost if on localhost, to make configs easier */
  if ((conf->server->ip.s_addr == ntohl(0x7f000001)) || (!hostname[0])) {
    ipaddr = 0x7f000001;
  } else {
    struct hostent *hp;

    if ((hp = gethostbyname(hostname)) == (struct hostent *) NULL) {
      ipaddr = 0x00000000;	/* no client IP address */
    } else {
      ipaddr = ntohl(*(UINT4 *) hp->h_addr); /* use the first one available */
    }
  }

  /* If we can't find an IP address, then don't add one */
  if (ipaddr) {
    add_int_attribute(request, PW_NAS_IP_ADDRESS, ipaddr);
  }

  /* There's always a NAS identifier */
  if (conf->client_id && *conf->client_id) {
    add_attribute(request, PW_NAS_IDENTIFIER, conf->client_id,
		  strlen(conf->client_id));
  }

  /* Set remote ID as Calling-Station-ID */
  if (conf->remote_id && *conf->remote_id) {
    add_attribute(request, PW_CALLING_STATION_ID, conf->remote_id,
		  strlen(conf->remote_id));
  }

  /*
   *  Add in the port (pid) and port type (virtual).
   *
   *  We might want to give the TTY name here, too.
   */
  add_int_attribute(request, PW_NAS_PORT_ID, getpid());
  add_int_attribute(request, PW_NAS_PORT_TYPE, PW_NAS_PORT_TYPE_VIRTUAL);
}

/*
 * Talk RADIUS to a server.
 * Send a packet and get the response
 */
static int
talk_radius(radius_conf_t *conf, AUTH_HDR *request, AUTH_HDR *response,
	    char *password, char *old_password, int tries)
{
  int salen, total_length;
  fd_set set;
  struct timeval tv;
  time_t now, end;
  int rcode;
  struct sockaddr saremote;
  struct sockaddr_in *s_in = (struct sockaddr_in *) &saremote;
  radius_server_t *tmp = NULL;
  radius_server_t *rad_live_srv = NULL;
  radius_server_t *server = conf->server;
  int ok;
  int server_tries;
  struct addrinfo hints, *addrlist, *res = NULL;
  int err;
  char port_str[PORT_STR_LEN];
  char *livserver_hostname = NULL;
   const int       *fips_mode = NULL;
  /* ************************************************************ */
  /* Now that we're done building the request, we can send it */

  /*
    Hmm... on password change requests, all of the found server information
    could be saved with a pam_set_data(), which means even the radius_conf_t
    information will have to be malloc'd at some point

    On the other hand, we could just try all of the servers again in
    sequence, on the off chance that one may have ended up fixing itself.

    */
  /* loop over all available servers */
  if (request->code == PW_ACCOUNTING_REQUEST) {
      int hostlen = 0;
      PAM_LOG(LOG_DEBUG, "talk_radius: Entering session for fetching liveserver information \n");

      /* During authentication phase first liveserver information is fetched and if      it is different from existing liveserver information in pam then liveserver     information is written to pam environment */ 
      pam_get_data(pamhandle, PAM_RADIUS_LIVESERV, 
			  (void **)&livserver_hostname);

	  if (livserver_hostname == NULL) { 
     		PAM_LOG(LOG_DEBUG, "Live Sever hostname is NULL\n");
			return PAM_AUTH_ERR; 
	} 
      hostlen = strlen(livserver_hostname);
      if (hostlen > 0) {
		  char *hostname = NULL;
	 	  for (tmp = conf->server; tmp != NULL ; tmp=tmp->next) {
			  if ( (hostname = gethostname_from_server(tmp)) == NULL) 
				  continue;
			  if ((strlen(hostname)== hostlen) &&
				  (memcmp(hostname,livserver_hostname,hostlen) == 0)) 
				  break;
		  }
		  
		  if(NULL != hostname) {
			  free(hostname);
			  hostname = NULL;
		  }
	  	  if (tmp != NULL) {
		  	  rad_live_srv = tmp;
                          /* During session phase the live server that has responded during authentication phase is contacted first instead of looping server list from the beginning */
			  server = rad_live_srv;
			  PAM_LOG(LOG_DEBUG, "talk_radius:The server contacted during session is %s", server->hostname);
	  	  }
      }
  }
  pam_get_data(pamhandle, PAM_RADIUS_FIPSMODE,
                                                   (const void **) &fips_mode);

  if (fips_mode != NULL)
	fipsmode = *fips_mode;

  PAM_LOG(LOG_DEBUG, "fips mode:%d", fipsmode);

  while (server != NULL) {

	/*
	 * Check whether the authentication selected is PEAP
	 , */
	if ((PEAP_PROTOCOL == conf->server->pap) &&
		(PW_AUTHENTICATION_REQUEST == request->code)) {
		int retval = 0;
		char user_name[MAXPWNAM] = {'\0'};
		attribute_t * peap_attr  = 0;

		peap_attr = find_attribute(request, PW_USER_NAME);
		if (!peap_attr) {
			_pam_log(LOG_DEBUG, "attribute type: %u not found", PW_USER_NAME);
			return PAM_CRED_ERR;
		}
		strncpy(user_name, peap_attr->data, peap_attr->length - sizeof(short) );
		user_name [peap_attr->length - sizeof(short)] = '\0';
        PAM_LOG(LOG_DEBUG, "talk_radius: Auth type is PEAP \n");

		retval = do_peap_authentication(conf, user_name, password, response);
		if(0 == retval) {
			PAM_LOG(LOG_DEBUG, "talk_radius: PEAP authentication successful \n");
			ok = TRUE;
			goto next;
			/* return PAM_SUCCESS; */
		} else if (PAM_AUTHINFO_UNAVAIL == retval) {
			ok = FALSE;
			goto next;
			/* server failed to respond */
		} else if (X509_V_ERR_HOSTNAME_MISMATCH == retval) {
			PAM_LOG(LOG_DEBUG, "talk_radius: PEAP authentication failed for HOSTNAME mismatch error \n");


                        ok = FALSE;
                        goto next;
		} else if (X509_V_ERR_CERT_UNKNOWN_TO_OCSP == retval) {
			PAM_LOG(LOG_DEBUG, "talk_radius: PEAP authentication failed for Cert Status Unknown error \n");

			/* host validaiton failed */
			int *ocsp_unknown = malloc( sizeof(int) );
			*ocsp_unknown = X509_V_ERR_CERT_UNKNOWN_TO_OCSP;
			pam_set_data( pamhandle, PAM_RADIUS_CERT_VERIFY, (void *) ocsp_unknown, _int_free );

                        ok = FALSE;
                        goto next;
		} else if (X509_V_ERR_CERT_REVOKED_TO_OCSP == retval) {
			PAM_LOG(LOG_DEBUG, "talk_radius: PEAP authentication failed for Cert Status Revoked error \n");

			/* host validaiton failed */
			int *ocsp_revoked = malloc( sizeof(int) );
			*ocsp_revoked = X509_V_ERR_CERT_REVOKED_TO_OCSP;
			pam_set_data( pamhandle, PAM_RADIUS_CERT_VERIFY, (void *) ocsp_revoked, _int_free );

                        ok = FALSE;
                        goto next;
		} else if (X509_V_ERR_OCSP_APPLICATION_VERIFICATION == retval) {
			PAM_LOG(LOG_DEBUG, "talk_radius: PEAP authentication failed for OCSP Application Verification error \n");

			/* OCSP validaiton failed */
			int *ocsp_app_failed = malloc( sizeof(int) );
			*ocsp_app_failed = X509_V_ERR_OCSP_APPLICATION_VERIFICATION;
			pam_set_data( pamhandle, PAM_RADIUS_CERT_VERIFY, (void *) ocsp_app_failed, _int_free );

                        ok = FALSE;
                        goto next;
		} else {
			/*
			 * User authentication failed. The radius log will show the
			 * information on the failure
			 */
			PAM_LOG(LOG_DEBUG, "talk_radius: peap authentication failed for the user %s error %d \n",
								user_name, retval);
			/*
			 * Once the PEAP authentication fails, we need not send PAP 
			 */
			 if (retval == -2) {
			 	PAM_LOG(LOG_DEBUG, "talk_radius: peap authentication %d \n", retval);
				 retval = PAM_AUTH_ERR;
			 } else {
			 	return (PAM_AUTH_ERR);
			}
		}
	}
    /* only look up IP information as necessary */
    if (host2server(server) != PAM_SUCCESS) {
      _pam_log(LOG_ERR, "Failed looking up IP address for RADIUS server %s",
	       server->hostname);
      ok = FALSE;
      goto next;		/* skip to the next server */
    }

    total_length = ntohs(request->length);

    if (!password) { 		/* make an RFC 2139 p6 request authenticator */
      get_accounting_vector(request, server);
    }

    server_tries = tries;

    /* Zero out hints before initializing */
    memset(&hints, 0, sizeof(hints));
    hints.ai_family = PF_UNSPEC; /* accept any protocol family */
    hints.ai_socktype = SOCK_DGRAM;
    /*
     * the IP address portion of the socket address structure will be set to
     * INADDR_ANY for an IPv4 address or IN6ADDR_ANY_INIT for an IPv6 address.
     */
    hints.ai_flags = AI_PASSIVE;
    hints.ai_protocol = 0;  /* accept any protocol */

    snprintf(port_str, PORT_STR_LEN, "%d", server->port);
    err = getaddrinfo(server->hostname, port_str, &hints, &res);
    if (err) {
      _pam_log(LOG_ERR, "talk_radius: getaddrinfo err = %d\n",  err);
      ok = FALSE;
      goto next;
    }

send:
    for (addrlist = res; addrlist; addrlist = addrlist->ai_next) {
    	/* send the packet */
    	if (sendto(conf->sockfd, (char *) request, total_length, 0,
            addrlist->ai_addr, addrlist->ai_addrlen) < 0) {
            continue; /* skip to the next address */
        } else  {
            break; /* sendto() is successful */
        }
    }

    if (!addrlist) {
        /* Could not send the packet to the server */
        _pam_log(LOG_ERR, "Error sending RADIUS packet to server %s: %s",
            server->hostname, strerror(errno));
        ok = FALSE;
        if (res) {
            freeaddrinfo(res);
        	res = NULL;
        }
        goto next;		/* skip to the next server */
    }

    /* ************************************************************ */
    /* Wait for the response, and verify it. */
    salen = sizeof(struct sockaddr);
    tv.tv_sec = server->timeout; /* wait for the specified time */
    tv.tv_usec = 0;
    FD_ZERO(&set);		/* clear out the set */
    FD_SET(conf->sockfd, &set);	/* wait only for the RADIUS UDP socket */

    time(&now);
    end = now + tv.tv_sec;

    /* loop, waiting for the select to return data */
    ok = TRUE;
    while (ok) {

      rcode = select(conf->sockfd + 1, &set, NULL, NULL, &tv);

      /* select timed out */
      if (rcode == 0) {
	_pam_log(LOG_ERR, "RADIUS server %s failed to respond",
		 server->hostname);
	if (--server_tries)
	  goto send;
	ok = FALSE;
	break;			/* exit from the select loop */
      } else if (rcode < 0) {

	/* select had an error */
	if (errno == EINTR) {	/* we were interrupted */
	  time(&now);

	  if (now > end) {
	    _pam_log(LOG_ERR, "RADIUS server %s failed to respond",
		     server->hostname);
	    if (--server_tries)
	      goto send;
	    ok = FALSE;
	    break;			/* exit from the select loop */
	  }

	  tv.tv_sec = end - now;
	  if (tv.tv_sec == 0) {	/* keep waiting */
	    tv.tv_sec = 1;
	  }

	} else {		/* not an interrupt, it was a real error */
	  _pam_log(LOG_ERR, "Error waiting for response from RADIUS server %s: %s",
		   server->hostname, strerror(errno));
	  ok = FALSE;
	  break;
	}

	/* the select returned OK */
      } else if (FD_ISSET(conf->sockfd, &set)) {

	/* try to receive some data */
	if ((total_length = recvfrom(conf->sockfd, (char *) response,
			BUFFER_SIZE, 0,
			addrlist->ai_addr, &addrlist->ai_addrlen)) < 0) {
	  _pam_log(LOG_ERR, "error reading RADIUS packet from server %s: %s",
		   server->hostname, strerror(errno));
	  ok = FALSE;
	  break;

	  /* there's data, see if it's valid */
	} else {
	  char *p = server->secret;

	  if ((ntohs(response->length) != total_length) ||
	      (ntohs(response->length) > BUFFER_SIZE)) {
	    _pam_log(LOG_ERR, "RADIUS packet from server %s is corrupted",
		     server->hostname);
	    ok = FALSE;
	    break;
	  }

	  /* Check if we have the data OK.  We should also check request->id */

	  if (password) {
	    if (old_password) {
#ifdef LIVINGSTON_PASSWORD_VERIFY_BUG_FIXED
	      p = old_password;	/* what it should be */
#else
	      p = "";		/* what it really is */
#endif
	    }
	    /*
	     * RFC 2139 p.6 says not do do this, but the Livingston 1.16
	     * server disagrees.  If the user says he wants the bug, give in.
	     */
	  } else {		/* authentication request */
	    if (conf->accounting_bug) {
	      p = "";
	    }
	  }

	  if (!verify_packet(p, response, request)) {
	    _pam_log(LOG_ERR, "packet from RADIUS server %s fails verification: The shared secret is probably incorrect.",
		     server->hostname);
	    ok = FALSE;
	    break;
	  }

	  /*
	   * Check that the response ID matches the request ID.
	   */
	  if (response->id != request->id) {
	    _pam_log(LOG_WARNING, "Response packet ID %d does not match the request packet ID %d: verification of packet fails", response->id, request->id);
	      ok = FALSE;
	    break;
	  }
	}

	/*
	 * Whew!  The select is done.  It hasn't timed out, or errored out.
	 * It's our descriptor.  We've got some data.  It's the right size.
	 * The packet is valid.
	 * NOW, we can skip out of the select loop, and process the packet
	 */
	_pam_log(LOG_DEBUG, "got proper response from server %s.",
		 server->hostname);
	break;
      }
      /* otherwise, we've got data on another descriptor, keep select'ing */
    }

    if (res) {
        freeaddrinfo(res);
        res = NULL;
    }

    /* go to the next server if this one didn't respond */
  next:
    if (!ok) {
      radius_server_t *old;	/* forget about this server */

      old = server;
     if (rad_live_srv != NULL) {
       server = conf->server;
       rad_live_srv = NULL;
     } else {
       server = server->next;
     }
      conf->server = server;


      if (server) {		/* if there's more servers to check */
	/* get a new authentication vector, and update the passwords */
	get_random_vector(request->vector);
	request->id = request->vector[0];

	/* update passwords, as appropriate */
	if (password) {
	  get_random_vector(request->vector);
	  if (old_password) {	/* password change request */
	    add_password(request, PW_PASSWORD, password, old_password);
	    add_password(request, PW_OLD_PASSWORD, old_password, old_password);
	  } else if (server->pap == 0) {	/* authentication request */
	    add_chap_password(request, CHAP_MD5, password);
	  } else {
	    add_password(request, PW_PASSWORD, password, server->secret);
	  }
	}
      }
      continue;

    } else {
      /* we've found one that does respond, forget about the other servers */
      cleanup(server->next);
      server->next = NULL;
      live_server = server;	/* we've got a live one! */
     if (request->code == PW_AUTHENTICATION_REQUEST) {
       char *serv_name = strdup(live_server->hostname);
       (void)pam_set_data(pamhandle, PAM_RADIUS_LIVESERV, (void *)serv_name, _char_free);
     }	 
      break;
    }
  }
  
  if (!server) {
    _pam_log(LOG_ERR, "All RADIUS servers failed to respond.");
    return PAM_AUTHINFO_UNAVAIL;
  }

  return PAM_SUCCESS;
}

/**************************************************************************
 * MIDLEVEL PAM CODE
 **************************************************************************/

/* this is our front-end for module-application conversations */

#undef PAM_FAIL_CHECK
#define PAM_FAIL_CHECK if (retval != PAM_SUCCESS) { _pam_log(LOG_ERR, "pam auth failed: %d.", retval); return PAM_AUTH_ERR;}

static int rad_converse(pam_handle_t *pamh, int msg_style, char *message, char **password)
{
  CONST struct pam_conv *conv;
  struct pam_message resp_msg;
  CONST struct pam_message *msg[1];
  struct pam_response *resp = NULL;
  int retval;

  resp_msg.msg_style = msg_style;
  resp_msg.msg = message;
  msg[0] = &resp_msg;

  /* grab the password */
  retval = pam_get_item(pamh, PAM_CONV, (CONST void **) &conv);
  PAM_FAIL_CHECK;

  retval = conv->conv(1, msg, &resp,conv->appdata_ptr);
  PAM_FAIL_CHECK;

  if (password) {		/* assume msg.type needs a response */
    /* I'm not sure if this next bit is necessary on Linux */
#ifdef sun
    /* NULL response, fail authentication */
    if ((resp == NULL) || (resp->resp == NULL)) {
      return PAM_SYSTEM_ERR;
    }
#endif

    *password = resp->resp;
    free(resp);
  }

  return PAM_SUCCESS;
}

/**************************************************************************
 * GENERAL CODE
 **************************************************************************/

#undef PAM_FAIL_CHECK
#define PAM_FAIL_CHECK if (retval != PAM_SUCCESS) {	\
	int *pret = malloc( sizeof(int) );		\
	*pret = PAM_AUTH_ERR;					\
	pam_set_data( pamh, "rad_setcred_return"	\
	              , (void *) pret, _int_free );	\
  	_pam_log(LOG_ERR, "pam auth failed: %d.", retval); \
	return PAM_AUTH_ERR; }

#define DPRINT if (ctrl & PAM_DEBUG_ARG) _pam_log

/*
 * do_peap_authentication()
 *
 * This function takes the RADIUS server list, user name and password
 * and authenticate them against the RADIUS server using PEAP protocol.
 * For PEAP authentication, it uses the libradpeapclient.so library.
 *
 * Input:
 *		radius_conf_t * config
 *			Pointer to the radius configuration
 *		char * user
 *			User name
 *		char * passord
 *			User password
 *		AUTH_HDR *response
 *			Buffer for RADIUS accept response packet
 *
 * Output:
 *		Data contained in the RADIUS accept packet (in case of success)
 *
 * Returns:
 *		0 for success
 *		-1 for Failure
 *		PAM_AUTHINFO_UNAVAIL
 */
static int
do_peap_authentication(radius_conf_t * config, CONST char * user, char * password,
	AUTH_HDR *response)
{
	radius_server_t *server = 0;
	int retval = 0;

	server = config->server;


		if (host2server(server) != PAM_SUCCESS) {
			PAM_LOG(LOG_DEBUG, "do_peap_authentication: Error while getting server \n");
			PAM_LOG(LOG_DEBUG, "do_peap_authentication: Skipping the current server \n");
		}

		if ((server->ip.s_addr = get_ipaddr(server->hostname)) == (UINT4)0) {
		}
	/*
	 * Call the PEAP authentication module for further processing
	 */
	retval = pam_peap_authenticate(user, password, config, response);

	return retval;
}

PAM_EXTERN int
pam_sm_authenticate(pam_handle_t *pamh,int flags,int argc,CONST char **argv)
{
  CONST char *user;
  char *password = NULL;
  char *resp2challenge = NULL;
  int ctrl;
  int retval = PAM_AUTH_ERR;
  char host_name[_SC_HOST_NAME_MAX] = {0};
  char recv_buffer[4096];
  char send_buffer[4096];
  AUTH_HDR *request = (AUTH_HDR *) send_buffer;
  AUTH_HDR *response = (AUTH_HDR *) recv_buffer;
  radius_conf_t config;
  int tries,default_vf = 0;
  FILE * pfile = NULL;
  const char * vf_enabled_str = NULL, *end_ptr;
  int vf_enabled = 0;
  char cmd_buf[256] = {0};

  /* Find whether VF is enabled or not from PAM envlist */
  vf_enabled_str = pam_getenv(pamh, "VF_ENABLED");
  if (vf_enabled_str) {
		vf_enabled = strtol(vf_enabled_str , &end_ptr, 10);
  }
  PAM_LOG(LOG_DEBUG, "VF_Enabled:%d", vf_enabled);

  /* initialize debug log level */
  dbg_init();

  ctrl = _pam_parse(argc, argv, &config);
  /* Read the tries from PAM configuration file */
  tries = ((ctrl & PAM_RETRY) >> 4) + 1;
  /* grab the user name */
  retval = pam_get_user(pamh, &user, NULL);
  PAM_FAIL_CHECK;

  /* check that they've entered something, and not too long, either */
  if ((user == NULL) ||
      (strlen(user) > MAXPWNAM)) {
    int *pret = malloc( sizeof(int) );
    *pret = PAM_USER_UNKNOWN;
    pam_set_data( pamh, "rad_setcred_return", (void *) pret, _int_free );

    DPRINT(LOG_DEBUG, "User name was NULL, or too long");
    return PAM_USER_UNKNOWN;
  }

  DPRINT(LOG_DEBUG, "Got user name %s", user);

  /*
   * Get the IP address of the authentication server
   * Then, open a socket, and bind it to a port
   */
  retval = initialize(&config, FALSE);
  PAM_FAIL_CHECK;

  /*
   * If there's no client id specified, use the service type, to help
   * keep track of which service is doing the authentication.
   */
  if (!config.client_id) {
    retval = pam_get_item(pamh, PAM_SERVICE, (CONST void **) &config.client_id);
    PAM_FAIL_CHECK;
  }
  /* now we've got a socket open, so we've got to clean it up on error */
#undef PAM_FAIL_CHECK
#define PAM_FAIL_CHECK if (retval != PAM_SUCCESS) {goto error; }

  /*
   * Defect 85640
   * Nas-Identifier field whose value was PAM_SERVICE
   * will now have  value as `switchname'
   */
  if (vf_enabled) {
  	snprintf(cmd_buf, sizeof (cmd_buf),
      "/fabos/libexec/getDefaultFID");
  	retval = system(cmd_buf);
  	if (retval == -1) {
      PAM_LOG(LOG_ERR, "Getting default FID failed");
	  retval = PAM_AUTH_ERR;
	  goto error;
	}
  	default_vf = WEXITSTATUS(retval);
  	snprintf(cmd_buf, sizeof (cmd_buf),
      "ROLE_ID=root LOGIN_ID=root CURRENT_VF=%d /fabos/link_bin/switchname", default_vf);
  	pfile = popen(cmd_buf, "r");
  } else {
  	pfile = popen("ROLE_ID=root LOGIN_ID=root CURRENT_AD=0 /fabos/link_bin/switchname", "r");
  }

  if (pfile == NULL) {
    PAM_LOG(LOG_WARNING, "popen failed %d (%s)", errno,strerror(errno));
    retval = PAM_AUTH_ERR;
  } else {
    fscanf(pfile,"%s",host_name);
    retval = pclose(pfile);
  }

  /*
   * If SIGCHLD is set to SIG_IGN, pclose will
   * always return -1 and errno will be set to
   * ECHILD. So to ignore this case, if errno is
   * ECHILD always set retval to 0
   */
  if ((retval == -1) && (errno == ECHILD))
    retval = PAM_SUCCESS;
  PAM_FAIL_CHECK;
  /* Set NAS-Identifier with hostname */
  config.client_id = host_name;

  /* Set remote ID as Calling-Station-ID */
  config.remote_id = pam_getenv(pamh, "REMOTEHOST");

  /* build and initialize the RADIUS packet */
  request->code = PW_AUTHENTICATION_REQUEST;
  get_random_vector(request->vector);
  request->id = request->vector[0]; /* this should be evenly distributed */

  /* grab the password (if any) from the previous authentication layer */
  retval = pam_get_item(pamh, PAM_AUTHTOK, (CONST void **) &password);
  PAM_FAIL_CHECK;

  if(password) {
    password = strdup(password);
    DPRINT(LOG_DEBUG, "Got password %s", password);
  }

  /* no previous password: maybe get one from the user */
  if (!password) {
    if (ctrl & PAM_USE_FIRST_PASS) {
      retval = PAM_AUTH_ERR;	/* use one pass only, stopping if it fails */
      goto error;
    }

    /* check to see if we send a NULL password the first time around */
    if (!(ctrl & PAM_SKIP_PASSWD)) {
      retval = rad_converse(pamh, PAM_PROMPT_ECHO_OFF, "Password: ", &password);
      PAM_FAIL_CHECK;

    }
  } /* end of password == NULL */
  build_radius_packet(request, user, password, &config);

  DPRINT(LOG_DEBUG, "Sending RADIUS request code %d", request->code);
  pamhandle = pamh;
  retval = talk_radius(&config, request, response, password, NULL, tries);
	PAM_FAIL_CHECK;

  DPRINT(LOG_DEBUG, "Got RADIUS response code %d", response->code);

  /*
   *  If we get an authentication failure, and we sent a NULL password,
   *  ask the user for one and continue.
   *
   *  If we get an access challenge, then do a response, for as many
   *  challenges as we receive.
   */
  while (response->code == PW_ACCESS_CHALLENGE) {
#if 0
    attribute_t *a_state, *a_reply;
    char challenge[BUFFER_SIZE];

    /* Now we do a bit more work: challenge the user, and get a response */
    if (((a_state = find_attribute(response, PW_STATE)) == NULL) ||
	((a_reply = find_attribute(response, PW_REPLY_MESSAGE)) == NULL)) {
      _pam_log(LOG_ERR, "RADIUS access-challenge received with State or Reply-Message missing");
      retval = PAM_AUTHINFO_UNAVAIL;
      goto error;
    }

    memcpy(challenge, a_reply->data, a_reply->length - 2);
    challenge[a_reply->length - 2] = 0;

    /* It's full challenge-response, we might as well have echo on */
    retval = rad_converse(pamh, PAM_PROMPT_ECHO_OFF, challenge, &resp2challenge);

    /* now that we've got a response, build a new radius packet */
    build_radius_packet(request, user, resp2challenge, &config);
    /* request->code is already PW_AUTHENTICATION_REQUEST */
    request->id++;		/* one up from the request */

    /* copy the state over from the servers response */
    add_attribute(request, PW_STATE, a_state->data, a_state->length - 2);

    retval = talk_radius(&config, request, response, resp2challenge, NULL, 1);
    PAM_FAIL_CHECK;

    DPRINT(LOG_DEBUG, "Got response to challenge code %d", response->code);
#endif
	retval = PAM_AUTH_ERR; 
	goto error;
  }

accept:
  /* Whew! Done the pasword checks, look for an authentication acknowledge */
  if (response->code == PW_AUTHENTICATION_ACK) {
	PAM_LOG(LOG_DEBUG, "pam_sm_authenticate: Getting role of the user \n");
   	retval = role_to_user(pamh, response, user, default_vf);
  } else {
    retval = PAM_AUTH_ERR;	/* authentication failure */

  error:
	PAM_LOG(LOG_DEBUG, "pam_sm_authenticate: %d: returned error %d: \n", __LINE__, retval);
    /* If there was a password pass it to the next layer */
    if (password && *password) {
      pam_set_item(pamh, PAM_AUTHTOK, password);
    }
  }

  if (ctrl & PAM_DEBUG_ARG) {
    _pam_log(LOG_DEBUG, "authentication %s"
	     , retval==PAM_SUCCESS ? "succeeded":"failed" );
  }

  /*
   * Fix for defects 609674 and 585688: Return PAM_AUTHINFO_UNAVAIL when RADIUS
   * Server is not reachable, to log this information correctly in
   * pam_fabos_log_auth module
   */
  if ((retval != PAM_SUCCESS) && (retval != PAM_AUTHINFO_UNAVAIL)) {
      retval = PAM_AUTH_ERR;
  }

  close(config.sockfd);
  cleanup(config.server);
  #ifdef DEBUG_ZEROIZATION
  FILE *fp = fopen("/dev/console", "w+");
  fprintf(fp, "ZeroizationLog: In pam_sm_authenticate(): Before zeroization, password = %s\n", password);
  #endif
  int passlen = 0;
  if (password != NULL) {
     passlen = strlen(password);
    memset( password, 0, passlen);
  }
  #ifdef MEMCMP_ZEROIZATION
/* Code for CC */
  pam_memcmp_zero(password, passlen);
  #endif

  #ifdef DEBUG_ZEROIZATION
  fprintf(fp, "ZeroizationLog: In pam_sm_authenticate(): After afterzeroization, password = %s\n", password);
  fclose(fp);
  #endif
  if (password) {
	  free(password);
	  password = NULL;
  }
  _pam_forget(resp2challenge);
  {
    int *pret = malloc( sizeof(int) );
    *pret = retval;
    pam_set_data( pamh, "rad_setcred_return", (void *) pret, _int_free );
  }
  return retval;
}

/*
 * Return a value matching the return value of pam_sm_authenticate, for
 * greatest compatibility.
 * (Always returning PAM_SUCCESS breaks other authentication modules;
 * always returning PAM_IGNORE breaks PAM when we're the only module.)
 */
PAM_EXTERN int
pam_sm_setcred(pam_handle_t *pamh,int flags,int argc,CONST char **argv)
{
  int retval, *pret;

  retval = PAM_SUCCESS;
  pret = &retval;
  pam_get_data( pamh, "rad_setcred_return", (CONST void **) &pret );
  return *pret;
}

#undef PAM_FAIL_CHECK
#define PAM_FAIL_CHECK if (retval != PAM_SUCCESS) { return PAM_SESSION_ERR; }

static int
pam_private_session(pam_handle_t *pamh, int flags,
		    int argc, CONST char **argv,
		    int status)
{
  CONST char *user = NULL;
  int ctrl;
  int retval = PAM_AUTH_ERR;

  char recv_buffer[4096];
  char send_buffer[4096];
  AUTH_HDR *request = (AUTH_HDR *) send_buffer;
  AUTH_HDR *response = (AUTH_HDR *) recv_buffer;
  radius_conf_t config;

  FILE *pfile = NULL;
  char host_name[_SC_HOST_NAME_MAX] = {0};
  const char * vf_enabled_str = NULL, *end_ptr;
  int vf_enabled = 0, default_vf = 0;
  char cmd_buf[256] = {0};

  /* Find whether VF is enabled or not from PAM envlist */
  vf_enabled_str = pam_getenv(pamh, "VF_ENABLED");
  if (vf_enabled_str) {
		vf_enabled = strtol(vf_enabled_str , &end_ptr, 10);
  }
  PAM_LOG(LOG_DEBUG, "VF_Enabled:%d", vf_enabled);

  ctrl = _pam_parse(argc, argv, &config);

  /* grab the saved RADIUS user name */
  retval = pam_get_data(pamh, PAM_RADIUS_UNAME, (CONST void **) &user);
  PAM_FAIL_CHECK;

  /* check that they've entered something, and not too long, either */
  if ((user == NULL) ||
      (strlen(user) > MAXPWNAM)) {
    return PAM_USER_UNKNOWN;
  }

  /*
   * Get the IP address of the authentication server
   * Then, open a socket, and bind it to a port
   */
  retval = initialize(&config, TRUE);
  PAM_FAIL_CHECK;

  /*
   * If there's no client id specified, use the service type, to help
   * keep track of which service is doing the authentication.
   */

  if (!config.client_id) {
    retval = pam_get_item(pamh, PAM_SERVICE, (CONST void **) &config.client_id);
    PAM_FAIL_CHECK;
  }
  /* now we've got a socket open, so we've got to clean it up on error */
#undef PAM_FAIL_CHECK
#define PAM_FAIL_CHECK if (retval != PAM_SUCCESS) {goto error; }
  /*
   * Defect 85640
   * Nas-Identifier field whose value was PAM_SERVICE
   * will now have  value as `switchname'
   */
  if (vf_enabled) {
  	snprintf(cmd_buf, sizeof (cmd_buf),
      "/fabos/libexec/getDefaultFID");
  	retval = system(cmd_buf);
  	if (retval == -1) {
      PAM_LOG(LOG_ERR, "Getting default FID failed");
	  retval = PAM_AUTH_ERR;
	  goto error;
	}
  	default_vf = WEXITSTATUS(retval);
  	snprintf(cmd_buf, sizeof (cmd_buf),
      "ROLE_ID=root LOGIN_ID=root CURRENT_VF=%d /fabos/link_bin/switchname", default_vf);
  	pfile = popen(cmd_buf, "r");
  } else {
  	pfile = popen("ROLE_ID=root LOGIN_ID=root CURRENT_AD=0 /fabos/link_bin/switchname", "r");
  }
  if (pfile == NULL) {
    PAM_LOG(LOG_WARNING, "popen failed %d (%s)", errno,strerror(errno));
    retval = PAM_AUTH_ERR;
  } else {
    fscanf(pfile,"%s",host_name);
    retval = pclose(pfile);
  }

  /*
   * If SIGCHLD is set to SIG_IGN, pclose will
   * always return -1 and errno will be set to
   * ECHILD. So to ignore this case, if errno is
   * ECHILD always set retval to 0
   */
  if ((retval == -1) && (errno == ECHILD))
    retval = PAM_SUCCESS;
  PAM_FAIL_CHECK;
  /* Set NAS-Identifier with hostname */
  config.client_id = host_name;

  /* Set remote ID as Calling-Station-ID */
  config.remote_id = pam_getenv(pamh, "REMOTEHOST");

  /* build and initialize the RADIUS packet */
  request->code = PW_ACCOUNTING_REQUEST;
  get_random_vector(request->vector);
  request->id = request->vector[0]; /* this should be evenly distributed */

  build_radius_packet(request, user, NULL, &config);

  add_int_attribute(request, PW_ACCT_STATUS_TYPE, status);

  sprintf(recv_buffer, "%08d", (int) getpid());
  add_attribute(request, PW_ACCT_SESSION_ID, recv_buffer, strlen(recv_buffer));

  add_int_attribute(request, PW_ACCT_AUTHENTIC, PW_AUTH_RADIUS);

  if (status == PW_STATUS_START) {
    session_time = time(NULL);
  } else {
    add_int_attribute(request, PW_ACCT_SESSION_TIME, time(NULL) - session_time);
  }

  retval = talk_radius(&config, request, response, NULL, NULL, 1);
  PAM_FAIL_CHECK;

  /* oops! They don't have the right password.  Complain and die. */
  if (response->code != PW_ACCOUNTING_RESPONSE) {
    retval = PAM_PERM_DENIED;
    goto error;
  }

  retval = PAM_SUCCESS;

error:

  close(config.sockfd);
  cleanup(config.server);

  return retval;
}

PAM_EXTERN int
pam_sm_open_session(pam_handle_t *pamh, int flags,
		    int argc, CONST char **argv)
{
  return pam_private_session(pamh, flags, argc, argv, PW_STATUS_START);
}

PAM_EXTERN int
pam_sm_close_session(pam_handle_t *pamh, int flags,
		     int argc, CONST char **argv)
{
  return pam_private_session(pamh, flags, argc, argv, PW_STATUS_STOP);
}

#undef PAM_FAIL_CHECK
#define PAM_FAIL_CHECK if (retval != PAM_SUCCESS) {_pam_log(LOG_ERR, "pam auth failed: %d.", retval); return PAM_AUTH_ERR; }
#define MAX_PASSWD_TRIES 3

PAM_EXTERN int
pam_sm_chauthtok(pam_handle_t *pamh, int flags, int argc, CONST char **argv)
{
  CONST char *user;
  char *password = NULL;
  char *new_password = NULL;
  char *check_password = NULL;
  int ctrl;
  int retval = PAM_AUTHTOK_ERR;
  int attempts;

  char recv_buffer[4096];
  char send_buffer[4096];
  AUTH_HDR *request = (AUTH_HDR *) send_buffer;
  AUTH_HDR *response = (AUTH_HDR *) recv_buffer;
  radius_conf_t config;

  ctrl = _pam_parse(argc, argv, &config);

  /* grab the user name */
  retval = pam_get_user(pamh, &user, NULL);
  PAM_FAIL_CHECK;

  /* check that they've entered something, and not too long, either */
  if ((user == NULL) ||
      (strlen(user) > MAXPWNAM)) {
    return PAM_USER_UNKNOWN;
  }

  /*
   * Get the IP address of the authentication server
   * Then, open a socket, and bind it to a port
   */
  retval = initialize(&config, FALSE);
  PAM_FAIL_CHECK;

  /*
   * If there's no client id specified, use the service type, to help
   * keep track of which service is doing the authentication.
   */
  if (!config.client_id) {
    retval = pam_get_item(pamh, PAM_SERVICE, (CONST void **) &config.client_id);
    PAM_FAIL_CHECK;
  }

  /* now we've got a socket open, so we've got to clean it up on error */
#undef PAM_FAIL_CHECK
#define PAM_FAIL_CHECK if (retval != PAM_SUCCESS) {goto error; }

  /* grab the old password (if any) from the previous password layer */
  retval = pam_get_item(pamh, PAM_OLDAUTHTOK, (CONST void **) &password);
  PAM_FAIL_CHECK;
  if(password) password = strdup(password);

  /* grab the new password (if any) from the previous password layer */
  retval = pam_get_item(pamh, PAM_AUTHTOK, (CONST void **) &new_password);
  PAM_FAIL_CHECK;
  if(new_password) new_password = strdup(new_password);

  /* preliminary password change checks. */
  if (flags & PAM_PRELIM_CHECK) {
    if (!password) {		/* no previous password: ask for one */
      retval = rad_converse(pamh, PAM_PROMPT_ECHO_OFF, "Password: ", &password);
      PAM_FAIL_CHECK;
    }

    /*
     * We now check the password to see if it's the right one.
     * If it isn't, we let the user try again.
     * Note that RADIUS doesn't have any concept of 'root'.  The only way
     * that root can change someone's password is to log into the RADIUS
     * server, and and change it there.
     */

    /* build and initialize the access request RADIUS packet */
    request->code = PW_AUTHENTICATION_REQUEST;
    get_random_vector(request->vector);
    request->id = request->vector[0]; /* this should be evenly distributed */

    build_radius_packet(request, user, password, &config);
    add_int_attribute(request, PW_USER_SERVICE_TYPE, PW_AUTHENTICATE_ONLY);

    retval = talk_radius(&config, request, response, password, NULL, 1);
    PAM_FAIL_CHECK;

    /* oops! They don't have the right password.  Complain and die. */
    if (response->code != PW_AUTHENTICATION_ACK) {
      _pam_forget(password);
      retval = PAM_PERM_DENIED;
      goto error;
    }

    /*
     * We're now sure it's the right user.
     * Ask for their new password, if appropriate
     */

    if (!new_password) {	/* not found yet: ask for it */
      int new_attempts;
      attempts = 0;

      /* loop, trying to get matching new passwords */
      while (attempts++ < 3) {

	/* loop, trying to get a new password */
	new_attempts = 0;
	while (new_attempts++ < 3) {
	  retval = rad_converse(pamh, PAM_PROMPT_ECHO_OFF,
				"New password: ", &new_password);
	  PAM_FAIL_CHECK;

	  /* the old password may be short.  Check it, first. */
	  if (strcmp(password, new_password) == 0) { /* are they the same? */
	    rad_converse(pamh, PAM_ERROR_MSG,
			 "You must choose a new password.", NULL);
	    _pam_forget(new_password);
	    continue;
	  } else if (strlen(new_password) < 6) {
	    rad_converse(pamh, PAM_ERROR_MSG, "it's WAY too short", NULL);
	    _pam_forget(new_password);
	    continue;
	  }

	  /* insert crypt password checking here */

	  break;		/* the new password is OK */
	}

	if (new_attempts >= 3) { /* too many new password attempts: die */
	  retval = PAM_AUTHTOK_ERR;
	  goto error;
	}

	/* make sure of the password by asking for verification */
	retval =  rad_converse(pamh, PAM_PROMPT_ECHO_OFF,
			       "New password (again): ", &check_password);
	PAM_FAIL_CHECK;

	retval = strcmp(new_password, check_password);
	_pam_forget(check_password);

	/* if they don't match, don't pass them to the next module */
	if (retval != 0) {
	  _pam_forget(new_password);
	  rad_converse(pamh, PAM_ERROR_MSG,
		       "You must enter the same password twice.", NULL);
	  retval = PAM_AUTHTOK_ERR;
	  goto error;		/* ??? maybe this should be a 'continue' ??? */
	}

	break;			/* everything's fine */
      }	/* loop, trying to get matching new passwords */

      if (attempts >= 3) { /* too many new password attempts: die */
	retval = PAM_AUTHTOK_ERR;
	goto error;
      }
    } /* now we have a new password which passes all of our tests */

    /*
     * Solaris 2.6 calls pam_sm_chauthtok only ONCE, with PAM_PRELIM_CHECK
     * set.
     */
#ifndef sun
    /* If told to update the authentication token, do so. */
  } else if (flags & PAM_UPDATE_AUTHTOK) {
#endif

    if (!password || !new_password) { /* ensure we've got passwords */
      retval = PAM_AUTHTOK_ERR;
      goto error;
    }

    /* build and initialize the password change request RADIUS packet */
    request->code = PW_PASSWORD_REQUEST;
    get_random_vector(request->vector);
    request->id = request->vector[0]; /* this should be evenly distributed */

    /* the secret here can not be know to the user, so it's the new password */
    _pam_forget(config.server->secret);
    config.server->secret = strdup(password); /* it's free'd later */

    build_radius_packet(request, user, new_password, &config);
    add_password(request, PW_OLD_PASSWORD, password, password);

    retval = talk_radius(&config, request, response, new_password, password, 1);
    PAM_FAIL_CHECK;

    /* Whew! Done password changing, check for password acknowledge */
    if (response->code != PW_PASSWORD_ACK) {
      retval = PAM_AUTHTOK_ERR;
      goto error;
    }
  }

  /*
   * Send the passwords to the next stage if preliminary checks fail,
   * or if the password change request fails.
   */
  if ((flags & PAM_PRELIM_CHECK) || (retval != PAM_SUCCESS)) {
  error:

    /* If there was a password pass it to the next layer */
    if (password && *password) {
      pam_set_item(pamh, PAM_OLDAUTHTOK, password);
    }

    if (new_password && *new_password) {
      pam_set_item(pamh, PAM_AUTHTOK, new_password);
    }
  }

  if (ctrl & PAM_DEBUG_ARG) {
    _pam_log(LOG_DEBUG, "password change %s"
	     , retval==PAM_SUCCESS ? "succeeded":"failed" );
  }

  close(config.sockfd);
  cleanup(config.server);

  _pam_forget(password);
  _pam_forget(new_password);
  return retval;
}

#ifdef PAM_STATIC

/* static module data */

struct pam_module _pam_radius_modstruct = {
  "pam_radius_auth",
  pam_sm_authenticate,
  pam_sm_setcred,
  NULL,
  pam_sm_open_session,
  pam_sm_close_session,
  pam_sm_chauthtok,
};
#endif

/*
 *  This module is a merger of an old version of pam_radius.c, and
 * code which went into mod_auth_radius.c, with further modifications
 * by Alan DeKok of CRYPTOCard Inc..
 *
 * The original pam_radius.c code is copyright (c) Cristian Gafton, 1996,
 *                                             <gafton@redhat.com>
 *
 * The additional code is copyright (c) CRYPTOCard Inc, 1998.
 *
 *                                              All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, and the entire permission notice in its entirety,
 *    including the disclaimer of warranties.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote
 *    products derived from this software without specific prior
 *    written permission.
 *
 * ALTERNATIVELY, this product may be distributed under the terms of
 * the GNU Public License, in which case the provisions of the GPL are
 * required INSTEAD OF the above restrictions.  (This clause is
 * necessary due to a potential bad interaction between the GPL and
 * the restrictions contained in a BSD-style copyright.)
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */
