/*
 *    Copyright (c) 1996-2007 Brocade Communications Systems, Inc.
 *    All rights reserved.
 *
 *    Module name: radpeapclient.c
 *
 *    Description:
 *      This modules does with PEAP-MSCHAPv2 user authentication
 *
 */

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <openssl/x509v3.h>
#include "eap_tls.h"
#include "conf.h"
#include "radpaths.h"
#include "missing.h"
#include "md5.h"
#include "eap_types.h"
#include "eap_sim.h"
#include "radpeapclient.h"
#include "mschapv2.h"
#include "pam_radius_auth.h"

static unsigned char cached_auth_challenge[16] = { 0 };
static unsigned char cached_peer_challenge[16] = { 0 };
static unsigned char cached_nt_response[24] = { 0 };
static peap_instance_t peap_instance;
static const char * xlookuptbl = {"0123456789ABCDEF"};
const char * radius_dir = RADDBDIR;
int debug_flag = 0;

#define SW_CERT_FILE "/etc/fabos/certs/radius/swRadcert.pem"
#define PVT_KEY "/etc/fabos/certs/radius/pvt_key"
#define RADIUS_CA_FILE "/etc/fabos/certs/radius/swRadca.pem"

/*
 * uint8_t *compose_nak_response 
 *
 * Description
 *	Function  builds nak response packet for unknow request
 *	send by the server based on the state machine's  state
 *
 * Input
 *	pointer for NAK data length	
 *
 * Output
 *	Composed NAK response packets's length
 *
 * Return
 *	Returns the inner tlv NAK data
 */
static uint8_t *
compose_nak_respone(int *nak_data_len)
{
	uint8_t * ptr = 0;
	VALUE_PAIR * first_vp = 0;
	RADIUS_PACKET * rep = 0;

	/* Get the latest server reply */
	rep = peap_instance.server_rep;
	if (0 == rep) {
		peap_log_debug("compose_mschap_response : server rep is NULL!!!");
		return 0;
	}
	/* Extract the EAP Message from the AVP's */
	first_vp = pairfind(rep->vps, PW_EAP_MESSAGE);
	if (0 == first_vp) {
		peap_log_debug("compose_mschap_response : %d: first_vp is NULL \n",
						__LINE__);
		return 0;
	}

	*nak_data_len = 1 /* PW_EAP_NAK */  + 1 /* PW_EAP_MSCHAPV2 */;

	ptr = (uint8_t *)calloc(*nak_data_len, sizeof(uint8_t));
	if (0 == ptr) {
		peap_log_debug("compose_nak_response: Memory allocation failed ");
		return 0;
	}
	*ptr = PW_EAP_NAK;					/* EAP-Type */
	*(ptr + 1) = PW_EAP_MSCHAPV2; 	/* Authentication Type */

	return ptr; 	/* Return from here itself, don't fall through */
}

/*
 * int radlog
 *
 * Description
 *			Function used for logging by freeradius
 *
 * Input
 *   		Var Args
 *
 * Output
 *        	None
 *
 * Return
 *      	None
 */
int
radlog(int lvl, const char *msg, ...)
{
	va_list ap;
	int r;
	r = lvl;			/* shut up compiler */
	va_start(ap, msg);
	r = vfprintf(stderr, msg, ap);
	va_end(ap);
	fputc('\n', stderr);
	return r;
}

/*
 * int peap_log_debug
 *
 * Description
 *			Function used for logging
 *
 * Input
 *   		Var Args
 *
 * Output
 *        	None
 *
 * Return
 *      	None
 */
int
peap_log_debug(const char * msg, ...)
{
	FILE * fp;
	va_list ap;
	int r;
#define FILENAME "/var/log/peaplog"

	fp = fopen(FILENAME, "a");
	if(0 == fp) {
		return -1;
	}
	va_start(ap, msg);
	r = vfprintf(fp, msg, ap);
	va_end(ap);
	fputc('\n', fp);
	fclose(fp);

	return r;
}

/*
 * static void encode_asc
 *
 * Description
 *    	Convert a hexadecimal value into corresponding ascii value
 *
 * Input
 *     	unsigned int num
 *
 * Output
 *     char * encoded
 *
 * Return
 *     None
 */
static void
encode_asc(uint8_t num, char * encoded)
{
	encoded[0] = xlookuptbl[(num & 0xF0) >> 4];
	encoded[1] = xlookuptbl[(num & 0x0F)];
}

/*
 * int compose_eap_identity_pkt
 *
 * Description
 * 	Compose the first radius packet from the client
 *
 * Input
 * 	RADIUS_PACKET * req
 *
 * Output
 *	RADIUS_PACKET * req
 *
 * Return
 *	Success or failure
 */
static int
compose_eap_identity_pkt(RADIUS_PACKET * req)
{
	char mesg_auth[AUTH_VECTOR_LEN] = { 0 };
	char eap_identity_str[MAX_NT_USERNAME] = { 0 };
	uint16_t eap_pkt_size = 0;
	uint16_t eap_msg_size = 0;
	uint16_t rad_pkt_size = 0;
	uint16_t tv_size = 0;
	uint16_t temp_len = 0 ;
	uint32_t nas_ip_addr = 0;
	int user_tlv_size = 0;
	int nas_tlv_size = 0;
	int port_tlv_size = 0;
	uint8_t * ptr = 0;
	eap_message_t * eap_msg = 0;
	eap_pkt_t * eap_pkt = 0;
	radius_packet_t * rad_pkt = 0;
	type_length_value_t * tv = 0;
	type_length_value_t * user_tlv = 0;
	type_length_value_t * nas_tlv = 0;
	type_length_value_t * port_tlv = 0;

#undef CLEAN_UP
#define CLEAN_UP()		  				\
	do {							\
		if(eap_pkt) { free(eap_pkt); eap_pkt = 0; }	\
		if(eap_msg) { free(eap_msg); eap_msg = 0; } 	\
		if(tv) { free(tv); tv = 0; }			\
		if(user_tlv) { free(user_tlv); user_tlv = 0; }	\
		if(nas_tlv) { free(nas_tlv); nas_tlv = 0; }	\
		if(port_tlv) { free(port_tlv); port_tlv = 0; }	\
	} while (0);

	strncpy(eap_identity_str, peap_instance.user_name , strlen(peap_instance.user_name));

	eap_pkt_size = sizeof(eap_pkt_t) + strlen(eap_identity_str);
	eap_pkt = calloc(eap_pkt_size, sizeof(uint8_t));
	if (0 == eap_pkt) {
		peap_log_debug("compose_eap_identity_pkt: Memory allocation failed,eap_pkt is NULL\n ");
		return -1;
	}
	eap_pkt->code = PW_EAP_RESPONSE;
	eap_pkt->id = EAP_ID; 		/* Some Random ID it is */

	temp_len = htons(eap_pkt_size);
	eap_pkt->length = temp_len;
	eap_pkt->type = PW_EAP_IDENTITY;
	memcpy(eap_pkt->type_data, eap_identity_str, strlen(eap_identity_str));

	eap_msg_size = sizeof(eap_message_t) + eap_pkt_size;

	/*
	 * Allocate memory for EAP Message Radius Attribute
	 */
	eap_msg = calloc(eap_msg_size, sizeof(uint8_t));
	if(0 == eap_msg_size) {
		peap_log_debug("compose_eap_identity_pkt:Memory allocation failed,eap_msg is NULL\n");
		CLEAN_UP();
		return -1;
	}
	eap_msg->code = PW_EAP_MESSAGE;
	memcpy(eap_msg->data, (eap_packet_t *) eap_pkt, eap_pkt_size);
	eap_msg->length = eap_msg_size;
	tv_size = sizeof(type_length_value_t) + AUTH_VECTOR_LEN;
	tv = calloc(tv_size, sizeof(uint8_t));
	if(0 == tv) {
		peap_log_debug("compose_eap_identity_pkt:Memory allocation failed,eap_msg is NULL\n");
		CLEAN_UP();
		return -1;
	}
	tv->type = PW_MESSAGE_AUTHENTICATOR;
	tv->length = tv_size;
	/*
	 * User TLV ,Contains the user credentials initailly only user_name
	 */
	user_tlv_size =	sizeof(type_length_value_t) + strlen(eap_identity_str);
	user_tlv = calloc(user_tlv_size, sizeof(uint8_t));
	if(0 == user_tlv) {
		peap_log_debug("compose_eap_identity_pkt:Memory allocation failed,eap_msg is NULL\n");
		CLEAN_UP();
		return -1;
	}
	user_tlv->type = PW_USER_NAME;
	user_tlv->length = user_tlv_size;
	memcpy(user_tlv->data, eap_identity_str, strlen(eap_identity_str));
	/*
	 * NAS IP address, Remote Access policy dictates this to be present
	 */
	nas_tlv_size = sizeof(type_length_value_t) + IP_ADDR_SIZE; /* 4 bytes of IPV4 address */
	nas_tlv = calloc(nas_tlv_size, sizeof(uint8_t));
	if(0 == nas_tlv) {
		peap_log_debug("compose_eap_identity_pkt:Memory allocation failed,eap_msg is NULL\n");
		CLEAN_UP();
		return -1;
	}
	nas_tlv->type = PW_NAS_IP_ADDRESS;
	nas_tlv->length = nas_tlv_size;
	nas_ip_addr = peap_instance.dst_ipaddr;

	memcpy(nas_tlv->data, (uint32_t *) & nas_ip_addr, sizeof(uint32_t));
	/*
	 * NAS port tlv, value can be zero also
	 */
	port_tlv_size = sizeof(type_length_value_t) + sizeof(uint32_t);
	port_tlv = calloc(port_tlv_size, sizeof(uint8_t));
	if(0 == port_tlv) {
		peap_log_debug("compose_eap_identity_pkt:Memory allocation failed,eap_msg is NULL\n");
		CLEAN_UP();
		return -1;
	}
	port_tlv->type = PW_NAS_PORT;
	port_tlv->length = port_tlv_size;
	memset(port_tlv->data, 0, sizeof(uint32_t));

	rad_pkt_size = sizeof(radius_packet_t) + eap_msg_size + tv_size +
			user_tlv_size + nas_tlv_size + port_tlv_size;

	rad_pkt = calloc(rad_pkt_size, sizeof(uint8_t));
	if(0 == rad_pkt) {
		peap_log_debug("compose_eap_identity_pkt:Memory allocation failed,eap_msg is NULL\n");
		CLEAN_UP();
		return -1;
	}
	rad_pkt->code = RADIUS_ACCESS_REQUEST;	/* Radius Access-Request */
	rad_pkt->id = RANDOM_ID;		/* Random ID to start with */
	memset(rad_pkt->vector, 0, AUTH_VECTOR_LEN);

	ptr = (uint8_t *)rad_pkt->data;	/*
					 * Now we are going to pack all these tlv's into a packet
					 * We gonna play with these temporarry variable `ptr'
					 */
	memcpy(ptr, user_tlv, user_tlv_size);
	ptr += user_tlv_size; 		/*
					 * These tlv's can be any order,
					 * let user_tlv get packed first
					 */
	memcpy(ptr, nas_tlv, nas_tlv_size);
	ptr += nas_tlv_size; 	/* pack NAS port tlv, advance the pointer accordingly */

	memcpy(ptr, port_tlv, port_tlv_size);
	ptr += port_tlv_size;	/* PORT tlv , does not seems to be that much useful */

	memcpy(ptr, eap_msg, eap_msg_size);	/* Here comes the real EAP Message */

	ptr += eap_msg_size;
	memcpy(ptr, tv, tv_size);
						/* Protect the entire radiius packet with
						 * the message authenticator
						 */

	temp_len = htons(rad_pkt_size);
	rad_pkt->length = temp_len;

	/* Calculate the radius paket vector */
	librad_md5_calc(rad_pkt->vector, rad_pkt->vector, sizeof(rad_pkt->vector));

	lrad_hmac_md5((uint8_t *) rad_pkt, rad_pkt_size, peap_instance.shared_secret,
			strlen(peap_instance.shared_secret), mesg_auth);

	ptr += 2 ;	/* Skip the length field */
	memcpy(ptr, mesg_auth, AUTH_VECTOR_LEN);
	req->data = (uint8_t *)rad_pkt;
	req->data_len = rad_pkt_size;
	CLEAN_UP();
	return 0;
}

/*
 * static void record_peap_init
 *
 * Description
 *	TLS session initialisation function
 *
 * Input
 *	record_t * rec
 *
 * Output
 *	None
 *
 * Return
 *	None
 */
static void
record_peap_init(record_t * rec)
{
	rec->used = 0;
}

/*
 * static void record_peap_close
 *
 * Description
 *	TLS session Termination function
 *
 * Input
 *	record_t * rec
 *
 * Output
 *	None
 *
 * Return
 *	None
 */
static void
record_peap_close(record_t * rec)
{
	rec->used = 0;
}

/*
 * static unsigned int record_peap_plus
 *
 * Description
 *	Copy data to the intermediate buffer, before we send it somewhere
 *
 * Input
 *	record_t * rec, const void * ptr, unsigned int size
 *
 * Output
 *	None
 *
 * Return
 *	Total no of bytes added to the buffer
 */
static unsigned int
record_peap_plus(record_t * rec, const void * ptr, unsigned int size)
{
	unsigned int added = MAX_RECORD_SIZE - rec->used;
	if (added > size)
		added = size;
	if (added == 0)
		return 0;
	memcpy(rec->data + rec->used, ptr, added);
	rec->used += added;
	return added;
}

/*
 * static unsigned int record_peap_minus
 *
 * Description
 *	Take data from the buffer, and give it to the caller
 *
 * Input
 *	record_t * rec
 *
 * Output
 *	None
 *
 * Return
 *	Total no of bytes taken out from the buffer
 */
static unsigned int
record_peap_minus(record_t * rec, void * ptr, unsigned int size)
{
    unsigned int taken = rec->used;
	if (taken > size)
		taken = size;
	if (taken == 0)
		return 0;
	if (ptr)
		memcpy(ptr, rec->data, taken);
	rec->used -= taken;
	/*
	 *      This is pretty bad...
	 */
	if (rec->used > 0)
		memmove(rec->data, rec->data + taken, rec->used);
	return taken;
}

/*
 * static void cb_peap_info
 *
 * Description
 *	Info call back function for SLL functions
 *
 * Input
 *	const SSL * s, int where
 *
 * Output
 *	None
 *
 * Return
 *	None
 */
static void
cb_peap_info(const SSL * s, int where, int unused)
{
	const char * str, * state;
	int w;
	w = where & ~SSL_ST_MASK;
	if (w & SSL_ST_CONNECT)
		str = "TLS_connect";
	else if (w & SSL_ST_ACCEPT)
		str = "TLS_accept";
	else
		str = "(other)";

	state = SSL_state_string_long(s);
	state = state ? state : "NULL";

	return;
}

/*
 * void cb_peap_msg
 *
 * Description
 *	Message call back  function for SSL.
 *	Function get invoked from SSL layer , when a SSL handshake
 *	happens ,so that we can know what messages are exchaned
 *
 * Input
 *	int write_p, int msg_version, int content_type,
 *	const void * buf, size_t len, void * arg
 *
 * Output
 *	None
 *
 * Return
 *	None
 */
static void
cb_peap_msg(int write_p, int msg_version, int content_type,
	    const void * buf, size_t len, SSL * ssl UNUSED, void * arg)
{
	tls_session_t * state = (tls_session_t *) arg;

	/*
	 *      Work around bug #298, where we may be called with a NULL
	 *      argument.  We should really log a serious error
	 */
	if (!arg)
		return;

	state->info.origin = (unsigned char) write_p;
	state->info.content_type = (unsigned char) content_type;
	state->info.record_len = len;
	state->info.version = msg_version;
	state->info.initialized = 1;

	if (content_type == SSL3_RT_ALERT) {
		state->info.alert_level = ((const unsigned char *) buf)[0];
		state->info.alert_description = ((const unsigned char *) buf)[1];
		state->info.handshake_type = 0x00;

	} else if (content_type == SSL3_RT_HANDSHAKE) {
		state->info.handshake_type = ((const unsigned char *) buf)[0];
		state->info.alert_level = 0x00;
		state->info.alert_description = 0x00;
	}
	tls_session_information(state);
}

/*
 *RADIUS_PACKET * compose_eap_packet
 *
 * Description
 *	This function composes the radius packets on-wire which contains
 *	TLVs like eap message , nas ip address , nas port ,
 *	state , message authenticator etc .
 *
 * Input
 *	RADIUS_PACKET * req , char *data,
 *	int len , uint8_t include_length .
 *
 * Output
 *	Radius packet on-wire
 *
 * Return
 *	RADIUS_PACKET * req
 */
static RADIUS_PACKET  *
compose_eap_packet(RADIUS_PACKET * req, char * data, int len, uint8_t include_length)
{
	char eap_identity_str[MAX_NT_USERNAME] = { 0 };
	char mesg_auth[AUTH_VECTOR_LEN] = { 0 };
	uint16_t eap_pkt_size = 0;
	uint16_t ep_pkt_size = 0;
	uint16_t eap_msg_size[MAX_EAP_PACKETS] = { 0 };
	uint16_t rad_pkt_size = 0;
	uint16_t tls_msg_length = 0;
	uint16_t tv_size = 0;
	uint16_t tls_record_length = 0;
	uint16_t state_tv_size = 0;
	uint16_t temp_len = 0 ;
	uint32_t temp_len_l = 0 ;
	uint32_t nas_ip_addr = 0;
	int user_tlv_size = 0;
	int nas_tlv_size = 0;
	int port_tlv_size = 0;
	uint8_t * srv_eap_message = 0;
	uint8_t * ptr = 0;
	tls_record_layer_t * tls_record = 0;
	type_length_value_t * tv = 0;
	type_length_value_t * state_tv = 0;
	type_length_value_t * user_tlv = 0;
	type_length_value_t * nas_tlv = 0;
	type_length_value_t * port_tlv = 0;
	VALUE_PAIR * eap_state = 0;
	RADIUS_PACKET * rep = 0;
	radius_packet_t * rad_pkt = 0;
	eap_tls_response_t * eap_pkt = 0;
	eap_tls_response_t * eap_pktptr = 0;
	eap_message_t * eap_msg_arr[MAX_EAP_PACKETS] = { 0 };
	int no_eap_message = 0;
	int tmpcount = 0;

#undef CLEAN_UP
#define CLEAN_UP()  							\
	do {								\
		if(user_tlv) { free(user_tlv); user_tlv = 0; }		\
		if(nas_tlv) { free(nas_tlv);nas_tlv = 0; }  		\
		if(port_tlv) { free(port_tlv); port_tlv = 0; }		\
		if(tls_record) { free(tls_record); tls_record = 0; }	\
		if(eap_pkt) { free(eap_pkt); eap_pkt = 0; } 		\
		if(state_tv) { free(state_tv); state_tv = 0; } 		\
		if(tv) { free(tv); tv = 0; }  				\
                for(tmpcount = 0; tmpcount < MAX_EAP_PACKETS; tmpcount++) { \
                        if(eap_msg_arr[tmpcount]) { \
                                free(eap_msg_arr[tmpcount]); \
                        } \
                } \
	} while (0);

	strncpy(eap_identity_str, peap_instance.user_name, strlen(peap_instance.user_name));

	tls_record_length = len;
	if (0 != len) {
		eap_pkt_size = sizeof(eap_tls_response_t) + len;
		if (FALSE == include_length) {	/* Check Length Included flag is set or not */
			eap_pkt_size -= sizeof(uint32_t);	/* Jump by tls_msg_length size */
		}
	} else {
		eap_pkt_size = sizeof(eap_pkt_t ) + sizeof(uint8_t);
	}

	tv_size = sizeof(type_length_value_t) + AUTH_VECTOR_LEN;
	tls_msg_length = sizeof(eap_tls_response_t)  + len;

	/*
	 * User TLV, in each packet it all these tlv's are needed .
	 */
	user_tlv_size =	sizeof(type_length_value_t) + strlen(eap_identity_str);
	user_tlv = calloc(user_tlv_size, sizeof(uint8_t));
	if(0 == user_tlv) {
		peap_log_debug("compose_eap_packet:Memory alocation failed, user_tlv is NULL \n");
		return 0;
	}
	user_tlv->type = PW_USER_NAME;	/* Pack the user name as a tlv */
	user_tlv->length = user_tlv_size;
	memcpy(user_tlv->data, eap_identity_str, strlen(eap_identity_str));
	/*
	 * NAS IP address, it is need in all packets as
	 * W2K3 server Remote Access policy is based on this
	 */
	nas_tlv_size = sizeof(type_length_value_t) + IP_ADDR_SIZE;	/* 4 bytes of IPV4 address */
	nas_tlv = calloc(nas_tlv_size, sizeof(uint8_t));
	if(0 == nas_tlv) {
		peap_log_debug("compose_eap_packet:Memory alocation failed, nas_tlv is NULL \n");
		CLEAN_UP();
		return 0;
	}
	nas_tlv->type = PW_NAS_IP_ADDRESS;
	nas_tlv->length = nas_tlv_size;

	nas_ip_addr = peap_instance.dst_ipaddr;

	memcpy(nas_tlv->data, (uint32_t *) & nas_ip_addr, IP_ADDR_SIZE);

	/*
	 * Port TLV
	 */
	port_tlv_size = sizeof(type_length_value_t) + sizeof(uint32_t);
	port_tlv = calloc(port_tlv_size, sizeof(uint8_t));
	if(0 == port_tlv) {
		peap_log_debug("compose_eap_packet:Memory alocation failed, port_tlv is NULL \n");
		CLEAN_UP();
		return 0;
	}
	port_tlv->type = PW_NAS_PORT;
	port_tlv->length = port_tlv_size;
	memset(port_tlv->data, 0, sizeof(uint32_t));

	tls_record = calloc(tls_record_length, sizeof(uint8_t));
	if(0 == tls_record) {
		peap_log_debug("compose_eap_packet:Memory alocation failed, tls_record is NULL \n");
		CLEAN_UP();
		return 0;
	}
	memcpy(tls_record, data, tls_record_length);
	/*
	 * Extract PW_EAP_MESSAGE
	 */
	rep = peap_instance.server_rep;

	srv_eap_message = peap_instance.srv_eap_message;
	if (0 == srv_eap_message) {
//		peap_log_debug("compose_eap_packet: %d: srv_eap_message is NULL \n", __LINE__);
		return 0;
	}
	eap_pkt = calloc(eap_pkt_size, sizeof(uint8_t));
	if(0 == eap_pkt) {
		peap_log_debug("compose_eap_packet:Memory alocation failed, eap_pkt is NULL \n");
		CLEAN_UP();
		return 0;
	}
	eap_pkt->code = PW_EAP_RESPONSE;
	eap_pkt->id = (uint8_t)*(srv_eap_message + 1);
	temp_len = htons(eap_pkt_size);
	eap_pkt->length = temp_len;
	eap_pkt->type = PW_EAP_PEAP;

	if (0 != len) {
		if (TRUE == include_length) {
			eap_pkt->flags = EAP_LENGTH_INCLUDED;
			// Length should be EAP message's length. Full EAP packet length should not be added
			temp_len_l = htonl(len);
			eap_pkt->tls_msg_len = temp_len_l;
			memcpy(eap_pkt->data, tls_record, tls_record_length);
		} else {
			eap_pkt->flags = 0x0;
			memcpy(&(eap_pkt->tls_msg_len), tls_record, tls_record_length);
		}
	} else {
		eap_pkt->flags = 0x0;
		eap_pkt_size = TLS_MSG_HEADER_LEN;
	}

	/*
	 * Maximum EAP message data can be only 253. For TLS data above 253
	 * data has to be fragmented and sent with multiple EAP-Message att
	 */

	eap_pktptr = eap_pkt;
	do {
		if (eap_pkt_size > EAP_MESSAGE_MAX){
			ep_pkt_size = EAP_MESSAGE_MAX;
		} else {
			ep_pkt_size = eap_pkt_size;
		}
		eap_msg_size[no_eap_message] = sizeof(eap_message_t) + ep_pkt_size;
		eap_msg_arr[no_eap_message] = calloc(eap_msg_size[no_eap_message], sizeof(uint8_t));
		if(0 == eap_msg_arr[no_eap_message]) {
			peap_log_debug("compose_eap_packet:Memory alocation failed, eap_msg is NULL \n");
			CLEAN_UP();
			return 0;
		}
		eap_msg_arr[no_eap_message]->code = PW_EAP_MESSAGE;
		memcpy(eap_msg_arr[no_eap_message]->data, (char *) eap_pktptr , ep_pkt_size);

		if (eap_pkt_size > EAP_MESSAGE_MAX) {
	        	eap_pktptr = (char *)eap_pktptr + ep_pkt_size;
		}

		if (eap_pkt_size > EAP_MESSAGE_MAX) {
		    eap_pkt_size = eap_pkt_size - EAP_MESSAGE_MAX;
		} else {
			eap_pkt_size = 0;
		}

		eap_msg_arr[no_eap_message]->length = eap_msg_size[no_eap_message];


		no_eap_message++;	
	} while ( eap_pkt_size > 0 );


	eap_state = pairfind(rep->vps, PW_STATE);
	if (0 == eap_state) {
		peap_log_debug("compose_eap_packet: Error while getting PW_STATE \n");
		CLEAN_UP();
		return 0;
	}
	state_tv_size = sizeof(type_length_value_t) + eap_state->length;
	state_tv = calloc(state_tv_size, sizeof(uint8_t));
	if(0 == state_tv) {
		peap_log_debug("compose_eap_packet:Memory alocation failed, state_tv is NULL \n");
		CLEAN_UP();
		return 0;
	}
	state_tv->type = PW_STATE;
	state_tv->length = state_tv_size;
	memcpy(state_tv->data, eap_state->strvalue, eap_state->length);
	tv_size = sizeof(type_length_value_t) + AUTH_VECTOR_LEN;
	tv = calloc(tv_size, sizeof(uint8_t));
	if(0 == tv) {
		peap_log_debug("compose_eap_packet:Memory alocation failed, tv is NULL \n");
		CLEAN_UP();
		return 0;
	}
	tv->type = PW_MESSAGE_AUTHENTICATOR;
	tv->length = tv_size;

	rad_pkt_size = sizeof(radius_packet_t) + state_tv_size +
			tv_size + user_tlv_size + nas_tlv_size + port_tlv_size;


	int no_eap_msg = no_eap_message;
	for ( tmpcount = 0; no_eap_msg > 0; no_eap_msg-- ) {
		rad_pkt_size = rad_pkt_size + eap_msg_size[tmpcount];
		tmpcount++;
	}

	rad_pkt = calloc(rad_pkt_size, sizeof(uint8_t));
	if(0 == rad_pkt) {
		peap_log_debug("compose_eap_packet:Memory alocation failed, nas_tlv is NULL \n");
		CLEAN_UP();
		return 0;
	}
	rad_pkt->code = RADIUS_ACCESS_REQUEST;	/* Radius Access-Request */
	rad_pkt->id = rep->id + 1;	/* Each EAP request should have unique ID*/
	memset(rad_pkt->vector, 0, AUTH_VECTOR_LEN);
	ptr = (uint8_t *) rad_pkt->data;
	temp_len = htons(rad_pkt_size);
	rad_pkt->length = temp_len;

	/* Calculate the vector digest and put into the radius packet */
	librad_md5_calc(req->vector, rep->vector, sizeof(rep->vector));
	memcpy(rad_pkt->vector, req->vector, sizeof(rad_pkt->vector));

	/* Pack the packet by copying all tlv's */
	memcpy(ptr, user_tlv, user_tlv_size);
	ptr += user_tlv_size;

	memcpy(ptr, nas_tlv, nas_tlv_size);
	ptr += nas_tlv_size;

	memcpy(ptr, port_tlv, port_tlv_size);
	ptr += port_tlv_size;

	memcpy(ptr, state_tv, state_tv_size);
	ptr += state_tv_size;

	tmpcount = 0;
	do {
		memcpy(ptr, eap_msg_arr[tmpcount], eap_msg_size[tmpcount]);

		ptr = ptr +  eap_msg_size[tmpcount];
		no_eap_message--;
		tmpcount++;
	} while(no_eap_message > 0);
	memcpy(ptr, tv, tv_size);

	lrad_hmac_md5((uint8_t *) rad_pkt, rad_pkt_size, peap_instance.shared_secret,
				strlen(peap_instance.shared_secret), mesg_auth);
	ptr += 2;	/* Jump eap message length */
	memcpy(ptr, mesg_auth, AUTH_VECTOR_LEN);

	req->data = (uint8_t *) rad_pkt;
	req->data_len = rad_pkt_size;
	CLEAN_UP();
	return req;
}

/*
 * int radpeap_verifyCipher()
 *
 * Description:
 *		This function obtains the cipher configured for RADIUS from the
 *		config file.
 *
 * Input (s):
 * 		Pointer to fill the cipher 
 *
 * Output (s):
 * 		Populate the input with the configured cipher 
 *
 * Returns:
 *   0 False; 1 True;
 */
int 
radpeap_verifyCipher (char* cipher) {

    FILE *fp;
    char cmd[COMMAND_SIZE];
    char path[COMMAND_SIZE];


    memset(cmd, 0, COMMAND_SIZE);
    memset(path, 0, COMMAND_SIZE);

	/* Obtain the configured cipher */
	
                uid_t realUid = getuid();
                uid_t effectiveUid = geteuid();

                setreuid (0,0);
	snprintf(cmd, COMMAND_SIZE, "FABOS_SW_TMP=$FABOS_SWITCHNO FABOS_SWITCHNO=0"
		" LOGIN_ID=root ROLE_ID=root CHASSIS_ROLEID=0"
		" /fabos/cliexec/config get cipher.radius 5; FABOS_SWITCHNO=$FABOS_SW_TMP");

    fp = popen(cmd, "r");
    setreuid(realUid,effectiveUid);
    if (fp == NULL) { 
             peap_log_debug("RADIUS Failed to run command to get configured cipher\n");
             return 0;
	}

    fgets(path, COMMAND_SIZE, fp);
	peap_log_debug("radpeap_verifyCipher path %s\n", path);
    if (strlen(path) > 1) {  // if the string is not empty 1 is returned
		snprintf(cipher, COMMAND_SIZE, "%s", path);
		peap_log_debug("radpeap_verifyCipher configured cipher %s\n", cipher);
		pclose(fp);   
    	return (1);
    } else {
		pclose(fp);   
        return (0);
	}
}

/*
 * tls_session_t * eaptls_new_client_session
 *
 * Description
 *              This is the low level function which initialise the
 *              SSL session with the passed SSL context variable
 *
 *
 * Input
 *              SSL context pointer ,SSL_CTX * ssl_ctx
 *
 * Output
 *              A new tls session is setup
 *
 * Return
 *              Pointer to a tls session object
 */
static tls_session_t *
eaptls_new_client_session(SSL_CTX * ssl_ctx)
{
	tls_session_t * state = 0;
	SSL * new_tls = 0;
        char conf_cipher[COMMAND_SIZE];
	int fipsmode = 0;

		peap_log_debug("eaptls_new_client_session 1: cipher %s \n ",conf_cipher);
	if ((new_tls = SSL_new(ssl_ctx)) == 0) {
		peap_log_debug("eaptls_new_client_session : Error creating new SSL");
		return 0;
	}
        /* If cipher configured for RADIUS, obtain and configure */
        if(radpeap_verifyCipher(conf_cipher)) {
	        /* Set TLS ciphers */ 
		peap_log_debug("eaptls_new_client_session 3: cipher %s \n ",conf_cipher);
		if( SSL_set_cipher_list(new_tls, conf_cipher)!= 1) {
			peap_log_debug("SSL_set_cipher_list : Setting ciphers failed\n");
			return 0;
		}
	}

	/*
	 * We use the SSL's "app_data" to indicate a call-back
	 */
	SSL_set_app_data(new_tls, 0);

	state = (tls_session_t *)calloc(sizeof(tls_session_t), sizeof(uint8_t));
	if(0 == state) {
		peap_log_debug("eaptls_new_client_session: Memory allocation failed, state is NULL \n");
		return 0;
	}
	memset(state, 0, sizeof(tls_session_t));
	session_init(state);
	state->ssl = new_tls;

	/*
	 *      Initialize callbacks
	 */
	state->record_init = record_peap_init;
	state->record_close = record_peap_close;
	state->record_plus = record_peap_plus;
	state->record_minus = record_peap_minus;

	/*
	 *      Create & hook the BIOs to handle the dirty side of the
	 *      SSL.  This is *very important* as we want to handle
	 *      the transmission part.  Now the only IO interface
	 *      that SSL is aware of, is our defined BIO buffers.
	 *
	 *      This means that all SSL IO is done to/from memory,
	 *      and we can update those BIOs from the EAP packets we've
	 *      received.
	 */
	state->into_ssl = BIO_new(BIO_s_mem());
	state->from_ssl = BIO_new(BIO_s_mem());
	SSL_set_bio(state->ssl, state->into_ssl, state->from_ssl);

	/*
	 *      Add the message callback to identify what type of
	 *      message/handshake is passed
	 */
	SSL_set_msg_callback(new_tls, cb_peap_msg);
	SSL_set_msg_callback_arg(new_tls, state);
	SSL_set_info_callback(new_tls, cb_peap_info);

	/*
	 *      In Client mode we only connect.
	 */
	SSL_set_connect_state(state->ssl);

	return state;
}


/*
 * int rad_peap_send
 *
 * Description
 *	Reply to the request
 *	Also attach reply attribute value pairs
 *	and any user message provided.
 *
 * Input
 *	RADIUS_PACKET * packet , const char *secret
 *
 * Output
 *	 None
 *
 * Return
 *	Number of octects it could really sent .
 */
static int
rad_peap_send(RADIUS_PACKET * packet, const RADIUS_PACKET * original,
	      const char * secret)
{
	const char * what;
	struct sockaddr_in saremote;
	struct sockaddr_in * sa;

	/*
	 *      Maybe it's a fake packet.  Don't send it.
	 */
	if (!packet || (packet->sockfd < 0)) {
		return 0;
	}
	if ((packet->code > 0) && (packet->code < 52)) {
		what = packet_codes[packet->code];
	} else {
		what = "Reply";
	}
	/*
	 *      And send it on it's way.
	 */
	sa = (struct sockaddr_in *) &saremote;
	memset((char *) sa, '\0', sizeof(saremote));
	sa->sin_family = AF_INET;
	packet->sockfd = peap_instance.sockfd;
	sa->sin_addr.s_addr = packet->dst_ipaddr;
	sa->sin_port = htons(packet->dst_port);

#ifndef WITH_UDPFROMTO
	return sendto(packet->sockfd, packet->data, (int) packet->data_len, 0,
					(struct sockaddr *) &saremote, sizeof(struct sockaddr_in));
#else
	{
	struct sockaddr_in salocal;
	memset((char *) &salocal, '\0', sizeof(salocal));
	salocal.sin_family = AF_INET;
	salocal.sin_addr.s_addr = packet->src_ipaddr;

	return sendfromto(packet->sockfd, packet->data,(int) packet->data_len, 0,
			  (struct sockaddr *) &salocal, sizeof(struct sockaddr_in),
			  (struct sockaddr *) &saremote, sizeof(struct sockaddr_in));
	}
#endif
}

/*
 * uint8_t * get_tls_data
 *
 * Description
 *              This function encrypts the data passed into it based on TLS
 *              TLS session parameters initaially exchanged .
 *
 * Input
 *              uint8_t * plain_data, uint32_t plain_data_len
 *
 * Output
 *              uint32_t * tls_data_len
 *
 * Return
 *              Returns encrypted TLS data
 */

uint8_t *
get_tls_data(uint8_t * plain_data, uint32_t plain_data_len, uint32_t * tls_data_len)
{
	(peap_instance.tls_session->record_init) (&peap_instance.tls_session->clean_in);
	(peap_instance.tls_session->record_init) (&peap_instance.tls_session->dirty_out);
	(peap_instance.tls_session->record_plus) (&peap_instance.tls_session->clean_in,
														plain_data, plain_data_len);
	/* Feed clean data into the SSL layer */
	SSL_write(peap_instance.tls_session->ssl, peap_instance.tls_session->clean_in.data,
			peap_instance.tls_session->clean_in.used);

	/* Let SSL layer process that, get it encrypted and read back */
	*tls_data_len = BIO_read(peap_instance.tls_session->from_ssl,
			peap_instance.tls_session->dirty_out.data,
			sizeof(peap_instance.tls_session->dirty_out.data));
	if (*tls_data_len > 0) {
		return peap_instance.tls_session->dirty_out.data;
	}
	return 0;
}

/*
 * uint8_t * compose_peap_identity
 *
 * Description
 *              This function sends the peap identity packet
 *              as part of PEAP phase 2 packet exchange, after
 *              setting up the TLS tunnel.
 *
 * Input
 *              int * tls_data_length, Server eap message from global peap instance
 *
 * Output
 *              None
 *
 * Return
 *              TLS Data
 */
static uint8_t *
compose_peap_identity(int * tls_data_length)
{
	char eap_identity_str[MAX_NT_USERNAME] = { 0 };
	int eap_pkt_size = 0;
	uint8_t * eap_message = 0;
	uint8_t * tls_data = 0;
	type_value_t * eap_pkt = 0;

#undef CLEAN_UP
#define CLEAN_UP()  						\
	do {							\
		if(eap_pkt) { free(eap_pkt); eap_pkt = 0; }	\
	} while (0);

	*tls_data_length = 0;

	strncpy(eap_identity_str, peap_instance.user_name , strlen(peap_instance.user_name));

	eap_pkt_size = sizeof(type_value_t) + strlen(eap_identity_str);
	eap_message = peap_instance.srv_eap_message;
	if (0 == eap_message) {
		return 0;
	}
	eap_pkt = calloc(eap_pkt_size, sizeof(uint8_t));
	if (0 == eap_pkt) {
		peap_log_debug("compose_peap_identity: Memory allocation failed ");
		return 0;
	}
	eap_pkt->type = PW_EAP_IDENTITY;
	memcpy(eap_pkt->data, eap_identity_str, strlen(eap_identity_str));
	/*
	 * Pass the clear eap packet and get it encrypted, Ya we can send it by our own
	 * We don't need SSL Layer to do that.
	 */
	tls_data = get_tls_data((uint8_t *)eap_pkt, eap_pkt_size, tls_data_length);
	if(0 == tls_data) {
		peap_log_debug("compose_peap_identity: Error getting tls_data\n");
		CLEAN_UP();
		return 0;
	}
	CLEAN_UP();
	return (uint8_t*)tls_data;
}

/*
 * static RADIUS_PACKET * rad_peap_recv
 *
 * Description
 *              Receive the radius packet , decode the eap message ,
 *              find the mapped eap types inside the packet
 *
 * Input
 *              int sockfd
 *
 * Output
 *              None
 *
 * Return
 *              RADIUS_PACKET * rep
 */
static RADIUS_PACKET *
rad_peap_recv()
{
	RADIUS_PACKET * rep = 0;
	rep = rad_recv(peap_instance.sockfd);
	if(0 == rep) {
		peap_log_debug("rad_peap_recv: rad_decode failed\n");
		return 0;
	}
	if ((rad_decode(rep, 0, peap_instance.shared_secret)) != 0) {
		peap_log_debug("rad_peap_recv: rad_decode failed\n");
		return 0;
	}
	/*
	 * okay got back the packet, go and decode the EAP-Message.
	 */
	unmap_eap_types(rep);
	return rep;
}

/*
 * int setup_TLS_session
 *
 * Description
 *	This function setup and initialize TLS session variables
 *	and necessary call backs
 *
 * Input
 *	None
 *
 * Output
 *	Global peap_instance's TLS context variables are updated
 *
 * Return
 *	Sucess or Failure
 */
static int
setup_TLS_session(void)
{
	SSL_METHOD * meth;
	SSL_CTX * ctx;
	int fipsOrSimulationMode = -1;
	int cert_present    =      -1;
	X509_VERIFY_PARAM *x509Param = NULL;

    int ret = 0, swcert_present, key_present, flag;
	char *tls_cert_file, *tls_key_file;
        char cmd_string[COMMAND_SIZE] = {'\0'}, tls_radius_protocol[20] = {'\0'};
        FILE *pfile = NULL;

	/*
	 * No need to check the return value since it always returns 1
	 */
	ret = SSL_library_init();
	SSL_load_error_strings();
	fipsOrSimulationMode = fipsmode;

	uid_t realUid = getuid();
	uid_t effectiveUid = geteuid();

	/* Appropriate TLS connection method is called as per the RADIUS TLS 
	 * protocol configured */
	setreuid (0,0);
        sprintf (cmd_string, "FABOS_SW_TMP=$FABOS_SWITCHNO FABOS_SWITCHNO=0"
	" LOGIN_ID=root ROLE_ID=root CHASSIS_ROLEID=0"
	" /bin/grep protocol.radius /etc/fabos/fabos.0.conf | /usr/bin/cut -d : -f 2;"
	" FABOS_SWITCHNO=$FABOS_SW_TMP  > /dev/null 2>&1");
        pfile = popen(cmd_string, "r");
	setreuid(realUid,effectiveUid);

	if (pfile != NULL) {
		fscanf(pfile, "%s", tls_radius_protocol);
		if (strncasecmp(tls_radius_protocol, "TLSv1.2", 8)== 0) {
			peap_log_debug("Radius protocol :%s TLSv1_2_client_method()\n", tls_radius_protocol);
			meth = TLSv1_2_client_method();
		} else if (strncasecmp(tls_radius_protocol, "TLSv1", 6)== 0) {
			peap_log_debug("Radius protocol :%s TLSv1_client_method()\n", tls_radius_protocol);
			meth = TLSv1_client_method();
		} else {
			peap_log_debug("Radius protocol :%s SSLv23_client_method()\n", tls_radius_protocol);
			meth = SSLv23_client_method();
		}
		pclose(pfile);
	}
	if (0 == meth) {
		peap_log_debug("setup_TLS_session: Error while setting up TLSv1_method");
		return (-1);
	}
                setreuid (0,0);

	/* Set up an SSL context object, which will be used for SSL session establishment */
	ctx = SSL_CTX_new(meth);
	/* Blocking SSLv2 and SSLv3 */
	SSL_CTX_set_options(ctx,SSL_OP_NO_SSLv2|SSL_OP_NO_SSLv3);
                setreuid(realUid,effectiveUid);

	if (0 == ctx) {
		peap_log_debug("setup_TLS_session: Error while setting up SSL infrastructure");
		return (-1);
	}

	/*
	 * Check for CA certificate. 
	 * In fips or simulation mode CA should be present.
	 */
	cert_present    =       access( RADIUS_CA_FILE, F_OK);

        peap_log_debug("FIPS MODE %d cert_present %d\n",fipsOrSimulationMode,cert_present);
        peap_log_debug("FIPS MODE PAM %d cert_present %d\n",fipsmode,cert_present);

	/* 
	 * If CA cert is present and FIPS or Simulation mode not ON.
	 * Set RADIUS CA certificate 
	 */
	if ((cert_present == SUCCESS)) {

		struct stat st;
    		if (stat(RADIUS_CA_FILE, &st) != 0) {
			peap_log_debug("eaptls_new_client_session: Unable to get size of CAfile\n");
        		return (-1);
		}

		peap_log_debug("file size: %zd\n", st.st_size);
		if ( st.st_size == 0) {
			peap_log_debug("eaptls_new_client_session: zero size cert CAfile\n");
			return (-1);
		}
			
		peap_log_debug("setup_TLS_session: CA cert present\n");
		if ( !SSL_CTX_load_verify_locations(ctx, RADIUS_CA_FILE,NULL))
		{
			peap_log_debug("eaptls_new_client_session: could not load verify locations for CAfile\n");
			return (-1);
		}
	
	} else if (fipsOrSimulationMode == 1) {
			peap_log_debug("FIPS MODE %d\n",fipsOrSimulationMode);
                        peap_log_debug("setup_TLS_session: CA cert must in FIPS or SIMULATE MODE.\n");
			return (-1);
	} else {
                peap_log_debug("setup_TLS_session: No cert validation follows.\n");
	}	
	/*
	 * If No CA cert and NO fips or simulation mode then no Validation of server cert
	 */


    swcert_present = access( SW_CERT_FILE, F_OK);
    peap_log_debug("setupTLSSession swcert_present=%d\n", swcert_present);
    if (swcert_present == 0) {
        tls_cert_file = (char *)SW_CERT_FILE;
        if (!SSL_CTX_use_certificate_file(ctx, tls_cert_file, SSL_FILETYPE_PEM)) {
            peap_log_debug("setupTLSSession Error while setting cert\n");
            return (1);
        }
        key_present = access(PVT_KEY, F_OK);
        peap_log_debug("setupTLSSession key_present=%d\n", key_present);
        if (key_present == 0) {
            tls_key_file = (char *)PVT_KEY;

            if (!SSL_CTX_use_PrivateKey_file(ctx, tls_key_file, SSL_FILETYPE_PEM)) {
                peap_log_debug("setupTLSSession Error while setting pvt key\n");
                return (1);
            }
        } else {
            peap_log_debug("setupTLSSession Could not set pvt key: key not found\n");
            return (1);
        }
    }
	peap_instance.tls_session = eaptls_new_client_session(ctx);
	if (0 == peap_instance.tls_session) {
		return (-1);
	}

	
	/* If CA cert presetnt 
	 * Enable automatic hostname and OCSP validation checks 
	 * */
	if (cert_present == SUCCESS){
		x509Param = SSL_get0_param(peap_instance.tls_session->ssl);

		if (X509_VERIFY_PARAM_set1_host(x509Param, peap_instance.hostname, 0) == 0) {
			peap_log_debug("setup_TLS_session: Setting of servername failed for host validation\n");
			return (-1);
		} else {
			peap_log_debug("setup_TLS_session: Host validation setting passed.\n");
			X509_VERIFY_PARAM_set_flags(x509Param, X509_V_FLAG_OCSP_CHECK);
		}
	} else {
                peap_log_debug("setup_TLS_session: No cert validation follows.\n");
	}
	return 0;
}

/*
 * RADIUS_PACKET * rad_peap_compose_ack
 *
 * Description
 *	This function composes a tls fragment handshake
 *	when a `more fragment' bit set packet is received
 * Input
 *	RADIUS_PACKET * req
 *
 * Output
 *	None
 *
 * Return
 *	RADIUS_PACKET * req
 */
static RADIUS_PACKET *
rad_peap_compose_ack(RADIUS_PACKET * req)
{
	char mesg_auth[AUTH_VECTOR_LEN] = { 0 };
	uint16_t eap_pkt_size = 0;
	uint16_t eap_msg_size = 0;
	uint16_t rad_pkt_size = 0;
	uint16_t tv_size = 0;
	uint16_t state_tv_size = 0;
	uint16_t temp_len = 0;
	char * ptr = 0;
	uint8_t * eap_message = 0;
	eap_message_t * eap_msg = 0;
	eap_tls_ack_t * eap_pkt = 0;
	type_length_value_t * tv = 0;
	type_length_value_t * state_tv = 0;
	VALUE_PAIR * eap_state;
	RADIUS_PACKET * rep = 0;
	radius_packet_t * rad_pkt = 0;

#undef CLEAN_UP
#define CLEAN_UP()  						\
	do {							\
		if(eap_pkt) { free(eap_pkt); eap_pkt = 0; }	\
		if(eap_msg) { free(eap_msg); eap_msg = 0; }	\
		if(state_tv) { free(state_tv); state_tv = 0; } 	\
		if(tv) { free(tv); tv = 0; }			\
	} while (0);

	rep = peap_instance.server_rep;
	eap_pkt_size = sizeof(eap_tls_ack_t);
	tv_size = sizeof(type_length_value_t) + AUTH_VECTOR_LEN;

	/*
	 * Extract PW_EAP_MESSAGE
	 */
	eap_message = peap_instance.srv_eap_message;
	if (0 == eap_message) {
		peap_log_debug("rad_peap_compose_ack: eap_message is NULL");
		return 0;
	}
	eap_pkt = calloc(eap_pkt_size, sizeof(uint8_t));
	if(0 == eap_pkt) {
		peap_log_debug("rad_peap_compose_ack:Memory allocation failed, eap_pkt is NULL\n");
		return 0;
	}
	eap_pkt->code = PW_EAP_RESPONSE;
	eap_pkt->id = (uint8_t) (*(eap_message + 1));
	temp_len = htons(eap_pkt_size);
	eap_pkt->length = temp_len;
	eap_pkt->type = PW_EAP_PEAP;

	/* Let it be a TLS fragment ack packet */
	eap_pkt->flags = EAPTLS_ACK;
	eap_msg_size = sizeof(eap_message_t) + eap_pkt_size;
	eap_msg = calloc(eap_msg_size, sizeof(uint8_t));
	if(0 == eap_msg) {
		peap_log_debug("rad_peap_compose_ack:Memory allocation failed, eap_msg is NULL\n");
		CLEAN_UP();
		return 0;
	}
	eap_msg->code = PW_EAP_MESSAGE;
	memcpy(eap_msg->data, (eap_tls_ack_t *) eap_pkt, eap_pkt_size);
	eap_msg->length = eap_msg_size;


	/* Look for state tlv in the server reply packet,
	 * so that we can copy that to client request packet
	 */
	eap_state = pairfind(rep->vps, PW_STATE);
	if (0 == eap_state) {
		peap_log_debug("rad_peap_compose_ack: Error while getting PW_STATE");
		CLEAN_UP();
		return 0;
	}

	/* State tlv, state is send by the server, Just copy it into each packet */
	state_tv_size = sizeof(type_length_value_t) + eap_state->length;
	state_tv = calloc(state_tv_size, sizeof(uint8_t));
	if(0 == state_tv) {
		peap_log_debug("rad_peap_compose_ack:Memory allocation failed, state_tv is NULL\n");
		CLEAN_UP();
		return 0;
	}
	state_tv->type = PW_STATE;
	state_tv->length = state_tv_size;
	memcpy(state_tv->data, eap_state->strvalue, eap_state->length);

	tv_size = sizeof(type_length_value_t) + AUTH_VECTOR_LEN;
	tv = calloc(tv_size, sizeof(uint8_t));
	if(0 == tv) {
		peap_log_debug("rad_peap_compose_ack:Memory allocation failed, tv is NULL\n");
		CLEAN_UP();
		return 0;
	}
	tv->type = PW_MESSAGE_AUTHENTICATOR;
	tv->length = tv_size;
	rad_pkt_size = sizeof(radius_packet_t) + eap_msg_size +
							state_tv_size +	tv_size;

	rad_pkt = calloc(rad_pkt_size, sizeof(uint8_t));
	if(0 == rad_pkt) {
		peap_log_debug("rad_peap_compose_ack:Memory allocation failed, rad_pkt is NULL\n");
		CLEAN_UP();
		return 0;
	}

	/* Client can only send Radius Access Requests */
	rad_pkt->code = RADIUS_ACCESS_REQUEST;	/* Radius Access-Request */
	rad_pkt->id = rep->id + 1;	/* Packet ID should different for each EAP request*/

	memset(rad_pkt->vector, 0, AUTH_VECTOR_LEN);

	/* Copy the EAP Message portion into the packet */
	ptr = (uint8_t *) rad_pkt->data;
	memcpy(ptr, eap_msg, eap_msg_size);

	temp_len = htons(rad_pkt_size);
	rad_pkt->length = temp_len;

	/* Calculate the md5 hash for the vector and pack into the packet */
	librad_md5_calc(req->vector, rep->vector, sizeof(rep->vector));
	memcpy(rad_pkt->vector, req->vector, sizeof(rad_pkt->vector));

	/* Copy the state tlv */
	ptr += eap_msg_size;
	memcpy(ptr, state_tv, state_tv_size);

	/* Copy the message authenticator */
	ptr += state_tv_size;
	memcpy(ptr, tv, tv_size);

	lrad_hmac_md5((uint8_t *) rad_pkt, rad_pkt_size, peap_instance.shared_secret,
			strlen(peap_instance.shared_secret), mesg_auth);
	ptr += 2; 	/*
			 * Skip the length field, Now its points to the data part
			 */
	memcpy(ptr, mesg_auth, AUTH_VECTOR_LEN);
	req->data = (uint8_t *) rad_pkt;
	req->data_len = rad_pkt_size;
	CLEAN_UP();
	return req;
}

/*
 * uint8_t * extract_tls_data
 *
 * Description
 *	This function extracts the TLS layer data from eap message data
 *
 * Input
 *	eap message data from global peap_instance
 *
 * Output
 *	int *data_len , uint8_t * more_tls_data
 *
 * Return
 *	Tls data
 */
static uint8_t *
extract_tls_data(int * data_len, uint8_t * more_tls_data)
{
	uint16_t tls_data_len = 0;
	uint8_t * eap_message = 0;
	uint8_t * tls_data = 0;

	/* Get a local pointer to the eap message */
	eap_message = peap_instance.srv_eap_message;
	if (0 == eap_message) {
		peap_log_debug("extract_tls_data: eap_message is NULL");
		return (0);
	}
	/* Check for `Length Included' flag */
	if (((eap_tls_response_t *) eap_message)->flags & EAP_LENGTH_INCLUDED) {
		tls_data = (uint8_t *) ((eap_tls_response_t *)eap_message)->data;
		tls_data_len = ((eap_tls_response_t *)eap_message)->length;
		tls_data_len = ntohs(tls_data_len) - TLS_HDR_LEN_PLUS_LENGTH;

	} else {
	/* `Length Included' flag is clear */
		tls_data = (uint8_t *)(&(((eap_tls_response_t *) eap_message)->tls_msg_len));
		tls_data_len = ((eap_tls_response_t *) eap_message)->length;
		tls_data_len = ntohs(tls_data_len) - TLS_HDR_LEN;
	}
	/* Check `more tls data' bit is set or not */
	if (((eap_tls_response_t *) eap_message)->flags & MORE_TLS_DATA) {
		*more_tls_data = 1;
	} else {
		*more_tls_data = 0;
	}
	*data_len = tls_data_len;
	return tls_data;
}

/*
 * uint8_t * process_peap_message
 *
 * Description
 *	This function  get the unencrypted peap data
 *	out of encrypted tls data
 *
 * Input
 *	Gets the encrypted tls data from global peap_instance
 *
 * Output
 *	Unencryted data from tls layer
 *
 * Return
 *	int *data_length
 */
static uint8_t *
process_peap_message(int * data_length)
{
	uint8_t more_peap_data = 0;
	uint16_t client_state = 0;
	int peap_data_length = 0;
	int read_bytes = 0;
	uint8_t  * peap_data = 0;

	peap_data = extract_tls_data((int *) &peap_data_length,	(uint8_t *) & more_peap_data);
	if (0 == peap_data) {
		peap_log_debug("process_peap_message: Error while extracting PEAP data: state 0x%x \n",
																 client_state);
		peap_log_debug("process_peap_message: Client state 0x%x, SSL state 0x%x \n",
				client_state, SSL_get_state(peap_instance.tls_session->ssl));
		return 0;
	}
	/* Pass it on to tls layer */
	(peap_instance.tls_session->record_init)(&peap_instance.tls_session->dirty_in);
	(peap_instance.tls_session->record_init)(&peap_instance.tls_session->clean_out);
	(peap_instance.tls_session->record_plus)(&peap_instance.tls_session->dirty_in,
														peap_data, peap_data_length);
	/* Feed the encrypted TLS data into SSL layer */
	BIO_write(peap_instance.tls_session->into_ssl, peap_instance.tls_session->dirty_in.data,				peap_data_length);
	/* Get clean data out of SSL layer */
	read_bytes = SSL_read(peap_instance.tls_session->ssl,
			peap_instance.tls_session->clean_out.data,
			sizeof(peap_instance.tls_session->clean_out.data));

	if (read_bytes <= 0) {
		peap_log_debug("process_peap_message: Error while decoding PEAP"
			" data or no PEAP data: state 0x%x \n",	client_state);
		peap_log_debug("process_peap_message: Client state 0x%x, SSL state 0x%x \n",
			client_state, SSL_get_state(peap_instance.tls_session->ssl));
		return 0;
	}
	*data_length = read_bytes;
	return (peap_instance.tls_session->clean_out.data);
}

/*
 * int  process_eap_message
 *
 * Description
 *	This function verify's what kind of eap packet we recieved
 *	To get the eap message , it refers to the global insatnce
 * Input
 *	Takes input from global peap_instace
 *
 * Output
 *	None
 *
 * Return
 *	Sucess or failure
 */
static int
process_eap_message(void)
{
	uint8_t * eap_msg = 0;
	uint8_t eap_code = 0;
	uint8_t eap_type = 0;

	/*
	 * Retrieve the EAP message from the server response
	 */
	eap_msg = (uint8_t *) peap_instance.srv_eap_message;
	if (0 == eap_msg) {
		peap_log_debug("process_eap_message: srv_eap_message is NULL!!!");
		return (-1);
	}
	eap_code = ((eap_tls_response_t *) eap_msg)->code; /* Get EAP code */
	eap_type = ((eap_tls_response_t *) eap_msg)->type;

	switch (eap_code) {
		case PW_EAP_REQUEST:
		/* Are we talking in PEAP ? */
		switch (eap_type) {
			case PW_EAP_PEAP:
	    			return 0; 	/* Yes!! We do talk in PEAP,  If Not.. Die !!! */
			case PW_EAP_SECUREID:
				return (PAM_AUTH_SECUREID);
			case PW_EAP_MSCHAPV2:
			case PW_EAP_IDENTITY:
			case PW_EAP_NAK:
			default:
				peap_log_debug("process_eap_message: Invalid EAP type 0x%x \n", eap_type);
				return (-1);
	}
	case PW_EAP_SUCCESS:
	case PW_EAP_FAILURE:
	default:
		peap_log_debug("process_eap_message: Invalid EAP code 0x%x \n", eap_code);
		return (-1);
	}
	return (-1);
}

/*
 * int extract_eap_message
 *
 * Description
 *	Extract the EAP-MESSAGE AVP from the
 *	radius packet and update into the global `peap_instance'
 *
 * Input
 *	None
 *
 * Output
 *	None
 *
 * Return
 *	Success or Failure
 */
static int
extract_eap_message(void)
{
	uint16_t len = 0;
	int total_len = 0;
	uint8_t * eap_message = 0;
	uint8_t * ptr = 0;
	VALUE_PAIR * first_vp = 0;
	VALUE_PAIR * tmp_vp = 0;
	VALUE_PAIR * vp = 0;
	RADIUS_PACKET * rep = 0;

	/* Get the latest server reply */
	rep = peap_instance.server_rep;
	if (0 == rep) {
		peap_log_debug("extrace_eap_message: server rep is NULL!!!");
		return (-1);
	}
	/*
	 * Extract PW_EAP_MESSAGE
	 */
	first_vp = pairfind(rep->vps, PW_EAP_MESSAGE);
	if (0 == first_vp) {
		peap_log_debug("extrace_eap_message: server first_vp is NULL!!!");
		return (-1);
	}
	memcpy(&len, first_vp->strvalue + 2, sizeof(len));
	len = ntohs(len);
	/*
	 * Compute the entire EAP message length
	 */
	for (vp = first_vp; vp != 0; vp = vp->next) {
		tmp_vp = pairfind(vp, PW_EAP_MESSAGE);
		if (0 == tmp_vp) {
		    break;
		}
		total_len += tmp_vp->length;
	}
	/*
	 * Check for error condition
	 */
	if (total_len > len) {
		peap_log_debug("extract_eap_message: Improper EAP message length");
		return (-1);
	}

	eap_message = calloc(total_len, sizeof(uint8_t));
	if (0 == eap_message) {
		peap_log_debug("extract_eap_message: Memory allocation failed for eap_message");
		return (-1);
	}
	ptr = eap_message;
	/*
	 * Extract whole EAP message present in the server response
	 */
	for (vp = first_vp; vp != 0; vp = vp->next) {
		tmp_vp = pairfind(vp, PW_EAP_MESSAGE);
		if (0 == tmp_vp) {
		    break;
		}
		memcpy(ptr, tmp_vp->strvalue, tmp_vp->length);
		ptr += tmp_vp->length;
	}
	/* Free the prevoius EAP message */
	if(peap_instance.srv_eap_message)
		free(peap_instance.srv_eap_message);

	/* Update into global peap_instance */
	peap_instance.srv_eap_message = eap_message;
	return 0;
}

/*
 * int process_rad_rep
 *
 * Description
 *	Process the radiud reply from server ,
 *	checks for Access-Reject
 *
 * Input
 *	None
 *
 * Output
 *	None
 *
 * Return
 *	Success or failure , 0 or -1
 */
static int
process_rad_rep()
{
	RADIUS_PACKET * rep = 0;
	RADIUS_PACKET * req = 0;

	rep = peap_instance.server_rep;
	req = peap_instance.current_rad_pkt;
	/*
	 * Check for access-reject message
	 */
	if (PW_AUTHENTICATION_REJECT == rep->code) {
		peap_log_debug("process_rad_rep: Access Reject received: state 0x%x \n",
			peap_instance.client_state);
		peap_log_debug("process_rad_rep: Dumping the req & rep packets");
		return (-1);
	}
	return 0;
}

/*
 * void compose_radius_pkt
 *
 * Description
 *	Add the necessary fields to the
 *	radius data structure maintained ..
 *	This function does not really compose a radius packet
 *
 * Input
 *	RADIUS_PACKET * req , int rad_code
 *
 * Output
 *	None
 *
 * Return
 *	None
 */
void
compose_radius_pkt(RADIUS_PACKET * req, int rad_code)
{
	req->code = rad_code;
	req->dst_ipaddr = peap_instance.dst_ipaddr;
	req->dst_port = peap_instance.dst_port;
	req->sockfd = peap_instance.sockfd;
}

/*
 * int peap_add_to_list
 *
 * Description
 *	Add the tls fragments into the linked list of tls fragments
 *
 * Input
 *	list ,  data , len
 *
 * Output
 *	None
 *
 * Return
 *	success or failure
 */
static int
peap_add_to_list(tls_data_list_t * list, uint8_t * data, int len)
{
	tls_data_list_t *tls_element = 0;

#undef CLEAN_UP
#define CLEAN_UP() 				\
	do {					\
		if(tls_element) {		\
			free(tls_element);	\
			tls_element = 0;	\
		}				\
	} while (0);

	tls_element = calloc(sizeof(tls_data_list_t), sizeof(uint8_t));
	if (0 == tls_element) {
		peap_log_debug("peap_add_to_list: Error while allocating memory for tls_element");
		return -1;
	}
	tls_element->tls_data = calloc(len, sizeof(uint8_t));
	if (0 == tls_element->tls_data) {
		peap_log_debug("peap_add_to_list: Error while allocating memory tls_element data");
		CLEAN_UP();
		return -1;
	}
	/* Add up the TLS fragments into a link list */
	memcpy(tls_element->tls_data, data, len);
	tls_element->tls_data_length = len;
	tls_element->next = 0;
	if (0 == list->list_end) {
		list->next = list->list_end = tls_element;
	} else {
		list->list_end->next = tls_element;
		list->list_end = tls_element;
	}
	return 0;
}

/*
 * uint8_t *  peap_remove_from_list
 *
 * Description
 *	Remove the tls fragments queue nodes ,once we receive
 *	all the tls fragments
 *
 * Input
 *	tls_data_list_t * list
 *
 * Output
 *	int * len
 *
 * Return
 *	Return tls data
 */
static uint8_t *
peap_remove_from_list(tls_data_list_t * list, int *len)
{
	uint8_t * data = 0;
	tls_data_list_t * tls_entry = 0;

	*len = 0;
	tls_entry = list->next;
	if (0 != tls_entry) {
		/*
		 * Free the nodes as they get removed
		 */
		list->next = tls_entry->next;
		data = tls_entry->tls_data;
		*len = tls_entry->tls_data_length;
	} else {
		/*
		 * We are at the end of the list
		 */
		list->list_end = list->next = 0;
	}
	free(tls_entry);
	return data;
}

/*
 * uint8_t * compose_mschap_response
 *
 * Description
 *	Function  builds an mschap response packet from the
 *	access challenge send by the server based on the
 *	state machine's  state
 *
 * Input
 *	Challenge issued by server ,username ,password
 *
 * Output
 *	Composed mschap reply packets's length
 *
 * Return
 *	Returns the inner tlv peap data
 */
static uint8_t *
compose_mschap_response(int * mschap_rep_len)
{
	char passwd[MAX_NT_PASSWORD] = { 0 };
	char username[MAX_NT_USERNAME] = { 0 };
	char gen_auth_response[MAX_AUTH_RESPONSE_LEN * 2] = { 0 };
	char recv_auth_response[MAX_AUTH_RESPONSE_LEN * 2] = { 0 };
	unsigned char auth_response[MAX_AUTH_RESPONSE_LEN] = { 0 };
	uint8_t count = 0;
	uint8_t nthash = 0;
	uint16_t res_len = 0 ;
	uint8_t * ptr = 0;
	VALUE_PAIR * first_vp = 0;
	mschapv2_request_packet_t * mschap_req = 0;
	mschapv2_response_packet_t * mschap_rep = 0;
	RADIUS_PACKET * rep = 0;

	strncpy(passwd, peap_instance.user_passwd, strlen(peap_instance.user_passwd));
	strncpy(username, peap_instance.user_name, strlen(peap_instance.user_name));

	/* Get the latest server reply */
	rep = peap_instance.server_rep;
	if (0 == rep) {
		peap_log_debug("compose_mschap_response : server rep is NULL!!!");
		return 0;
	}
	/* Extract the EAP Message from the AVP's */
	first_vp = pairfind(rep->vps, PW_EAP_MESSAGE);
	if (0 == first_vp) {
		peap_log_debug("compose_mschap_response : %d: first_vp is NULL \n",
						__LINE__);
		return 0;
	}
	if ((CLIENT_TLS_ESTABLISHED | CLIENT_RECEIVE_AUTH_RESULT) ==									peap_instance.client_state) {
		/*
		 * Get the authenticator response
	 	 */
		memcpy(recv_auth_response,&(peap_instance.tls_session->								clean_out.data[7]), MAX_AUTH_RESPONSE_LEN * 2);
		/*
		 * Generate authenticator response
		 */
		GenerateAuthenticatorResponse(passwd, cached_nt_response,							cached_peer_challenge,cached_auth_challenge,
					username, auth_response, nthash);
		/*
		 * Encode the 20 bytes response to 40 bytes ascii
		 */
		for (count = 0; count < MAX_AUTH_RESPONSE_LEN; count++) {
		    encode_asc(auth_response[count],&gen_auth_response[count * 2]);
		}
		/*
		 * Verify the authenticator response
		 */
		if (0 != strncmp(gen_auth_response, recv_auth_response,
										MAX_AUTH_RESPONSE_LEN * 2)) {
			peap_log_debug("compose_mschap_response: WARNING!!!!Authenticator response doesnt match!!!!");
			return 0;
		}

		*mschap_rep_len = 1 /* PW_EAP_MSCHAPV2 */  + 1 /* Mschap opcode size*/;

		ptr = (uint8_t *)calloc(*mschap_rep_len, sizeof(uint8_t));
		if (0 == ptr) {
		    peap_log_debug("compose_mschap_response: Memory allocation failed ");
		    return 0;
		}
		*ptr = PW_EAP_MSCHAPV2;					/*EAP-Type */
		*(ptr + 1) = PW_EAP_MSCHAPV2_SUCCESS; 	/*Mschap opcode */

		return ptr; 	/* Return from here itself, don't fall through */

	} /* else part */

	*mschap_rep_len = 1	/* Type PW_EAP_MSCHAPV2 */ +
			sizeof(mschapv2_response_packet_t) - 1 +
			strlen(peap_instance.user_name);

	(*mschap_rep_len) += 1;
				/* Ayeee!!!  ^^Hack^^  Added for Null terminating user-name string
				 * Here W2K3 Server deviates from RFC ...
				 */

	ptr = calloc(*mschap_rep_len, sizeof(uint8_t));
	if (0 == ptr) {
		peap_log_debug("compose_mschap_response: Memmory allocation failed");
		return 0;
	}
	*ptr = PW_EAP_MSCHAPV2;
	mschap_rep = (mschapv2_response_packet_t *) (ptr + 1 );	/*Skip Eap-Type*/
	mschap_rep->opcode = PW_EAP_MSCHAPV2_RESPONSE;
	mschap_rep->mschapv2_id = *((uint8_t *) (first_vp->strvalue + 1));

	res_len = htons(*mschap_rep_len);
	memcpy(mschap_rep->ms_length, &res_len, sizeof(uint16_t));
	mschap_rep->value_size = MSCHAP_RESPONSE_VALUE_SIZE & 0xff;
	mschap_req = (mschapv2_request_packet_t *)((peap_instance.tls_session->clean_out.data) + 1);

	for (count = 0; count < MSCHAPV2_CHALLENGE_LEN; count++) {
		mschap_rep->peer_challenge[count] = lrad_rand();
		/* Just random numbers, Thats it !! */
	}
	memset(mschap_rep->reserved, 0, MS_RESERVED_BYTES);	/* 8 bytes */

	/* Get Ms-Chap response as per the RFC */
	GenerateNTResponse(mschap_req->auth_challenge, mschap_rep->peer_challenge,
			username, passwd, mschap_rep->nt_response, nthash);
	/* Cache the challenges and responses for future use */
	memcpy(cached_auth_challenge, mschap_req->auth_challenge,
			MAX_CHALLENGE_LEN);
	memcpy(cached_peer_challenge, mschap_rep->peer_challenge,
			MAX_CHALLENGE_LEN);
	memcpy(cached_nt_response, mschap_rep->nt_response, MAX_RESPONSE_LEN);
	mschap_rep->flags = 0x0; /* Clear Flag field */
	memcpy(mschap_rep->data, username, strlen(username));
	return ptr;
}

/*
 * int wait_for_server_response
 *
 * Description
 *	Wait for the server response for the server specific timeout values
 *	Maintain the value of `retreis' between the function calls
 *
 * Input
 *	None
 *
 * Output
 *	None
 *
 * Return
 *	PEAP_SERVER_RETRY or Success or Failure
 */
int
wait_for_server_response()
{
	time_t now = { 0 };
	time_t end = { 0 };
	struct timeval tv = { 0 };
	fd_set set;
	radius_server_t * server = 0;
	int ret = 0;

	server = peap_instance.cur_serv;
	if (0 == server) {
		peap_log_debug("wait_for_server_response: cur_serv is NULL \n");
		return(-1);
	}

	/* wait for the server specific timeout */
	tv.tv_sec = server->timeout;
	tv.tv_usec = 0;

	/* Clear out the set */
	FD_ZERO(&set);
	/* wait only for the RADIUS UDP socket */
	FD_SET(peap_instance.sockfd, &set);

	time(&now);
	end = now + tv.tv_sec;
	ret = select(peap_instance.sockfd + 1, &set, NULL, NULL, &tv);

	/*
	 * Is it timed out???
	 */
	if (0 == ret) {
		peap_log_debug("wait_for_server_response: Server timed out. peap_instance.retriesleft %d \n",
				peap_instance.retries);
		if (--(peap_instance.retries)) {
			return PEAP_SERVER_RETRY;
		} else {
			return PAM_AUTHINFO_UNAVAIL;
		}
	}
	/*
	 * Interrupted or error
	 */
	if (ret < 0) {
		peap_log_debug("wait_for_server_response: Error in accessing the server \n");
		return -1;
	}
	if (FD_ISSET(peap_instance.sockfd, &set)) {
		peap_log_debug("wait_for_server_response: Server responded. Moving ahead \n");
		return 0;
	}
	return 0;
}
/*
 *int handle_req_rep
 *
 * Description
 *	This function handles the request and reply (response)
 *	excahnged between peap client and server
 *	It maintains a state machine
 * Input
 *	None
 *
 * Output
 *	None
 *
 * Return
 *	Retuns  Success or Failure
 */
static int
handle_req_rep(void)
{
	uint8_t exit = FALSE;
	uint8_t more_tls_data = 0;
	uint8_t result_avp[11] = { 0 };
	uint16_t client_state = 0;
	int read_bytes = 0;
	int tls_data_length = 0;
	int peap_data_length = 0;
	int total_length = 0;
	int ret = 0;
	uint8_t * tls_data = 0;
	uint16_t temp_len = 0;
	uint8_t * peap_data = 0;
	RADIUS_PACKET * req = 0;
	RADIUS_PACKET * rep = 0;
	int retval = 0;

	/* Take the initial cleint_state, to start with we user it here */
	client_state = peap_instance.client_state;

	/*
	 * Peap client state machine is maintained here
	 * For easy understandabillity, States are given in the order they transit
	 * form one to other, from top to bottom
	 */
	do {
		switch (client_state) {
		case CLIENT_INIT_STATE:
			free(req);
			req = 0;
			req = rad_alloc(RAD_NEW_VECTOR);
			if (0 == req) {
				peap_log_debug("handle_req_rep: Error while allocating req: state 0x%x \n",
					client_state);
				return (-1);
			}
			compose_radius_pkt(req, PW_AUTHENTICATION_REQUEST);

			/* Build the first EAP idendity packet */
			ret = compose_eap_identity_pkt(req);
			if (0 != ret) {
				peap_log_debug("handle_req_rep: Error while composing EAP identity pkt: state 0x%x \n",
					client_state);
				free(req);
				return (-1);
			}
			peap_instance.current_rad_pkt = peap_instance.prev_rad_pkt = req;
			client_state = peap_instance.client_state = CLIENT_EAP_CONNECT;
			break;
		case CLIENT_EAP_CONNECT:
			/* Now we have a server reply, go and process */
			ret = process_eap_message();

			peap_log_debug("handle_req_rep: after calling process_eap_message \n");
			if (ret < 0) {
				peap_log_debug("handle_req_rep: Error while processing radius response: state 0x%x \n",
					client_state);
				if (ret == PAM_AUTH_SECUREID) {
					peap_log_debug("handle_req_rep:ret%d\n",ret);
					return (PAM_AUTH_SECUREID);
				}
				return (-1);
			}
			ret = setup_TLS_session();

			peap_log_debug("handle_req_rep: after calling setup_TLS_session \n");
			if (ret < 0) {
				peap_log_debug("handle_req_rep: Error while setting up TLS session: state 0x%x \n",
					client_state);
				return (-1);
			}
			peap_log_debug("handle_req_rep: before record_init \n");

			(peap_instance.tls_session->record_init)(&peap_instance.tls_session->dirty_out);

			/*
			 * Send Client Hello
			 */
			read_bytes = SSL_read(peap_instance.tls_session->ssl,
					peap_instance.tls_session->dirty_out.data,
					sizeof(peap_instance.tls_session->dirty_out.data));
			peap_log_debug("handle_req_rep: after calling SSL_read \n");

			read_bytes = BIO_read(peap_instance.tls_session->from_ssl,
					peap_instance.tls_session->dirty_out.data,
					sizeof(peap_instance.tls_session->dirty_out.data));
			peap_log_debug("handle_req_rep: after calling BIO_read \n");

			free(req);
			req = 0;
			req = rad_alloc(RAD_NEW_VECTOR);
			if (0 == req) {
				peap_log_debug("handle_req_rep: Error while allocating req: state 0x%x \n",
					client_state);
				return (-1);
			}
			peap_log_debug("handle_req_rep: after calling rad_alloc \n");

			/* Get all details about radius server */
			compose_radius_pkt(req, PW_AUTHENTICATION_REQUEST);

			peap_log_debug("handle_req_rep: after calling compose_radius_pkt \n");

			/* Compose  radius packet packet on-wire */
			req = compose_eap_packet(req,peap_instance.tls_session->dirty_out.data,
																	read_bytes, TRUE);
			peap_log_debug("handle_req_rep: after calling compose_eap_packet \n");

			if (0 == req) {
				peap_log_debug("handle_req_rep: Error while composing EAP request: state 0x%x \n",
					client_state);
				free(req);
				return (-1);
			}
			/* Keep the SSL buffers updated */
			(peap_instance.tls_session->record_init)(&peap_instance.tls_session->dirty_out);
			peap_instance.prev_rad_pkt = peap_instance.current_rad_pkt;
			peap_instance.current_rad_pkt = req;
			client_state = peap_instance.client_state =	CLIENT_TLS_HANDSHAKE;
			break;
		case CLIENT_TLS_HANDSHAKE:
			ret = process_eap_message();
			if (ret < 0) {
				peap_log_debug("handle_req_rep: Error while processing radius response: state 0x%x \n",
					client_state);
				peap_log_debug("handle_req_rep: Client state 0x%x, SSL state 0x%x \n",
					client_state, SSL_get_state(peap_instance.tls_session->ssl));
				if (ret == PAM_AUTH_SECUREID) {
					return (PAM_AUTH_SECUREID);
				}
				return (-1);
			}
			/* Get the tls data out of eap message */
			tls_data = extract_tls_data((int *) &tls_data_length, (uint8_t *) & more_tls_data);
			if (0 == tls_data) {
				peap_log_debug("handle_req_rep: Error while extracting TLS data: state 0x%x \n",
					client_state);
				peap_log_debug("handle_req_rep: Client state 0x%x, SSL state 0x%x \n",
					client_state,
				SSL_get_state(peap_instance.tls_session->ssl));
				return (-1);
			}
			/* Wait untill till all TLS fragments are received, now we can proceed */
			if (0 == more_tls_data) {
				(peap_instance.tls_session->record_init)(&peap_instance.tls_session->dirty_in);
				peap_add_to_list(&peap_instance.srv_tls_data_list, tls_data, tls_data_length);
				while (TRUE) {
					tls_data = peap_remove_from_list(&peap_instance.srv_tls_data_list,
						&tls_data_length);
					if (0 == tls_data) {
						break;
					}
					(peap_instance.tls_session->record_plus)(&peap_instance.tls_session->dirty_in,
						tls_data, tls_data_length);
					free(tls_data);
					total_length += tls_data_length;
				}
			}
			if (1 == more_tls_data) {
				peap_add_to_list(&peap_instance.srv_tls_data_list, tls_data, tls_data_length);
				client_state = peap_instance.client_state = (CLIENT_TLS_HANDSHAKE | CLIENT_SEND_ACK);
				continue;	/* Loop in these two states untill we get all TLS fragments */
			}
			BIO_write(peap_instance.tls_session->into_ssl, peap_instance.tls_session->dirty_in.data,
				total_length);
			total_length = 0;
			(peap_instance.tls_session->record_init)(&peap_instance.tls_session->clean_out);
			read_bytes = SSL_read(peap_instance.tls_session->ssl,
					peap_instance.tls_session->clean_out.data,
					sizeof(peap_instance.tls_session->clean_out.data));
			if (read_bytes > 0) {
				peap_instance.tls_session->clean_out.used = read_bytes;
			}
			/* Get the dirty data from SSL layer */

			read_bytes = BIO_read(peap_instance.tls_session->from_ssl,
					peap_instance.tls_session->clean_out.data,
					sizeof(peap_instance.tls_session->clean_out.data));


		        int cert_present = access( RADIUS_CA_FILE, F_OK);

                	if (cert_present == SUCCESS) {
                        	peap_log_debug("handle_req_rep: CA cert Present.\n");

				int ret1 = -1;
				if(SSL_get_peer_certificate(peap_instance.tls_session->ssl)) {
	                                if ((ret1 = SSL_get_verify_result(peap_instance.tls_session->ssl)) && 
						(ret1 != X509_V_ERR_UNABLE_TO_FIND_OCSP_URL)) {
	                                        peap_log_debug("handle_req_rep: Server CA cert verification failed: ret1 is %d\n", ret1);
						if (ret1 == X509_V_ERR_HOSTNAME_MISMATCH) {
							if (checkCertValidationConfiguration() == CERTVALIDATION_STRICT) {
								X509_verify_cert_error_string(X509_V_ERR_HOSTNAME_MISMATCH);
								return (X509_V_ERR_HOSTNAME_MISMATCH);
							} else {
	                                        		peap_log_debug("handle_req_rep: Ignoring hostname mismatch error..");
								ret1 = 0;
							}
							}
						else if (ret1 == X509_V_ERR_CERT_UNKNOWN_TO_OCSP)
							return (X509_V_ERR_CERT_UNKNOWN_TO_OCSP);
						else if (ret1 == X509_V_ERR_CERT_REVOKED_TO_OCSP)
							return (X509_V_ERR_CERT_REVOKED_TO_OCSP);
						else if ((ret1 == X509_V_ERR_UNPARSABLE_OCSP_URL)||
							 (ret1 == X509_V_ERR_SSL_USED_FOR_OCSP)||
							 (ret1 == X509_V_ERR_OCSP_PROCESS_FUNC_NOT_SET)|| 
						 	 (ret1 == X509_V_ERR_OCSP_RESPONDER_STATUS_FAIL)||
							 (ret1 == X509_V_ERR_OCSP_NONCE_VERIFY_ERROR)||
							 (ret1 == X509_V_ERR_OCSP_RESPONSE_VERIFICATION)||
							 (ret1 == X509_V_ERR_OCSP_APPLICATION_VERIFICATION))
							return (X509_V_ERR_OCSP_APPLICATION_VERIFICATION);
						else 
	        	                                return (ret1);
					} else {
						if (ret1 == X509_V_ERR_UNABLE_TO_FIND_OCSP_URL) {
							peap_log_debug("handle_req_rep: OCSP Validation enabled, but bypassed as OCSP URL not found\n");
							ret1 = 0;
						}
						peap_log_debug("handle_req_rep: Server CA cert verification passed\n");
					}
				} else {
						peap_log_debug("handle_req_rep: Server CA cert not provided\n");
        	                                return (-1);
				}
			} else {
		                peap_log_debug("setup_TLS_session: No cert validation follows.\n");
			}
			
			if (SSL_ST_OK == SSL_get_state(peap_instance.tls_session->ssl)) {
				/*
				 * We are now done with establishing TLS channel.
				 */
				client_state = peap_instance.client_state = (CLIENT_TLS_ESTABLISHED | CLIENT_PEAP_START);
				free(tls_data);
				continue;
			}
			free(req);
			req = 0;
			req = rad_alloc(RAD_NEW_VECTOR);
			if (0 == req) {
				peap_log_debug("handle_req_rep: Error while allocating req: state 0x%x \n",
					client_state);
				peap_log_debug("handle_req_rep: Client state 0x%x, SSL state 0x%x \n",
					client_state, SSL_get_state(peap_instance.tls_session->ssl));
				return (-1);
			}
			compose_radius_pkt(req, PW_AUTHENTICATION_REQUEST);
			req = compose_eap_packet(req, peap_instance.tls_session->clean_out.data,
					read_bytes, TRUE);
			if (0 == req) {
				peap_log_debug("handle_req_rep: Error while composing EAP request: state 0x%x \n",
					client_state);
				peap_log_debug("handle_req_rep: Client state 0x%x, SSL state 0x%x \n", client_state,
					SSL_get_state(peap_instance.tls_session->ssl));
				free(req);
				return (-1);
			}
			peap_instance.prev_rad_pkt = peap_instance.current_rad_pkt;
			peap_instance.current_rad_pkt = req;
			break;
		case (CLIENT_TLS_HANDSHAKE | CLIENT_SEND_ACK):
			free(req);
			req = 0;
			req = rad_alloc(RAD_NEW_VECTOR);
			if (0 == req) {
				peap_log_debug("handle_req_rep: Error while allocating req: state 0x%x \n",
					client_state);
				peap_log_debug("handle_req_rep: Client state 0x%x, SSL state 0x%x \n",
					client_state, SSL_get_state(peap_instance.tls_session->ssl));
				return (-1);
			}
			compose_radius_pkt(req, PW_AUTHENTICATION_REQUEST);
			req = rad_peap_compose_ack(req);
			if (0 == req) {
				peap_log_debug("handle_req_rep: Error while composing TLS ack: state 0x%x \n",
					client_state);
				peap_log_debug("handle_req_rep: Client state 0x%x, SSL state 0x%x \n",
					client_state, SSL_get_state(peap_instance.tls_session->ssl));
				free(req);
				return (-1);
			}
			peap_instance.prev_rad_pkt = peap_instance.current_rad_pkt;
			peap_instance.current_rad_pkt = req;
			client_state = peap_instance.client_state =	CLIENT_TLS_HANDSHAKE;
			break;
		case (CLIENT_TLS_ESTABLISHED | CLIENT_PEAP_START):
			free(req);
			req = 0;
			req = rad_alloc(RAD_NEW_VECTOR);
			if (0 == req) {
				peap_log_debug("handle_req_rep: Error while allocating req: state 0x%x \n",
					client_state);
				peap_log_debug("handle_req_rep: Client state 0x%x, SSL state 0x%x \n",
					client_state, SSL_get_state(peap_instance.tls_session->ssl));
				return (-1);
			}
			compose_radius_pkt(req, PW_AUTHENTICATION_REQUEST);
			tls_data = 0;
			tls_data_length = 0;
			peap_instance.prev_rad_pkt = peap_instance.current_rad_pkt;
			peap_instance.current_rad_pkt = req;
			req = compose_eap_packet(req, tls_data, tls_data_length, FALSE);
			if (0 == req) {
				peap_log_debug("handle_req_rep: Error while composing EAP request: state 0x%x \n",
					client_state);
				peap_log_debug("handle_req_rep: Client state 0x%x, SSL state 0x%x \n",
					client_state, SSL_get_state(peap_instance.tls_session->ssl));
				free(req);
				return (-1);
			}
			client_state = peap_instance.client_state = (CLIENT_TLS_ESTABLISHED | CLIENT_SEND_PEAP_IDENTITY);
			break;
		case (CLIENT_TLS_ESTABLISHED | CLIENT_SEND_PEAP_IDENTITY):
			peap_data = process_peap_message(&peap_data_length);
			if (0 == peap_data) {
				peap_log_debug("handle_req_rep: Error while extracting PEAP data: state 0x%x \n",
					client_state);
				peap_log_debug("handle_req_rep: Client state 0x%x, SSL state 0x%x \n",
					client_state, SSL_get_state(peap_instance.tls_session->ssl));
				return (-1);
			}
			if (PW_EAP_IDENTITY != *peap_data) {
				peap_log_debug("handle_req_rep: No PEAP Identity request: state 0x%x \n",
					client_state);
				peap_log_debug("handle_req_rep: Client state 0x%x, SSL state 0x%x \n",
					client_state, SSL_get_state(peap_instance.tls_session->ssl));
				return (-1);
			}
			tls_data_length = 0;
			tls_data = compose_peap_identity(&tls_data_length);
			if (0 == tls_data) {
				peap_log_debug("handle_req_rep: Error while extracting TLS data: state 0x%x \n",
					client_state);
				peap_log_debug("handle_req_rep: Client state 0x%x, SSL state 0x%x \n",
					client_state, SSL_get_state(peap_instance.tls_session->ssl));
				return (-1);
			}
			free(req);
			req = 0;
			req = rad_alloc(RAD_NEW_VECTOR);
			if (0 == req) {
				peap_log_debug("handle_req_rep: Error while allocating req: state 0x%x \n",
					client_state);
				peap_log_debug("handle_req_rep: Client state 0x%x, SSL state 0x%x \n",
					client_state, SSL_get_state(peap_instance.tls_session->ssl));
				free(tls_data);
				return (-1);
			}
			compose_radius_pkt(req, PW_AUTHENTICATION_REQUEST);
			req = compose_eap_packet(req, tls_data, tls_data_length, FALSE);
			if (0 == req) {
				peap_log_debug("handle_req_rep: Error while composing EAP request: state 0x%x \n",
					client_state);
				peap_log_debug("handle_req_rep: Client state 0x%x, SSL state 0x%x \n",
					client_state, SSL_get_state(peap_instance.tls_session->ssl));
				free(req);
				free(tls_data);
				return (-1);
			}
			client_state = peap_instance.client_state =	(CLIENT_TLS_ESTABLISHED | CLIENT_SEND_MSCHAP_RESPONSE);
			break;
		case (CLIENT_TLS_ESTABLISHED | CLIENT_SEND_MSCHAP_RESPONSE):
			peap_data = process_peap_message(&peap_data_length);
			if (0 == peap_data) {
				peap_log_debug("handle_req_rep: Error while extracting PEAP data: state 0x%x \n",
					client_state);
				peap_log_debug("handle_req_rep: Client state 0x%x, SSL state 0x%x \n",
					client_state, SSL_get_state(peap_instance.tls_session->ssl));
				return (-1);
			}
			if (peap_data != 0) {
				peap_log_debug("handle_req_rep: peap_data = %x\n", *peap_data);
				if ((PW_EAP_MSCHAPV2 != *peap_data) && (PW_EAP_REQUEST != *peap_data)) {
					peap_log_debug("handle_req_rep: No PEAP MSCHAP request: state 0x%x \n",
							client_state);
					peap_log_debug("handle_req_rep: Client state 0x%x, SSL state 0x%x \n",
							client_state, SSL_get_state(peap_instance.tls_session->ssl));
					return (-1);
				}
			}
			free(req);
			req = 0;
			req = rad_alloc(RAD_NEW_VECTOR);
			if (0 == req) {
				peap_log_debug("handle_req_rep: Error while allocating req: state 0x%x \n",
					client_state);
				peap_log_debug("handle_req_rep: Client state 0x%x, SSL state 0x%x \n",
					client_state, SSL_get_state(peap_instance.tls_session->ssl));
				return (-1);
			}
			compose_radius_pkt(req, PW_AUTHENTICATION_REQUEST);
			peap_data_length = 0;
			/*
			 * If server response does not have MSCHAPV2 in PEAP data
			 * client has to send NAK response with the supported protocol
			 * details to server. Server on receiving NAK packet, it responds
			 * with its supported protocol.
			 */
			if (PW_EAP_MSCHAPV2 == *peap_data) {
				peap_data = compose_mschap_response(&peap_data_length);
				peap_log_debug("handle_req_rep: Compose MSCHAPv2 response of peap_data_length = %d\n",
						peap_data_length); 
				client_state = peap_instance.client_state =	(CLIENT_TLS_ESTABLISHED | CLIENT_RECEIVE_AUTH_RESULT);
			} else if (PW_EAP_REQUEST == *peap_data) {
				peap_data = compose_nak_respone(&peap_data_length);
				peap_log_debug("handle_req_rep: Compose NAK response of peap_data_length = %d\n",
						peap_data_length); 
				client_state = peap_instance.client_state = (CLIENT_TLS_ESTABLISHED | CLIENT_SEND_NAK_RESPONSE);
			}
	
			if (0 == peap_data) {
				peap_log_debug("handle_req_rep: PEAP data compose failed ");
				return -1;
			}

			tls_data = get_tls_data(peap_data, peap_data_length, &tls_data_length);
			if(0 == tls_data) {
				peap_log_debug("handle_req_rep: Error get_tls_data \n");
				free(req);
				free(peap_data);
				return (-1);
			}
			peap_log_debug("handle_req_rep: get_tls_data returned success\n");
			free(peap_data);
			peap_data = 0;

			req = compose_eap_packet(req, tls_data, tls_data_length, FALSE);
			if (0 == req) {
				peap_log_debug("handle_req_rep: Error while composing EAP request: state 0x%x \n",
					client_state);
				peap_log_debug("handle_req_rep: Client state 0x%x, SSL state 0x%x \n",
					client_state, SSL_get_state(peap_instance.tls_session->ssl));
				free(req);
				return (-1);
			}
			peap_log_debug("handle_req_rep: compose_eap_packet returned success\n");
			peap_log_debug("handle_req_rep: get_tls_data returned success\n");
			peap_log_debug("handle_req_rep: compose_eap_packet returned success\n");
			peap_log_debug("handle_req_rep: Client State = 0x%x\n", client_state);
			break;
		case (CLIENT_TLS_ESTABLISHED | CLIENT_SEND_NAK_RESPONSE):
			client_state = peap_instance.client_state =	(CLIENT_TLS_ESTABLISHED | CLIENT_SEND_MSCHAP_RESPONSE);
			peap_log_debug("handle_req_rep: Client State = 0x%x\n", client_state);
			break;
		case (CLIENT_TLS_ESTABLISHED | CLIENT_RECEIVE_AUTH_RESULT):
			peap_data = process_peap_message(&peap_data_length);
			if (0 == peap_data) {
				peap_log_debug("handle_req_rep: Error while extracting PEAP data: state 0x%x \n",
					client_state);
				peap_log_debug("handle_req_rep: Client state 0x%x, SSL state 0x%x \n",
					client_state, SSL_get_state(peap_instance.tls_session->ssl));
				return (-1);
			}
			if (PW_EAP_MSCHAPV2 != *peap_data) {
				peap_log_debug("handle_req_rep: No PEAP MSCHAP request: state 0x%x \n",
					client_state);
				peap_log_debug("handle_req_rep: Client state 0x%x, SSL state 0x%x \n",
					client_state, SSL_get_state(peap_instance.tls_session->ssl));
				return (-1);
			}
			free(req);
			req = 0;
			req = rad_alloc(RAD_NEW_VECTOR);
			if (0 == req) {
				peap_log_debug("handle_req_rep: Error while allocating req: state 0x%x \n",
					client_state);
				peap_log_debug("handle_req_rep: Client state 0x%x, SSL state 0x%x \n",
					client_state, SSL_get_state(peap_instance.tls_session->ssl));
				return (-1);
			}
			compose_radius_pkt(req, PW_AUTHENTICATION_REQUEST);
			if (PW_EAP_MSCHAPV2_SUCCESS != *(peap_data + 1)) {
				peap_log_debug("handle_req_rep:MSCHAPV2 authentication failure !!!!!!\n");
				return(-1);

			/* Probably we need to take these error checks later into account*/

			/*
			 * return -1 ;
			 */
			/*
			 * Check the failure reason
			 */
			/*
			 * error_code = get_mschapv2_error(peap_data);
			 */
			/*
			 * E=646 ERROR_RESTRICTED_LOGON_HOURS
			 */

			/*
			 * E=647 ERROR_ACCT_DISABLED
			 */

			/*
			 * E=648 ERROR_PASSWD_EXPIRED
			 */
			/*
			 * compose_mschap_changepwd_request();
			 */

			/*
			 * E=649 ERROR_NO_DIALIN_PERMISSION
			 */

			/*
			 * E=691 ERROR_AUTHENTICATION_FAILURE
			 */
			/*
			 * if (is_mschapv2_failure_retryable()) {
			 * compose_mschapv2_response(); }else{
			 * compose_mschapv2_failure_response(); }
			 */

			/*
			 * E=709 ERROR_CHANGING_PASSWORD
			 */

			} else {/*SUCCESS*/
				peap_data_length = 0;
				peap_data = compose_mschap_response(&peap_data_length);
				if (0 == peap_data) {
					peap_log_debug("handle_req_rep:  compose_mschap_response failed ");
					free(req);
					return -1;
				}
			}
			tls_data_length = 0;
			tls_data = get_tls_data(peap_data, peap_data_length, &tls_data_length);
			if(0 == tls_data) {
				peap_log_debug("handle_req_rep: Error get_tls_data \n");
				free(req);
				free(peap_data);
				return (-1);
			}
			free(peap_data);
			peap_data = 0;
			req = compose_eap_packet(req, tls_data, tls_data_length, FALSE);
			if (0 == req) {
				peap_log_debug("handle_req_rep: Error while composing EAP request: state 0x%x \n",
					client_state);
				peap_log_debug("handle_req_rep: Client state 0x%x, SSL state 0x%x \n",
					client_state, SSL_get_state(peap_instance.tls_session->ssl));
				free(req);
				free(peap_data);
				return (-1);
			}
			peap_data_length = 0;
			client_state = peap_instance.client_state =
					(CLIENT_TLS_ESTABLISHED | CLIENT_RECEIVE_FINAL_CHALLENGE);
			break ;
 		case (CLIENT_TLS_ESTABLISHED | CLIENT_RECEIVE_FINAL_CHALLENGE):
			peap_data_length = 0;
			peap_data = 0;
			peap_data = process_peap_message(&peap_data_length);
			if (0 == peap_data) {
				peap_log_debug("handle_req_rep: Error while extracting PEAP data: state 0x%x \n",
					client_state);
				peap_log_debug("handle_req_rep: Client state 0x%x, SSL state 0x%x \n",
					client_state, SSL_get_state (peap_instance.tls_session->ssl));
				return (-1);
			}

			temp_len = sizeof(result_avp);
			temp_len = htons(temp_len);
			result_avp[0] = PW_EAP_RESPONSE; 							/* Response */
			result_avp[1] = *((uint8_t*)(peap_data + 1));/* Id from previous packet  */
			(*(uint16_t*)(result_avp + 2)) = temp_len;/* Byte ordred length */

			result_avp[4] = PW_EAP_EXTENTIONS;/* EAP Extension method */
			result_avp[5] = EAP_LENGTH_INCLUDED;/*Length included */
			result_avp[6] = RESULT_AVP;/* Result AVP */

			(*(uint16_t*)(result_avp + 7)) = htons(RESULT_AVP_LEN);
			/*
			 * Byte ordred length
			 * 7th and 8th Bytes
			 */

			(*(uint16_t*)(result_avp + 9)) = htons(RESULT_AVP_RESULT);
			/* Byte ordred result value
			 * 9th and 10th Bytes
			 */
			peap_data_length = sizeof(result_avp);

			tls_data_length = 0;

			tls_data = get_tls_data(result_avp, peap_data_length,&tls_data_length);
			if(0 == tls_data) {
				peap_log_debug("handle_req_rep: Error get_tls_data \n");
				free(req);
				req = 0;
				free(peap_data);
				peap_data = 0 ;
				return (-1);
			}
			peap_data = 0;
			req = compose_eap_packet(req, tls_data, tls_data_length, FALSE);
			if (0 == req) {
				peap_log_debug("handle_req_rep: Error while composing EAP request: state 0x%x \n",
					client_state);
				peap_log_debug("handle_req_rep: Client state 0x%x, SSL state 0x%x \n",
					client_state, SSL_get_state(peap_instance.tls_session->ssl));
				return (-1);
			}
			peap_data_length = 0;
			client_state = peap_instance.client_state =
					(CLIENT_TLS_ESTABLISHED | CLIENT_TLS_SHUTDOWN);

			break;
		default:
			peap_log_debug("handle_req_rep: Invalid client state 0x%x \n", client_state);
			return (-1);
		} /* end of switch */
		if (exit == TRUE) {
			return ret;
		}

		ret = rad_peap_send(req, NULL, peap_instance.shared_secret);

		if (ret < 0) {
			peap_log_debug("handle_req_rep: Error while sending radius req: state 0x%x \n",
																			client_state);
			if(req->data) free(req->data);
			if(req) free(req);
			return (-1);
		}

		peap_log_debug("handle_req_rep: after rad_peap_send \n");

		retval = wait_for_server_response();
		if (retval < 0) {
			peap_log_debug("handle_req_rep:Error in getting server entry \n");
			return -1;
		}

		if (0 != retval) {
			if (PEAP_SERVER_RETRY == retval) {
				peap_log_debug("handle_req_rep: Retrying the request \n");
				return PEAP_SERVER_RETRY;
			} else if(PAM_AUTHINFO_UNAVAIL == retval) {
				peap_log_debug("handle_req_rep: Server failed to respond\n");
				return PAM_AUTHINFO_UNAVAIL;
			}

		}
		/*
		 * The radius server responded. We should go ahead with
		 * receiving the data
		 */
		peap_log_debug("handle_req_rep: calling rad_peap_recv \n");

		peap_instance.server_rep = rep = rad_peap_recv();
		if (0 == rep) {
			peap_log_debug("handle_req_rep: Error while receiving radius req 0x%x \n",
																			client_state);
			if(req->data) free(req->data);
			if(req) free(req);
			return (-1);
		}
		peap_log_debug("handle_req_rep: calling process_rad_rep \n");
		ret = process_rad_rep();
		if (ret < 0) {
			peap_log_debug("handle_req_rep: Error while processing radius response: state 0x%x \n",
				client_state);
			if(req->data) free(req->data);
			if(req) free(req);
			return (-1);
		}
		peap_log_debug("handle_req_rep: calling extract_eap_message \n");
		ret = extract_eap_message();
		if (ret < 0) {
			peap_log_debug("handle_req_rep: Error while processing EAP message: state 0x%x \n",
				client_state);
			if(req->data) free(req->data);
			if(req) free(req);
			return (-1);
		}
		if((CLIENT_TLS_ESTABLISHED | CLIENT_TLS_SHUTDOWN) == client_state) {
			exit=TRUE;
		}
		peap_log_debug("handle_req_rep: after calling extract_eap_message \n");
	} while (FALSE == exit);

	return (ret);
}


/*
 *int setup_client
 *
 * Description
 *	This function sets up the global variable values
 *	specific to a user and a server.
 * Input
 *	const char * user_name, const char * user_password, radius_conf_t* conf
 *
 * Output
 *	None
 *
 * Return
 *	Retuns  Success or Failure
 */
static int
setup_client(const char * user_name, const char * user_password, radius_conf_t * conf)
{
#undef CLEAN_UP
#define CLEAN_UP()		  				\
	do {							\
		if(peap_instance.shared_secret){		\
		free(peap_instance.shared_secret);		\
			peap_instance.shared_secret = 0;	\
		}						\
		if(peap_instance.user_name) {			\
			free(peap_instance.user_name); 		\
			peap_instance.user_name = 0;		\
		}						\
		if(peap_instance.hostname) {			\
			free(peap_instance.hostname); 		\
			peap_instance.hostname = 0;		\
		}						\
	} while (0);

	struct sockaddr_in sa; 
	char host[MAX_HOSTNAME_LEN];
	char service[20];
	int cert_present = -1;

	radius_server_t * cur_serv = conf->server;

	if (0 == cur_serv) {
		peap_log_debug("setup_client:No server info available");
		return -1 ;
	}
	if (strlen(user_name) >= MAX_NT_USERNAME) {
		peap_log_debug("setup_client:Username length is greater than allowed");
		return -1 ;
	}
	if (strlen(user_password) >= MAX_NT_PASSWORD) {
		peap_log_debug("setup_client:Password length is greater than allowed");
		return -1 ;
	}
	if(strlen(cur_serv->secret) >= MAX_SHARED_SEC_LEN){
		peap_log_debug("setup_client:Shared secret length is greater than allowed");
		return -1 ;
	}
	if ((peap_instance.sockfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		peap_log_debug("setup_client: error while getting the socket");
		return -1;
	}

	peap_instance.dst_port = cur_serv->port;
	peap_instance.retries = cur_serv->retries + 1;

	peap_instance.dst_ipaddr = (uint32_t) (cur_serv->ip.s_addr);

	/* get the remote host name using server IP*/


	memset(service,0,sizeof(service));
	memset(host,0,sizeof(host));
	sa.sin_family = AF_INET;
	sa.sin_port = htons(cur_serv->port);
	sa.sin_addr.s_addr = (uint32_t) (cur_serv->ip.s_addr);

	cert_present    =       access( RADIUS_CA_FILE, F_OK);
	if ( cert_present == SUCCESS) {
		peap_log_debug("setup_client: CA present");
 		if (getnameinfo(&sa, sizeof sa, host, sizeof host, service, sizeof service, NI_NAMEREQD) == EAI_NONAME) {
			printf("cannot find the host's name.\n ");
			return (-1);
		}

		peap_instance.hostname = calloc(strlen(host) + 1, sizeof(uint8_t));
		if (0 == peap_instance.hostname) {
			peap_log_debug("setup_client: Memory allocation failed for hostname");
			return -1;
		}
		memcpy((char *) peap_instance.hostname, host, strlen(host));
		peap_instance.hostname[strlen(host)] = '\0';
	} else {
                peap_log_debug("setup_TLS_session: No cert validation follows.\n");
	}

	peap_instance.shared_secret = calloc(strlen(cur_serv->secret) + 1, sizeof(uint8_t));
	if (0 == peap_instance.shared_secret) {
		peap_log_debug("setup_client: Memory allocation failed for shared_secret");
		return -1;
	}
	memcpy((char *) peap_instance.shared_secret, cur_serv->secret, strlen(cur_serv->secret));
	peap_instance.shared_secret[strlen(cur_serv->secret)] = '\0';

	peap_instance.user_name = calloc(strlen(user_name) + 1, sizeof(uint8_t));
	if (0 == peap_instance.user_name) {
		peap_log_debug("setup_client: Memory allocation failed for user_name");
		CLEAN_UP();
		return -1;
	}
	memcpy((char *) peap_instance.user_name, user_name, strlen(user_name));
	peap_instance.user_name[strlen(user_name)] = '\0';


	peap_instance.user_passwd = calloc(strlen(user_password) + 1, sizeof(uint8_t));
	if (0 == peap_instance.user_passwd) {
		peap_log_debug("setup_client: Memory allocation failed for user_passwd");
		CLEAN_UP();
		return -1;
	}
	memcpy((char *) peap_instance.user_passwd, user_password, strlen(user_password));
	peap_instance.user_passwd[strlen(user_password)] = '\0';

	peap_instance.client_state = CLIENT_INIT_STATE;

	return 0;
}
/*
 * int pam_peap_authenticate
 *
 * Description
 *	This function sets up the global variable values
 *	specific to a user and a server and initialises
 *	the state machine to CLIENT_INIT_STATE;
 * Input
 *  const char * user_name, const char * user_password,
 * 			radius_conf_t * conf, AUTH_HDR *response
 *
 * Output
 *	None
 *
 * Return
 *	Returns  Success or Failure or PAM_AUTHINFO_UNAVAIL when server
 *  is not reachable.
 */
int
pam_peap_authenticate(const char * user_name, const char * user_password,
			radius_conf_t* conf, AUTH_HDR *response)
{
	int ret = 0;
	int found_server = -1;
	RADIUS_PACKET * rep = 0;
	uint8_t * radius_pkt = 0;
	uint16_t rad_pkt_len = 0;
	peap_instance.cur_serv = conf->server;

	ret = setup_client(user_name, user_password, conf);
	if (0 != ret) {
		peap_log_debug("pam_peap_authenticate: setup_client error \n");
		return 7;
	}
	do {
		peap_instance.client_state = CLIENT_INIT_STATE;
		ret = handle_req_rep();
		if (0 == ret) {
			peap_log_debug("pam_peap_authenticate: Authentication succeeded"
				" Getting vendor specific attributes \n");

			free(peap_instance.shared_secret);
			free(peap_instance.user_name);
			free(peap_instance.user_passwd);
			free(peap_instance.tls_session);
			rep = peap_instance.server_rep;
			radius_pkt = rep->data;
			memcpy(&rad_pkt_len, radius_pkt + 2, sizeof(uint16_t));
			/* Entire radius packet length */

			rad_pkt_len = htons(rad_pkt_len);

			peap_log_debug("pam_peap_authenticate: rad_pkt_len is %d \n", rad_pkt_len);
			peap_log_debug("pam_peap_authenticate: Copying the response \n");
			memcpy(response, radius_pkt, rad_pkt_len);

			found_server = 0;
			break;
		}

    } while (PEAP_SERVER_RETRY == ret);

	if(PAM_AUTHINFO_UNAVAIL == ret) {
		peap_log_debug("\npam_peap_authenticate:Authentication Server failed to respond\n");
		return PAM_AUTHINFO_UNAVAIL;
	} else if (ret == X509_V_ERR_HOSTNAME_MISMATCH) {
		return X509_V_ERR_HOSTNAME_MISMATCH;
	} else if (ret == X509_V_ERR_CERT_UNKNOWN_TO_OCSP) {
		return X509_V_ERR_CERT_UNKNOWN_TO_OCSP;
	} else if (ret == X509_V_ERR_CERT_REVOKED_TO_OCSP) {
		return X509_V_ERR_CERT_REVOKED_TO_OCSP;
	} else if (ret == X509_V_ERR_OCSP_APPLICATION_VERIFICATION) {
		return X509_V_ERR_OCSP_APPLICATION_VERIFICATION;
	} 

	if (0 != found_server) {
		peap_log_debug("\npam_peap_authenticate:Authentication Failed : Exiting ...\n");
		if (ret == PAM_AUTH_SECUREID) {
			peap_log_debug("\npam_peap_authenticate:ret %d\n", ret);
			return (PAM_AUTH_SECUREID);
		}
			
		return -1;
	}
	else {
		peap_log_debug("\npam_peap_authenticate:Authentincated Successfully > Bye...\n");
		return 0;
	}
}

