/*
 *    Copyright (c) 1996-2007 Brocade Communications Systems, Inc.
 *    All rights reserved.
 *
 *    Module name: radpeapclient.h
 *
 *    Description:
 *      This modules deals with PEAP-MSCHAPv2 user authentication
 *
 */

#ifndef _RADPEAPCLIENT_H
# define _RADPEAPCLIENT_H

# ifndef ATTR_PACKED
#	define ATTR_PACKED	__attribute__ ((__packed__))
# endif

# define COMMAND_SIZE                   256
# define RESULT_AVP_LEN                2
# define RESULT_AVP_RESULT             1
# define RAD_NEW_VECTOR                1
# define EAP_ID                        210
# define BROCADE_ID                    1588

# define TLS_MSG_HEADER_LEN             6
# define MAX_CHALLENGE_LEN              16
# define MAX_RESPONSE_LEN               24
# define MAX_AUTH_RESPONSE_LEN          20

# define PW_EAP_MSCHAPV2_RESPONSE       2
# define PW_EAP_MSCHAPV2_SUCCESS        3

# define PW_EAP_MSCHAPV2                26 
# define PW_VENDOR_ATTR                 26
# define PW_EAP_EXTENTIONS              33 

# define MSCHAPV2_CHALLENGE_LEN         16
# define MSCHAP_RESPONSE_VALUE_SIZE     49 /* RFC demands it to be 49 */

# define RESULT_AVP                     3
# define CLIENT_INIT_STATE              0x0100
# define CLIENT_EAP_CONNECT             0x0200
# define CLIENT_TLS_HANDSHAKE           0x0400
# define CLIENT_TLS_ESTABLISHED	        0x0800

# define CLIENT_SEND_ACK                0x0001
# define CLIENT_PEAP_START              0x0002
# define CLIENT_SEND_PEAP_IDENTITY      0x0004
# define CLIENT_SEND_MSCHAP_RESPONSE    0x0008
# define CLIENT_RECEIVE_AUTH_RESULT     0x0010
# define CLIENT_RECEIVE_FINAL_CHALLENGE 0x0020
# define CLIENT_TLS_SHUTDOWN            0x0040
# define CLIENT_SEND_NAK_RESPONSE		0x0080

# define RADIUS_ACCESS_REQUEST          0x1
# define RANDOM_ID                      0x90
# define MORE_TLS_DATA                  0x40 
# define EAP_LENGTH_INCLUDED            0x80

# define TLS_HDR_LEN_PLUS_LENGTH        10		/* TLS header length */
# define TLS_HDR_LEN                    6
					/* TLS header length without including the 
					 * tls_msg_len field   
					 */
# define MAX_NT_PASSWORD               	256		/* Maximum Password length */
# define MAX_NT_USERNAME                256		/* Maximum User name length */
# define MAX_HOSTNAME_LEN		1024        	/* Maximum HOST name length */
# define MAX_SHARED_SEC_LEN             256		/* Maximum  shared secret length */

#define IP_ADDR_SIZE                   	4 		/* IPV4 -4 bytes */
#define MS_RESERVED_BYTES               8

# define PEAP_SERVER_RETRY              1

#define PAM_AUTHINFO_UNAVAIL			9
#define EAP_MESSAGE_MAX					253
#define MAX_EAP_PACKETS			26		/* Maximum Radius packet size is 4096.
							   Maximum EAP packets 4096/256=16.
							   Keeping buffer of 10 packets.*/	

#define TLS_CIPHERS                    "!ECDH:!DH:HIGH:-MD5:!CAMELLIA:!SRP:!PSK:!AESGCM:!TLSv1.2"
#define FIPS_SET			1

#define SUCCESS 0

static const char * packet_codes[] = {
  "",
  "Access-Request",
  "Access-Accept",
  "Access-Reject",
  "Accounting-Request",
  "Accounting-Response",
  "Accounting-Status",
  "Password-Request",
  "Password-Accept",
  "Password-Reject",
  "Accounting-Message",
  "Access-Challenge",
  "Status-Server",
  "Status-Client",
  "14",
  "15",
  "16",
  "17",
  "18",
  "19",
  "20",
  "Resource-Free-Request",
  "Resource-Free-Response",
  "Resource-Query-Request",
  "Resource-Query-Response",
  "Alternate-Resource-Reclaim-Request",
  "NAS-Reboot-Request",
  "NAS-Reboot-Response",
  "28",
  "Next-Passcode",
  "New-Pin",
  "Terminate-Session",
  "Password-Expired",
  "Event-Request",
  "Event-Response",
  "35",
  "36",
  "37",
  "38",
  "39",
  "Disconnect-Request",
  "Disconnect-ACK",
  "Disconnect-NAK",
  "CoA-Request",
  "CoA-ACK",
  "CoA-NAK",
  "46",
  "47",
  "48",
  "49",
  "IP-Address-Allocate",
  "IP-Address-Release"
};

/*
 * On wire radius packet header
 * The structure member `data' is used in the program
 * for copying the lower layer header.
 */
typedef struct radius_packet_t {
	uint8_t code;
	uint8_t id;
	uint16_t length;
	uint8_t vector[AUTH_VECTOR_LEN];
	uint8_t data[0];
} ATTR_PACKED radius_packet_t;

/*
 * Radius Attribute-Eap Message structure
 * `data' field is a pointer to the Eap message data portion 
 * Used for copying the real eap message in various packets
 */
typedef struct _eap_message {
	uint8_t	code;
	uint8_t	length;
	uint8_t data[0];
} ATTR_PACKED eap_message_t;

/*
 * Eap packet header
 * Based on the `length' field `type_data' can be of various size
 * So just a pointer is used to copy over the real data 
 */
typedef struct _eap_pkt_t {
	uint8_t	code;
	uint8_t	id;
	uint16_t length;
	uint8_t	type;
	uint8_t type_data[0];
} ATTR_PACKED eap_pkt_t;

/*
 * General TLV format, `data' can be of any length
 * a pointer is used to carry out copiying,
 * its not a actual structure member
 */
typedef struct _type_length_value {
	uint8_t	type;
	uint8_t length;
	uint8_t data[0];
} ATTR_PACKED type_length_value_t;

/*
 * Type-Value pair, used to represent EAP message
 * where there is no `length' field as compared to TLV
 */
typedef struct _type_value {
	uint8_t	type;
	uint8_t data[0];
} ATTR_PACKED type_value_t;

/*
 * TLS layer header, used for communicating to and from 
 * SSL library functions 
 */
typedef struct _eap_tls_response {
	uint8_t	code;
	uint8_t	id;
	uint16_t length;
	uint8_t	type;
	uint8_t	flags;
	uint32_t tls_msg_len;
	uint8_t data[0];
} ATTR_PACKED eap_tls_response_t;

/*
 * EAP TLS acknowledgement packet format
 */
typedef struct _eap_tls_ack_t {
	uint8_t	code;
	uint8_t	id;
	uint16_t length;
	uint8_t	type;
	uint8_t	flags;
} ATTR_PACKED eap_tls_ack_t;

/*
 * SSL protocol version structure, RFC 2246
 */
typedef struct _proto_ver {
	uint8_t	major;
	uint8_t	minor;
} ATTR_PACKED protocol_version_t;

/*
 * SSL content Type enums, content type specifies what type of
 * data SSL packet is containing, RFC 2246
 */
typedef enum {
	CHANGE_CIPHER_SPEC = 20, 
	ALERT,
	HANDSHAKE,
	APPLICATION_DATA
} content_type ;


/*
 * TLS record layer protocol header, RFC 2246 
 */
typedef struct TLS_RECORD_LAYER {
	content_type proto_type;
	protocol_version_t version;
	uint16_t length;
	uint8_t  fragment[1];
} ATTR_PACKED tls_record_layer_t;

/*
 * MSCHAP V2 response header
 * ietf draft Kamath & Palekar
 */
typedef struct _mschapv2_response_packet_t {
	uint8_t opcode;
	uint8_t mschapv2_id;
	uint8_t ms_length[2];
	uint8_t value_size;
	uint8_t peer_challenge[16];
	uint8_t reserved[8];
	uint8_t nt_response[24];
	uint8_t flags;
	uint8_t data[1];
} ATTR_PACKED mschapv2_response_packet_t;

/*
 * MSCHAP V2 request header
 * ietf draft Kamath & Palekar
 */
typedef struct _mschapv2_request_packet_t {
	uint8_t opcode;
	uint8_t mschapv2_id;
	uint8_t ms_length[2];
	uint8_t value_size;
	uint8_t auth_challenge[16];
	uint8_t name[1];
} ATTR_PACKED mschapv2_request_packet_t;

/*
 * Used for maintaining reciened TLS fragments in linked list format,
 * untill we get all the fragments which is needed for reassembly
 * Once all fragments are recieved, free the list
 */
typedef struct _tls_data_list {
	uint8_t * tls_data;
	int tls_data_length;
	struct _tls_data_list * next;
	struct _tls_data_list * list_end;
} tls_data_list_t;

/*
 * The global data structure used for maintaining *all* the 
 * information needed at any point of time during the 
 * execution of the program
 */
typedef struct _peap_instance {
	uint32_t dst_ipaddr;
	u_short dst_port;
	int sockfd;
	char * shared_secret;
	char * user_name;
	char * user_passwd;
	tls_session_t *tls_session;
	RADIUS_PACKET *prev_rad_pkt;
	RADIUS_PACKET *current_rad_pkt;
	RADIUS_PACKET *server_rep;
	uint8_t * srv_eap_message;
	tls_data_list_t srv_tls_data_list;
	uint16_t client_state;
	uint8_t ssl_state;
	void *cur_serv;
	char *hostname;
	int retries;
} peap_instance_t;

#endif 

