/***************************************************************************
 * Copyright (C) 2015 Brocade Communications Systems, Inc.
 * All rights reserved.
 * Author : Soundarya Ravi
 *
 * Brocade specific implementation of shadow style password hashing.
 * Salting and stretching mechanisms are adapted from Unix Crypt 
 * implementation. Openssl APIs are used to generate message digests.
 *************************************************************************/

#define _GNU_SOURCE
#include <stdio.h>
#include <string.h>
#include <openssl/evp.h>
#include <errno.h>
#include <stdint.h>
#include <stdbool.h>
#include <stdlib.h>
#include <sys/param.h>
#include <sys/types.h>
#include <unistd.h>
#include <alloca.h>
#include <time.h>		
#include <sys/time.h>

#include "brocade_hashing.h"

/* Prefix for optional rounds specification.  */
static const char sha2_rounds_prefix[] = "rounds=";

/* Maximum salt string length.  */
#define SALT_LEN_MAX 16
/* Default number of rounds if not explicitly specified.  */
#define ROUNDS_DEFAULT 5000
/* Minimum number of rounds.  */
#define ROUNDS_MIN 1000
/* Maximum number of rounds.  */
#define ROUNDS_MAX 999999999

#define MAGIC_LEN 3
#define MAX_LEN  32
#define SEPARATOR_LEN 1

/* Define Password Key and Initialization Vector */
#define PasswordKey "dsR4sOa7wPa0drabcDe9nmuTCiwArGzy"
#define PasswordIV "vceL6aKdp9asDWSfgEf08JcvSabfGnfs"

void convert_back(unsigned char *src, unsigned char *dst,
					 unsigned int loop, unsigned int rem);
void base642hex(unsigned char *val);
unsigned int base64_decode(unsigned char *src, unsigned char **dst,
							unsigned int src_len);
int decryptPassword(const char *encodedPassword, char *decryptedPassword);

/* Error values returned by Decryption */
enum {
	AUTHSPEC_OK		= 0,
	AUTH_INVALID_SERVER   = -1,
	AUTH_SERVER_UNKNOWN   = -2,
	AUTH_INVALID_SECRET  = -3,
	AUTH_INVALID_TIMEOUT  = -4,
	AUTH_INVALID_PORT_NUM = -5,
	AUTH_INVALID_PROTO = -6,
	AUTH_CONFIG_DUP		= -7,
	AUTH_CONFIG_MAXOUT    = -8,
	AUTH_INVALID_INPUT    = -9,
	AUTH_MALLOC_FAIL	= -10,
	AUTH_CONFIG_READ_ERR  = -11,
	AUTH_CONFIG_WRITE_ERR = -12,
	AUTH_CONFIG_NOT_FOUND = -13,
	AUTH_CONFIG_NOT_ENABLED		= -14,
	AUTH_CONFIG_LAST_SERVER		= -15,
	AUTH_CONFIG_NOT_CHANGED		= -16,
	AUTH_CONFIG_ENABLED		= -17,
	AUTH_LOCAL_CONFIG_ENABLED	= -18,
	AUTH_LOCAL_BACKUP_CONFIG_ENABLED	= -19,
	AUTH_LOCAL_ONLY_CONFIG_ENABLED		= -20,
	AUTH_INVALID_CONFTYPE		= -21,
	AUTH_INVALID_DOMAIN		= -22,
    AUTH_ZERO_LENGTH_SECRET = -23,
    AUTH_ENCRYPT_INIT_FAIL = -24,
    AUTH_ENCRYPT_FINAL_FAIL = -25,
    AUTH_ENCRYPT_UPDATE_FAIL = -26,
    AUTH_ENCRYPT_BASE64_ENCODE_FAIL = -27,
    AUTH_DECRYPT_INIT_FAIL = -28,
    AUTH_DECRYPT_FINAL_FAIL = -29,
    AUTH_DECRYPT_UPDATE_FAIL = -30,
    AUTH_DECRYPT_BASE64_DECODE_FAIL = -31
} secAuthErrCode;

static const struct
{
  const char *hashtype;
  const char *prefix;
  int len;
  int needed_len;
  int rounds_len;
} hash_info[]=
{
  {"md5", "$1$", 16, 22, 0},
  {"sha256", "$5$", 32, 43, sizeof (sha2_rounds_prefix)},
  {"sha512", "$6$", 64, 86, sizeof (sha2_rounds_prefix)}
};

/* Table with characters for base64 transformation.  */
static const char b64t[64] =
"./0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";

static void to64(char *s, unsigned long v, int n)
{
	while (--n >= 0) {
		*s++ = b64t[v & 0x3f];
		v >>= 6;
	}
}

/*
 * i64c - convert an integer to a radix 64 character
 */
static int i64c(int i)
{
	if (i < 0)
		return ('.');
	else if (i > 63)
		return ('z');
	if (i == 0)
		return ('.');
	if (i == 1)
		return ('/');
	if (i >= 2 && i <= 11)
		return ('0' - 2 + i);
	if (i >= 12 && i <= 37)
		return ('A' - 12 + i);
	if (i >= 38 && i <= 63)
		return ('a' - 38 + i);
	return ('\0');
}

static char*
md5_crypt_r(const char *pw, const char *salt, char *buffer, int buflen, const char* hashtype, const char *magic, int md5_hash_len)
{
	static char passwd[120], *p; 
	static const char *sp, *ep;
	unsigned char final[EVP_MAX_MD_SIZE];
	char *cp;
	int sl, pl, i, j;
	unsigned long l;
	size_t pw_len;
	size_t rounds = ROUNDS_MIN;
	unsigned int md_len;
        EVP_MD_CTX *ctx;
        EVP_MD_CTX *ctx1;
        const EVP_MD *md;

        OpenSSL_add_all_digests();
        ctx = EVP_MD_CTX_create();
        ctx1 = EVP_MD_CTX_create();
        md = EVP_get_digestbyname(hashtype);

	/* Refine the Salt first */
	sp = salt;

	/* If it starts with the magic string, then skip that */
	if (!strncmp(sp, magic, strlen(magic)))
		sp += strlen(magic);

	/* It stops at the first '$', max 8 chars */
	for (ep = sp; *ep && *ep != '$' && ep < (sp + 8); ep++)
		continue;

	/* get the length of the true salt */
	sl = ep - sp;

	pw_len = strlen(pw);

	EVP_DigestInit_ex(ctx, md, NULL); 

	/* The password first, since that is what is most unknown */
	EVP_DigestUpdate(ctx, pw, pw_len);

	/* Then our magic string */
	EVP_DigestUpdate(ctx, magic, strlen(magic));

	/* Then the raw salt */
	EVP_DigestUpdate(ctx, sp, sl);

	/* Then just as many characters of the MD5(pw,salt,pw) */
	EVP_DigestInit_ex(ctx1, md, NULL); 
	EVP_DigestUpdate(ctx1, pw, pw_len);
	EVP_DigestUpdate(ctx1, sp, sl);
	EVP_DigestUpdate(ctx1, pw, pw_len);
	EVP_DigestFinal_ex(ctx1, final, &md_len);

	for (pl = pw_len; pl > 0; pl -= md5_hash_len)
		EVP_DigestUpdate(ctx, final, pl>md5_hash_len ? md5_hash_len : pl);

	/* Don't leave anything around in vm they could use. */
	memset(final, 0, sizeof final);

	/* Then something really weird... */
	for (j = 0, i = pw_len; i; i >>= 1)
		if (i & 1)
			EVP_DigestUpdate(ctx, final+j, 1);
		else
			EVP_DigestUpdate(ctx, pw+j, 1);

	/* Now make the output string */
	strcpy(passwd, magic);
	strncat(passwd, sp, sl);
	strcat(passwd, "$");

	EVP_DigestFinal_ex(ctx, final, &md_len);

	/*
	 * and now, just to make sure things don't run too fast
	 * On a 60 Mhz Pentium this takes 34 msec, so you would
	 * need 30 seconds to build a 1000 entry dictionary...
	 */
	for (i = 0; i < rounds; i++) {
		EVP_DigestInit_ex(ctx1, md, NULL); 
		if (i & 1)
			EVP_DigestUpdate(ctx1, pw, pw_len);
		else
			EVP_DigestUpdate(ctx1, final, md5_hash_len);

		if (i % 3)
			EVP_DigestUpdate(ctx1, sp, sl);

		if (i % 7)
			EVP_DigestUpdate(ctx1, pw, pw_len);

		if (i & 1)
			EVP_DigestUpdate(ctx1, final, md5_hash_len);
		else
			EVP_DigestUpdate(ctx1, pw, pw_len);
		EVP_DigestFinal_ex(ctx1, final, &md_len);
	}

	p = passwd + strlen(passwd);

	l = (final[0] << 16) | (final[6] << 8) | final[12];
	to64(p, l, 4);
	p += 4;
	l = (final[1] << 16) | (final[7] << 8) | final[13];
	to64(p, l, 4);
	p += 4;
	l = (final[2] << 16) | (final[8] << 8) | final[14];
	to64(p, l, 4);
	p += 4;
	l = (final[3] << 16) | (final[9] << 8) | final[15];
	to64(p, l, 4);
	p += 4;
	l = (final[4] << 16) | (final[10] << 8) | final[5];
	to64(p, l, 4);
	p += 4;
	l = final[11];
	to64(p, l, 2);
	p += 2;
	*p = '\0';

	/* Don't leave anything around in vm they could use. */
	EVP_MD_CTX_destroy(ctx);
  	EVP_MD_CTX_destroy(ctx1);
	memset(final, 0, sizeof final);

	 cp = stpcpy(buffer,passwd);
	*cp = '\0'; 
	
	return buffer;
}


static char *
sha2_crypt_r (const char *key, const char *salt, char *buffer, int buflen, const char* hashtype, const char *sha2_salt_prefix, int sha2_hash_len)
{
  /*printf("crypt_r params %s %s %s %s %d\n", key, salt, hashtype, sha2_salt_prefix, sha2_hash_len); */
 
  EVP_MD_CTX *ctx;
  EVP_MD_CTX *alt_ctx;
  const EVP_MD *md;
  unsigned char alt_result[EVP_MAX_MD_SIZE];
  unsigned char temp_result[EVP_MAX_MD_SIZE];  
  unsigned int md_len;

  size_t salt_len;
  size_t key_len;
  size_t cnt;
  char *cp;
  char *copied_key = NULL;
  char *copied_salt = NULL;
  char *p_bytes;
  char *s_bytes;
  /* Default number of rounds.  */
  size_t rounds = ROUNDS_DEFAULT;
  bool rounds_custom = false;

  OpenSSL_add_all_digests();
  ctx = EVP_MD_CTX_create();
  alt_ctx = EVP_MD_CTX_create();
  md = EVP_get_digestbyname(hashtype);

  /* printf("%s %s\n", key, salt); */
  
  /* Find beginning of salt string.  The prefix should normally always
     be present.  Just in case it is not.  */
  if (strncmp (sha2_salt_prefix, salt, strlen (sha2_salt_prefix)) == 0)
    /* Skip salt prefix.  */
    salt += strlen (sha2_salt_prefix);

  if (strncmp (salt, sha2_rounds_prefix, sizeof (sha2_rounds_prefix) - 1)
      == 0)
    {
      const char *num = salt + sizeof (sha2_rounds_prefix) - 1;
      char *endp;
      unsigned long int srounds = strtoul (num, &endp, 10);
      if (*endp == '$')
	{
	  salt = endp + 1;
	  rounds = MAX (ROUNDS_MIN, MIN (srounds, ROUNDS_MAX));
	  rounds_custom = true;
	}
    }

  salt_len = MIN (strcspn (salt, "$"), SALT_LEN_MAX);
  key_len = strlen (key);
  /* printf("%s %s %d\n", key, salt, salt_len); */

  if(sha2_hash_len == 32)
  {
  if ((key - (char *) 0) % __alignof__ (uint32_t) != 0)
    {
      char *tmp = (char *) alloca (key_len + __alignof__ (uint32_t));
      key = copied_key =
	memcpy (tmp + __alignof__ (uint32_t)
		- (tmp - (char *) 0) % __alignof__ (uint32_t),
		key, key_len);
    }

  if ((salt - (char *) 0) % __alignof__ (uint32_t) != 0)
    {
      char *tmp = (char *) alloca (salt_len + __alignof__ (uint32_t));
      salt = copied_salt =
	memcpy (tmp + __alignof__ (uint32_t)
		- (tmp - (char *) 0) % __alignof__ (uint32_t),
		salt, salt_len);
    }
  }

  if(sha2_hash_len == 64)
  {
  if ((key - (char *) 0) % __alignof__ (uint64_t) != 0)
    {
      char *tmp = (char *) alloca (key_len + __alignof__ (uint64_t));
      key = copied_key =
	memcpy (tmp + __alignof__ (uint64_t)
		- (tmp - (char *) 0) % __alignof__ (uint64_t),
		key, key_len);
    }

  if ((salt - (char *) 0) % __alignof__ (uint64_t) != 0)
    {
      char *tmp = (char *) alloca (salt_len + __alignof__ (uint64_t));
      salt = copied_salt =
	memcpy (tmp + __alignof__ (uint64_t)
		- (tmp - (char *) 0) % __alignof__ (uint64_t),
		salt, salt_len);
    }
 }
  /* printf("%s\n",salt); */
  
  /* Prepare for the real work.  */
  EVP_DigestInit_ex(ctx, md, NULL);

  /* Add the key string.  */
  EVP_DigestUpdate(ctx, key, key_len);

  /* The last part is the salt string.  This must be at most 16
     characters and it ends at the first `$' character (for
     compatibility with existing implementations).  */
  EVP_DigestUpdate(ctx, salt, salt_len);

  /* Compute alternate SHA256 sum with input KEY, SALT, and KEY.  The
     final result will be added to the first context.  */
  EVP_DigestInit_ex(alt_ctx, md, NULL);

  /* Add key.  */
  EVP_DigestUpdate(alt_ctx, key, key_len);

  /* Add salt.  */
  EVP_DigestUpdate(alt_ctx, salt, salt_len);

  /* Add key again.  */
  EVP_DigestUpdate(alt_ctx, key, key_len);

  /* Now get result of this (sha2_hash_len bytes) and add it to the other
     context.  */
  EVP_DigestFinal_ex(alt_ctx, alt_result, &md_len);
  
  /* Add for any character in the key one byte of the alternate sum.  */
  for (cnt = key_len; cnt > sha2_hash_len; cnt -= sha2_hash_len)
    EVP_DigestUpdate(ctx, alt_result, sha2_hash_len);
  EVP_DigestUpdate(ctx, alt_result, cnt);

  /* Take the binary representation of the length of the key and for every
     1 add the alternate sum, for every 0 the key.  */
  for (cnt = key_len; cnt > 0; cnt >>= 1)
    if ((cnt & 1) != 0)
      EVP_DigestUpdate(ctx, alt_result, sha2_hash_len);
    else
      EVP_DigestUpdate(ctx, key, key_len);

  /* Create intermediate result.  */
  EVP_DigestFinal_ex(ctx, alt_result, &md_len);

  /* Start computation of P byggte sequence.  */
  EVP_DigestInit_ex(alt_ctx, md, NULL);

  /* For every character in the password add the entire password.  */
  for (cnt = 0; cnt < key_len; ++cnt)
    EVP_DigestUpdate(alt_ctx, key, key_len);

  /* Finish the digest.  */
  EVP_DigestFinal_ex(alt_ctx, temp_result, &md_len);

  /* Create byte sequence P.  */
  cp = p_bytes = alloca (key_len);
  for (cnt = key_len; cnt >= sha2_hash_len; cnt -= sha2_hash_len)
    cp = mempcpy (cp, temp_result, sha2_hash_len);
  memcpy (cp, temp_result, cnt);

  /* Start computation of S byte sequence.  */
  EVP_DigestInit_ex(alt_ctx, md, NULL);

  /* For every character in the password add the entire password.  */
  for (cnt = 0; cnt < 16 + alt_result[0]; ++cnt)
    EVP_DigestUpdate(alt_ctx, salt, salt_len);

  /* Finish the digest.  */
  EVP_DigestFinal_ex(alt_ctx, temp_result, &md_len);

  /* Create byte sequence S.  */
  cp = s_bytes = alloca (salt_len);
  for (cnt = salt_len; cnt >= sha2_hash_len; cnt -= sha2_hash_len)
    cp = mempcpy (cp, temp_result, sha2_hash_len);
  memcpy (cp, temp_result, cnt);

  /* Repeatedly run the collected hash value through SHA256 to burn
     CPU cycles.  */
  for (cnt = 0; cnt < rounds; ++cnt)
    {
      /* New context.  */
      EVP_DigestInit_ex(ctx, md, NULL);

      /* Add key or last result.  */
      if ((cnt & 1) != 0)
	EVP_DigestUpdate(ctx, p_bytes, key_len);
      else
	EVP_DigestUpdate(ctx, alt_result, sha2_hash_len);

      /* Add salt for numbers not divisible by 3.  */
      if (cnt % 3 != 0)
	EVP_DigestUpdate(ctx, s_bytes, salt_len);

      /* Add key for numbers not divisible by 7.  */
      if (cnt % 7 != 0)
	EVP_DigestUpdate(ctx, p_bytes, key_len);

      /* Add key or last result.  */
      if ((cnt & 1) != 0)
	EVP_DigestUpdate(ctx, alt_result, sha2_hash_len);
      else
	EVP_DigestUpdate(ctx, p_bytes, key_len);

      /* Create intermediate result.  */
      EVP_DigestFinal_ex(ctx, alt_result, &md_len);
    }

  /* Now we can construct the result string.  It consists of three
     parts.  */
  cp = stpncpy (buffer, sha2_salt_prefix, MAX (0, buflen));
  buflen -= strlen (sha2_salt_prefix);

  if (rounds_custom)
    {
      int n = snprintf (cp, MAX (0, buflen), "%s%lu$",
			sha2_rounds_prefix, (unsigned long)rounds);
      cp += n;
      buflen -= n;
    }

  cp = stpncpy (cp, salt, MIN ((size_t) MAX (0, buflen), salt_len));
  buflen -= MIN ((size_t) MAX (0, buflen), salt_len);

  if (buflen > 0)
    {
      *cp++ = '$';
      --buflen;
    }

  #define b64_from_24bit(B2, B1, B0, N)					      \
  do {									      \
    unsigned int w = ((B2) << 16) | ((B1) << 8) | (B0);			      \
    int n = (N);							      \
    while (n-- > 0 && buflen > 0)					      \
      {									      \
	*cp++ = b64t[w & 0x3f];						      \
	--buflen;							      \
	w >>= 6;							      \
      }									      \
  } while (0)

  if(sha2_hash_len == 32)
  {
  b64_from_24bit (alt_result[0], alt_result[10], alt_result[20], 4);
  b64_from_24bit (alt_result[21], alt_result[1], alt_result[11], 4);
  b64_from_24bit (alt_result[12], alt_result[22], alt_result[2], 4);
  b64_from_24bit (alt_result[3], alt_result[13], alt_result[23], 4);
  b64_from_24bit (alt_result[24], alt_result[4], alt_result[14], 4);
  b64_from_24bit (alt_result[15], alt_result[25], alt_result[5], 4);
  b64_from_24bit (alt_result[6], alt_result[16], alt_result[26], 4);
  b64_from_24bit (alt_result[27], alt_result[7], alt_result[17], 4);
  b64_from_24bit (alt_result[18], alt_result[28], alt_result[8], 4);
  b64_from_24bit (alt_result[9], alt_result[19], alt_result[29], 4);
  b64_from_24bit (0, alt_result[31], alt_result[30], 3);
  }
 
  if(sha2_hash_len == 64)
  {
  b64_from_24bit (alt_result[0], alt_result[21], alt_result[42], 4);
  b64_from_24bit (alt_result[22], alt_result[43], alt_result[1], 4);
  b64_from_24bit (alt_result[44], alt_result[2], alt_result[23], 4);
  b64_from_24bit (alt_result[3], alt_result[24], alt_result[45], 4);
  b64_from_24bit (alt_result[25], alt_result[46], alt_result[4], 4);
  b64_from_24bit (alt_result[47], alt_result[5], alt_result[26], 4);
  b64_from_24bit (alt_result[6], alt_result[27], alt_result[48], 4);
  b64_from_24bit (alt_result[28], alt_result[49], alt_result[7], 4);
  b64_from_24bit (alt_result[50], alt_result[8], alt_result[29], 4);
  b64_from_24bit (alt_result[9], alt_result[30], alt_result[51], 4);
  b64_from_24bit (alt_result[31], alt_result[52], alt_result[10], 4);
  b64_from_24bit (alt_result[53], alt_result[11], alt_result[32], 4);
  b64_from_24bit (alt_result[12], alt_result[33], alt_result[54], 4);
  b64_from_24bit (alt_result[34], alt_result[55], alt_result[13], 4);
  b64_from_24bit (alt_result[56], alt_result[14], alt_result[35], 4);
  b64_from_24bit (alt_result[15], alt_result[36], alt_result[57], 4);
  b64_from_24bit (alt_result[37], alt_result[58], alt_result[16], 4);
  b64_from_24bit (alt_result[59], alt_result[17], alt_result[38], 4);
  b64_from_24bit (alt_result[18], alt_result[39], alt_result[60], 4);
  b64_from_24bit (alt_result[40], alt_result[61], alt_result[19], 4);
  b64_from_24bit (alt_result[62], alt_result[20], alt_result[41], 4);
  b64_from_24bit (0, 0, alt_result[63], 2);  
  }

  if (buflen <= 0)
    {
      errno = ERANGE;
      buffer = NULL;
    }
  else
    *cp = '\0';		/* Terminate the string.  */

  /* Clear the buffer for the intermediate result so that people
     attaching to processes or reading core dumps cannot get any
     information.  We do it in this way to clear correct_words[]
     inside the SHA256 implementation as well.  */
  EVP_MD_CTX_destroy(ctx);
  EVP_MD_CTX_destroy(alt_ctx);
  memset (temp_result, '\0', sizeof (temp_result));
  memset (p_bytes, '\0', key_len);
  memset (s_bytes, '\0', salt_len);
  if (copied_key != NULL)
    memset (copied_key, '\0', key_len);
  if (copied_salt != NULL)
    memset (copied_salt, '\0', salt_len);

  return buffer;
}

static char *brcd_salt (const char *hashtype, const char *salt_prefix)
{
	struct timeval tv;
	EVP_MD_CTX *ctx;
	const EVP_MD *md;
	unsigned char final[EVP_MAX_MD_SIZE];
	unsigned int md_len;
	static char salt[MAX_LEN];
	int i;

	char *cp;
	OpenSSL_add_all_digests();
	ctx =  EVP_MD_CTX_create();
	md = EVP_get_digestbyname(hashtype);

	EVP_DigestInit_ex(ctx, md, NULL);
	gettimeofday(&tv, (struct timezone *) 0);
	EVP_DigestUpdate(ctx, (void *) &tv, sizeof tv);
	i = getpid();
	EVP_DigestUpdate(ctx, (void *) &i, sizeof i);
	i = clock();
	EVP_DigestUpdate(ctx, (void *) &i, sizeof i);
	EVP_DigestUpdate(ctx, salt, sizeof salt);
	EVP_DigestFinal_ex(ctx, final, &md_len);
	cp = stpncpy(salt, salt_prefix, MAX_LEN);	/* magic $x$ for the hashtype */
	for (i = 0; i < 16; i++)
		*cp++ = i64c(final[i] & 077);
	*cp = '\0';
	/* printf("|| Prefixed Generated Salt: %s",salt); */
	return salt;
}

/* This entry point is equivalent to the `crypt' function in Unix
   libcs.  */
char *
brcd_crypt (const char *key, const char *salt, const char *hashtype)
{
  /* We don't want to have an arbitrary limit in the size of the
     password.  We can compute an upper bound for the size of the
     result in advance and so we can prepare the buffer we pass to
     `sha256_crypt_r'.  */
  int buflen;
  const char *salt_prefix;
  char *new_buffer = NULL;
  int cnt = 3, i = 0, hash_len = 0, needed_len = 0, needed = 0, rounds_len = 0;

  for (i=0; i < cnt; i++)
  {
    if (strcmp (hashtype, hash_info[i].hashtype) == 0)
    {
      salt_prefix = hash_info[i].prefix;
      hash_len = hash_info[i].len;
      needed_len = hash_info[i].needed_len;
      rounds_len = hash_info[i].rounds_len;
    }
  }

  if(salt == NULL)
  {
	salt = brcd_salt(hashtype, salt_prefix);
  }
  
  needed = (sizeof (salt_prefix) - 1
		+ rounds_len + 9 + 1
		+ strlen (salt) + 1 + needed_len + 1);

  new_buffer = (char *) realloc (new_buffer, needed);
  if (new_buffer == NULL)
	return NULL;

  buflen = needed;
  
  if (strcmp (hashtype, "md5") == 0)
  {
    return md5_crypt_r (key, salt, new_buffer, buflen, hashtype, salt_prefix, hash_len);
  }
  else
  { /* printf("params %s %s %s %s %d\n", key, salt, hashtype, sha2_salt_prefix, sha2_hash_len); */
    return sha2_crypt_r (key, salt, new_buffer, buflen, hashtype, salt_prefix, hash_len);
  }
}

static const struct
{
  const char *salt;
  const char *input;
  const char *hashtype;
} test[] =
{
  { NULL, "fibranne", "md5" },
  { NULL, "fibranne", "sha256" },
  { NULL, "fibranne", "sha512" },
  { "$1$salt", "fibranne", "md5" },
  { "$5$salt", "fibranne", "sha256" },
  { "$6$salt", "fibranne", "sha512" },
  { "$1$CbSMIAL.$EjXJw1wnNm2IZPpMrhgyg1", "fibranne", "md5"},
  { "$5$puGzGeV.q/O2uKoZ$cThYwRYrvoNdILkA8vQ8vmCMam5M45m0ppm3Wztbt70", "fibranne", "sha256"},
  { "$6$XTzSUqGPDm.Tr1xV$xM/UbqwkowI/cokGkTwGdpfsn/jSPkld7k5AqiEyJGbYsnEZKEI0ptSR5Tfv9DDBbsBRvQRXu8v8zFyfnCaLH1", "fibranne", "sha512"}
};
#define ntest (sizeof (test) / sizeof (test[0]))

int main(void)
{
	int i = 0;
	char *cp;
        for (i = 0;i < (int) ntest; ++i)
        {
	printf ("Hashtype: %s || Input: %s || Prefixed Input Salt: %s ", test[i].hashtype, test[i].input, test[i].salt);
	cp = brcd_crypt (test[i].input, test[i].salt, test[i].hashtype);
	printf ("\nHash: %s\n\n", cp);
        }

	return 0;
}

/* Helper functions used by decryptPassword for decryption */
void convert_back(unsigned char *src, unsigned char *dst,
					unsigned int loop, unsigned int rem)
{
    if (rem > 2) {
        dst[loop] = (((src[0] & 0x20) << 2)| ((src[0] & 0x10) << 2) |
                ((src[0] & 0x08) << 2)|
            ((src[0] & 0x04) << 2) | ((src[0] & 0x02) << 2) |
            ((src[0] & 0x01) << 2) |
            ((src[1] & 0x20) >> 4) | ((src[1] & 0x10) >> 4));

        dst[loop + 1] = (((src[1] & 0x08) << 4) | ((src[1] & 0x04) << 4) |
                ((src[1] & 0x02) << 4) |
            ((src[1] & 0x01) << 4) | ((src[2] & 0x20) >> 2) |
            ((src[2] & 0x10) >> 2) |
            ((src[2] & 0x08) >> 2) | ((src[2] & 0x04) >> 2));

        dst[loop + 2] = (((src[2] & 0x02) << 6) | ((src[2] & 0x01) << 6) |

                (src[3] & 0x20) |
            (src[3] & 0x10) | (src[3] & 0x08) | (src[3] & 0x04) |
            (src[3] & 0x02) | (src[3] & 0x01));

    } else if (rem == 2) {
        dst[loop] = (((src[0] & 0x20) << 2)| ((src[0] & 0x10) << 2) |
                ((src[0] & 0x08) << 2)|
            ((src[0] & 0x04) << 2) | ((src[0] & 0x02) << 2) |
            ((src[0] & 0x01) << 2) |
            ((src[1] & 0x20) >> 4) | ((src[1] & 0x10) >> 4));

        dst[loop + 1] = (((src[1] & 0x08) << 4) | ((src[1] & 0x04) << 4) |
                ((src[1] & 0x02) << 4) |
            ((src[1] & 0x01) << 4) | ((src[2] & 0x20) >> 2) |
            ((src[2] & 0x10) >> 2) |
            ((src[2] & 0x08) >> 2) | ((src[2] & 0x04) >> 2));

    } else if (rem == 1) {
        dst[loop] = (((src[0] & 0x20) << 2)| ((src[0] & 0x10) << 2) |
                ((src[0] & 0x08) << 2)|
            ((src[0] & 0x04) << 2) | ((src[0] & 0x02) << 2) |
            ((src[0] & 0x01) << 2) |
            ((src[1] & 0x20) >> 4) | ((src[1] & 0x10) >> 4));
    }
}

void
base642hex(unsigned char *val)
{
    unsigned char tmp = *val;

    if ((tmp >= 'A') && (tmp <= 'Z')) {
        *val = tmp - 'A';
    } else if ((tmp >= 'a') && (tmp <= 'z')) {
        *val = ((tmp - 'a') + 26);
    } else if ((tmp >= '0') && (tmp <= '9')) {
        *val = ((tmp - '0') + 52);
    } else if (tmp == '+') {
        *val = 62;
    } else if (tmp == '/') {
        *val = 63;
    } else if (tmp == '=') {
        tmp = 0;
    }
}

unsigned int base64_decode(unsigned char *src, unsigned char **dst,
							unsigned int src_len)
{
    unsigned int dst_len;
    unsigned int rem;
    unsigned int src_loop = 0;
    unsigned int dst_loop = 0;

    if ((*(src + src_len - 1) == '=') && (*(src + src_len - 2) == '=')) {
        rem = 2;
    } else if ((*(src + src_len - 1) == '=')) {
        rem = 1;
    } else {
       rem = 0;
    }

    for (src_loop = 0; src_loop < src_len; src_loop++) {
        base642hex((src + src_loop));
    }
    src_loop = 0;

    dst_len = (((src_len / 4) * 3) - rem);

    *dst = calloc((dst_len + 1), sizeof (unsigned char));

    if (!(*dst)) {
        printf("Failed to get memory\n");
        return (0);
    }

    while (src_loop < src_len) {
        convert_back((src + src_loop), *dst, dst_loop, (dst_len - dst_loop));
        src_loop += 4;
        dst_loop += 3;
    }

    dst_loop = 0;
    src_loop = 0;

    return (dst_len);
}

/*
 * This function does the base64 decoding and decrypt the password.
 */
int
decryptPassword(const char *encodedPassword, char *decryptedPassword)
{
    unsigned char *passwd = NULL, *decodedBuf = NULL;
    int declen, padlen = 0;
    EVP_CIPHER_CTX ctx;
    int dstlen = 0, rc = 0;
    int len = strlen(encodedPassword);

	if (len == 0) {
        printf("Decryption failed: %d", AUTH_ZERO_LENGTH_SECRET);
        return (-1);
    }

    /*
     * base64 data arrived here. We should decode it  and get encrypted
     * data.
     */
    dstlen = base64_decode((unsigned char *)encodedPassword, &decodedBuf,
						    strlen(encodedPassword));
    if (!decodedBuf) {
        printf("Decryption failed: %d", AUTH_DECRYPT_BASE64_DECODE_FAIL);
        return (-1);
    }
    /* Initializes the cipher context ctx. */
    memset(&ctx, 0, sizeof (EVP_CIPHER_CTX));
    EVP_CIPHER_CTX_init(&ctx);

    /*
     * Plain text length is  always equal or lesser than cipher text.
     * Cipher text length is always equal or lesser than cipher text.
     * So we are safe to allocate the length.
     */
    passwd = (unsigned char *)calloc(sizeof (unsigned char *), len);
    if (!passwd) {
        printf("Decryption failed: %d", AUTH_MALLOC_FAIL);
        EVP_CIPHER_CTX_cleanup(&ctx);
        free(decodedBuf);
        return (-1);
    }


    /*
     * Sets up the cipher context "ctx" for decryption.
     * Third parameter is Engine, where we can provide our own
     * cipher/digest implementations, which we are not doing anyway.
     * All the remaining parameters are self explanatory except the
     * third parameter.
     */
    rc = EVP_DecryptInit_ex(&ctx, EVP_aes_256_cbc(), NULL,
                            (const unsigned char *) PasswordKey,
                            (const unsigned char *) PasswordIV);
    if (rc) {
        rc = EVP_DecryptUpdate(&ctx, passwd, &declen, decodedBuf, dstlen);
        if (rc) {
            rc = EVP_DecryptFinal_ex(&ctx, passwd + declen, &padlen);
            if (rc) {
                passwd[declen + padlen] = '\0';
            } else {
				printf("Decryption failed: %d", AUTH_DECRYPT_FINAL_FAIL);
				free(decodedBuf);
				free(passwd);
			    return (-1);
            }
        } else {
			printf("Decryption failed: %d", AUTH_DECRYPT_UPDATE_FAIL);
            free(decodedBuf);
            free(passwd);
			return (-1);
        }
    } else {
        printf("Decryption failed: %d", AUTH_DECRYPT_INIT_FAIL);
        free(decodedBuf);
        free(passwd);
		return (-1);
    }

    /* Cleanup the cipher context. */
    EVP_CIPHER_CTX_cleanup(&ctx);

    /* Freeing the decoded buf. */
    free(decodedBuf);

    /* Return the decoded string */
    strcpy(decryptedPassword, (char *)passwd);
   /* decryptedPassword = (char *)passwd;  */
    free(passwd);

    return (0);
}
